	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	BDGRD - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; BDGRD - Server/client Queue tests for MMPS
	;; this test will test a mechanism of the
	;; asynchronous queues. These queues are
	;; anonymous.

	;; The first instance of BDGRD will create the
	;; queue qBDGRD. It has the read access and
	;; becomes the server. Only one server may
	;; exists, because MMPS will fails to create
	;; another queue of the same name. All messages
	;; read from the queue will be written to the
	;; out descriptor D.

	;; All the following instances will try to
	;; open the queue qBDGRD. They will have the
	;; write access, and become client.
	;; Client will read from its in descriptor
	;; and write to the server through the queue.

	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

BDGRD:
	.LOCAL

	;; MMPS executable header.
	;;
	.BYTE
	BLDHEADER	BDGRD.asm$$._length 01 0000

	;;
	.CODE
	;; The process enters here
	;;   BC = global address
	;;   D in and E out descriptor
	;;   HL = argument address

	;; Expected to be server, so put 's' in H
	;;
	LD	H,$s

	;; Server queue name in BC
	;;
	LD	BC,PC
	LDA	*_*qName
	ADD	BC
	LD	L,01
	;; Create a queue
	;;
	MMPS	krn.QUE
	JR	C,doalloc

	;; Unable to create a queue. Not server
	;; So client just tres to open it
	;;
	LD	E,04
	MMPS	krn.OPN
	JR	NC,outerr

	;; Unable to create a queue. Not server
	;; So client just tries to open it
	;;
	LD	E,04
	MMPS	krn.OPN
	JR	NC,outerr

	;; We are client, so put 'c' in H
	;;
	LD	H,$c

	;; If we open the queue, we read from D
	;; and write to the queue. So exchange
	;; If we open the queue, we read from D
	;; and write to the queue. So exchange
	;; the D and E descriptors
	;;
	LDA	D
	STA	L
	LDA	E
	STA	D
	LDA	L
	STA	E

doalloc:
	;; Ok. So we need a page (64 bytes) as
	;; a buffer to receive and send messages.
	;; We will allocate it.
	;; Request 1 page in L and the address
	;; of the block is returned into BC
	;;
	LD	L,01
	MMPS	krn.BAL
	JR	NC,outerr

forever:
	;; Main loop: wait from descriptor E
	;; MMPS will suspend for wait only if E is
	;; a queue else we return with CARRY set
	;;
	MMPS	krn.WAI
	JR	NC,outerr

	;; Get our PID to build PID to hexa digit
	;; and add 's' or 'c'
	;;
	PUSH	HL
	MMPS	krn.PID
	LDA	H
	RCF
	ADC	$0
	STI	(BC)
	POP	HL
	LDA	H
	STI	(BC)

	;; Because in header we request 1 page
	;; for global area, so 1 page = 64 bytes
	;; Prepare to read 63 bytes max from
	;; descriptor R
	LD	L,3D
	MMPS	krn.REA
	JR	NC,outerr

	PUSH	DE
	;; Now we exchange D and E to write on
	;; the out descriptor
	;;
	LDA	D
	STA	E

	;; Refresh the output. Nice if E is a
	;; console...
	MMPS	krn.RFR

	;; Write the received buffer, L contains
	;; the number of bytes really read but we
	;; have to add 2 bytes and  decrement BC
	;; to point to the PID} header
	;;
	DEC	BC
	DEC	BC
	INC	L
	INC	L
	MMPS	krn.WRI

	;; Restore BC for next message to be read
	;;
	INC	BC
	INC	BC

	;; Back to forever loop waiting the next
	;; incoming message
	POP	DE
	JR	C,forever

outerr:
	;; Error. Return error code from A into
	;; E and set D to 0.
	LD	D,00
	STA	E
	RET

	.TEXT
	;; Name of the server queue: qBDGRD
qName:
	"qBDGR\00"

	.END

