	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	DEBUG - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; DEBUG - A process debugger for MMPS
	;; This process has in charge to launch under
	;; debugger (DBG) the process named in argument.
	;; The remaining arguments are passed to the
	;; debugged son.

	;; It displays the traced process CPU registers
	;; and when a key is pressed, trace the next
	;; instruction and updates the CPU registers
	;; display.

	;; DEBUG creates and activates its own console
	;; for its usage, so it keeps unchanged the
	;; original IN and OUT descriptors set by the
	;; father process for the debugged process.

	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

DEBUG:
	.LOCAL

	;; MMPS executable header.
	;;
	.BYTE
	BLDHEADER	DEBUG.asm$$._length 01 0040 

	.CODE
	;; The process enters here
	;;   BC contains the global address
	;;   DE the in and out descriptors
	;;   HL points to the arguments
	;;
	PUSH	DE

	;; BC points to the seed for our console
	;; and we copy the name to the top of our
	;; stack.
	LD	BC,PC
	LDA	*_*debugCON
	ADD	BC
	LD	DE,BC
	LD	BC,SP
	LDA	C
	AND	C0
	STA	C
	PUSH	HL
	PUSH	BC
	LD	L,03
copyseed:
	LDI	(DE)
	STI	(BC)
	DJC	copyseed
	POP	BC
	POP	HL

	;; Create a TMP name from seed DEBp.xxxx
	MMPS	krn.TMP

	;; Open a console for DEBUG. Like that
	;; the in and out descriptor coming
	;; from our father will be passed to
	;; the debugged process. Nice, no ?
	MMPS	krn.CON

	STA	D
	JR	NC,popouterr

	;; Activate  our console
	MMPS	krn.SCN

	;; We expect to have the name of the
	;; process to debug in the argument page.
	;; Fetch the next word, and this will
	;; start the argument for the debugged
	;; process we launch. We restore also
	;; the original in and out descriptor
	;; receive from our father
	CALL	NEXTARG
	LDA	E
	POP	DE
	PUSH	A
	MMPS	krn.DBG

	STA	D
	POP	A
	;; Error on debug. Save the error code
	;; to A and retrieve the out descriptor
	JR	NC,outerr

	STA	E
	AND	(BC),00
	LDI	(BC)

forever:
	LDA	H
	PUSH	A
	;; Get the CPU register of the debugged
	;; process
	MMPS	krn.REG

	PUSH	DE
	LD	HL,BC
	LDA	20
	ADD	BC
	LD	DE,BC
	PUSH	DE

	;; Here call the FMTSRING from the run
	;; time with in
	;;    BC the format string
	;;    DE the formatted string
	;;    HL the structure to format
	LD	BC,PC
	LDA	*_*debugREGstr
	ADD	BC
	CALL	FMTSTRING

	;; The formatted string is 26 characters:
	;;    SSSS HHLLDDEEBBCCAA PCPCFF
	LD	L,#26
	POP	BC
	POP	DE

	;; Refresh the console. This will avoid
	;; to be susped if a message is pending.
	MMPS	krn.RFR

	;; Write the formatted string of the CPU
	;; registers to the out descriptor
	MMPS	krn.WRI

	LDA	C
	AND	C0
	STA	C

	;; Wait for one key pressed
	LD	L,01
	MMPS	krn.REA
	POP	A

	;; Trace the debugged process. We will
	;; suspend until the process has executed
	;; its next instruction
	STA	H
	MMPS	krn.TRA
	JR	C,forever

	;; Error when tracing. We exit...
out:
	;; Return to father with &0000
	CLA
	STA	D
	STA	E
	RET

popouterr:
	POP	BC
outerr:
	;; Error code to E and 0 to D
	LDA	D
	STA	E
	LD	D,00
	RET

	.TEXT
debugREGstr:
	".W .W.W.W.B .W.B\00"

	.TEXT
	;; Seed for the DEBUGger console name
debugCON:
	"DEB\00"

	.END

