	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	DISPLAY - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; DISPLAY - A display file utility for MMPS
	;; DISPLAY will dump in hexadecimal the file
	;; or the entry named in argument.

	;; It reads the file 4 bytes by 4 bytes, and
	;; prints the file name, the offset and the
	;; 4 bytes in hexadecimal form.

	;; When the console output is refreshed by
	;; pressing the key RCL, the next 4 bytes
	;; are displayed

	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

DISPLAY:
	.LOCAL

	;; MMPS executable header.
	;;
	.BYTE
	BLDHEADER	DISPLAY.asm$$._length 01 0040

	.CODE
	;; The process enters here
	;;   BC contains the global address
	;;   DE the in and out descriptors
	;;   HL points to the arguments
	;;
	LDA	E
	PUSH	A

	;; Get file name to display from the
	;; argument page and open this file
	;; in read-only mode
	LD	BC,HL
	LD	D,04
	MMPS	krn.OPN
	JR	C,fileopen
outerr:
	;; Set &00 to D and MMPS error code
	;; into E and return to father
	LD	D,00
	STA	E
	POP	A
	RET

fileopen:
	POP	A
	STA	D

	;; HL points to argument page. We use
	;; this dummy page allocated to pass
	;; the arguments. Because we do not
	;; expect some, it is our buffer !
	;; This is not nice, but it reduces
	;; the space needed.
	PUSH	HL

	;; Skip the name. The buffer will start
	;; after, so the name, the offset and
	;; 4 bytes will be formatted.
	LD	C,08
skipname:
	LDI	(HL)
	JR	Z,formatsp
	DEC	C
	JR	C,skipname
formatsp:
	LDA	\x20 	;; Space character
	DEC	HL
loopsp:
	STI	(HL)
	DEC	C
	JR	C,loopsp

	POP	BC
forever:

	;; Main loop
	LD	HL,BC
	LDA	1A
	ADD	BC
	PUSH	HL

	;; Get into HL the current offset in the
	;; file pointed by the descriptor E
	MMPS	krn.TEL

	LDA	H
	STI	(BC)
	LDA	L
	STI	(BC)

	;; Read 4 bytes from descriptor E
	LD	L,04
	MMPS	krn.REA
	LDA	L
	POP	HL

	;; Error or no byte read (L=0), so we
	;; expect that is the end-of-file
	JR	NC,out
	JR	Z,out

	PUSH	DE
	PUSH	HL
	PUSH	A
	LDA	09
	ADD	HL
	PUSH	HL
	POP	DE
	LD	HL,BC

	LD	BC,PC
	LDA	*_*displaySTR
	ADD	BC
	DEC	HL
	DEC	HL
	;; Here call the FMTSRING from the run
	;; time with in
	;;    BC the format string
	;;    DE the formatted string
	;;    HL the structure to format
	CALL	FMTSTRING

	POP	A
	POP	BC
	STA	H
	SL
	RCF
	ADC	H
	ADC	0E
	STA	L
	POP	A
	PUSH	A
	STA	E
	;; Write the formatted string to the out
	;; descriptor
	MMPS	krn.WRI

	POP	DE
	JR	forever

out:
	LDA	D
	STA	E
	;; Flush the out descriptor. This is nice
	;; if we output to a console, because we
	;; will suspend until RCL is pressed. It
	;; gives time for user to read the output.
	MMPS	krn.FLU

	;; Return to father with &0000
	CLA
	STA	D
	STA	E
	RET
	SBC	C

	.TEXT
displaySTR:
	":.W .B .B .B .B\00"

	.END

