	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	FILES - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; FILES - A file-system node entries list
	;; FILES fetches all the file-system entries
	;; nodes (FSR), gets the status of each (FEN)
	;; and display the node type, the access
	;; rights, the number of pages allocated, the
	;; real entry ;; length and the entry name.

	;; When the console output is refreshed by
	;; pressing the key RCL, the next entry is
	;; displayed

	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

FILES:
	.LOCAL

	;; MMPS executable header.
	;;
	.BYTE
	BLDHEADER	FILES.asm$$._length 01 0040

	.CODE
	;; The process enters here
	;;   BC contains the global address
	;;   DE the in and out descriptors
	;;   HL points to the arguments
	;;
	;; FSR need to have &FF in D at first
	;; call
	LD	D,FF

forever:
	;; HL points to argument page. We use
	;; this dummy page allocated to pass
	;; the arguments. Because we do not
	;; expect some, it is our buffer !
	;; This is not nice, but it reduces
	;; the space needed.
	LD	BC,HL

	;; Read the file system entries in
	;; argument page + 5
	LDA	05
	ADD	BC
	MMPS	krn.FSR
	JR	NC,out_FS_END

	PUSH	HL
	LDA	0B
	ADD	BC
	PUSH	DE
	LD	DE,BC
	PUSH	DE

	LD	BC,PC
	LDA	*_*filesSTR
	ADD	BC
	;; Here call the FMTSRING from the run
	;; time with in
	;;    BC the format string
	;;    DE the formatted string
	;;    HL the structure to format
	CALL	FMTSTRING

	POP	BC
	POP	DE

	;; Restore all registers and write the
	;; formatted string to the out descriptor
	LD	L,1A
	MMPS	krn.WRI

	POP	HL
	JR	forever

out_FS_END:
	;; Flush the out descriptor. This is nice
	;; if we output to a console, because we
	;; will suspend until RCL is pressed. It
	;; gives time for user to read the output.
	MMPS	krn.FLU

	;; Return to father with &0000
	CLA
	STA	D
	STA	E
	RET
	SBC	C

	.TEXT
filesSTR:
		".N.-.A .B .B .W .S        \00"

	.END

