	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	MAKER - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; MAKER - Build a executable file from the
	;; BASIC area. The foreign assembled executable
	;; should be loaded into the BASIC area with
	;; a CLOAD
	;; -**- DO NOT TRY TO EDIT THIS PROGRAM -**-

	;; MAKER will create a file with the length
	;; retrieved from the BASIC area, copy the
	;; code and build the MMPS header.
	;; The name of the executable is passed in
	;; argument to MAKER

	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

MAKER:
	.LOCAL

	;; MMPS executable header.
	;
	.BYTE
	BLDHEADER	MAKER.asm$$._length 01 0040

	;;
	.CODE
	;; The process enters here
	;;   BC = global address
	;;   D in and E out descriptor
	;;   HL = argument address
	PUSH	DE

	;; First we create our lock to avoid another
	;; MAKER process to run on the same executable
	LD	BC,PC
	LDA	*_*makerLOK
	ADD	BC
	MMPS	krn.LOK
	POP	DE
	JR	NC,outerrnodel

	PUSH	DE
	;; Compute the size of the executable to install.
	;; This is ENDBASE - STARTBASIC.
	LDS	(>STARTBASIC)
	LD	DE,BC
	LDS	(>ENDBASIC)
	SCF
	LDA	C
	SBC	E
	STA	C
	LDA	B
	SBC	D
	STA	B
	PUSH	BC

	;; Under MMPS, the space for a file should be
	;; at the creation. The space to allocate is in
	;; number of pages (a page = 64 bytes). So we
	;; first add 63 bytes and divide the sum by 64
	LDA	3F
	ADD	BC
	LDA	C
	AEX
	AND	0C
	SR
	SR
	STA	C
	LDA	B
	AND	3F
	SL
	SL
	RCF
	ADC	C
	JR	NZ,l3

	;; The number of page is 0. The BASIC is empty.
	POP	MN
	JR	popout

l3:
	;; Copy the name of the executable to create
	;; for the argument line and create the file.
	;; We use the access RWX (07)
	LD	BC,HL
	LD	E,07
	STA	L
	MMPS	krn.FIL
	JR	NC,pop2outerr

	;; For the copy, MAKER access directly the
	;; BASIC area. Nice, we set the source address
	;; and require 64 bytes (or less) to write
	;; into the file
	LDS	(>STARTBASIC)
copyfrom:
	POP	HL
	PUSH	BC
	PUSH	DE
	LDA	H
	JR	NZ,l1
	;; If H = 0 and L < 64, so this is the last
	;; page, and we copy just the required amount
	;; of bytes
	LDA	L
	CPA	40
	JR	NC,l2
l1:
	;; We substract 64 bytes to the size
	SCF
	LDA	L
	SBC	40
	STA	L
	LDA	H
	SBC	N
	STA	H
	LDA	40
l2:
	POP	DE
	POP	BC
	PUSH	HL
	STA	L

	;; Load the amount of byte to be written in L
	;; Duplicate it in H to check if all is copied
	STA	H
	MMPS	krn.WRI
	JR	NC,pop2outerr

	;; Check if all bytes where  written
	LDA	L
	CPA	H
	JR	NZ,pop2outeof

	;; Next page to be copied
	ADD	BC

	;; It is finished ?	
	CPA	40
	JR	C,copyfrom

	POP	MN

	;; Now we prepare to call the build of the
	;; executable header (HDR). The user code
	;; contains the number of pages to allocate
	;; when launching the program (XEQ) and the
	;; offset of the global address. This is the
	;; amount of memory reserved for the stack
	;; and the variables. In most case, this
	;; will be 0 and 0. Because HDR will add
	;; 1 pages for the stack and an offset of
	;; &0040 to reserve the stack space.
	LDS	(>STARTBASIC)
	LDA	05
	ADD	BC
	LDD	(BC)
	STA	L
	LDD	(BC)
	STA	H
	LDA	(BC)
	STA	D
	MMPS	krn.HDR
	JR	NC,pop1outerr

	;; Done. Close the file and exit with return
	;; code &0000
	MMPS	krn.CLO
popout:
	POP	DE
	CLA
	STA	D
	STA	E
	RET

pop2outeof:
	LDA	err.EOF
	;; Error when calling the kernel-jumps.
pop2outerr:
	POP	MN
pop1outerr:
	POP	MN

	;; If the error code specify that the file
	;; already exist, we skip the file deletion.
	CPA	err.EIS
	JR	Z,outerrnodel

	;; Push the error code, because we want to
	;; return it to our father.
	PUSH	A

	;; The file has to be closed before to try
	;; to delete it.
	MMPS	krn.CLO

	;; Retrieve our global address. The name is
	;; the argument page.
	MMPS	krn.GBL

	DEC	BC
	LDD	(BC)
	STA	L
	LDA	(BC)
	STA	B
	LDA	L
	STA	C

	;; The file is deleted.
	MMPS	krn.DEL

	POP	A

outerrnodel:
	;; Error. Return error code from A into
	;; E and set D to 0.
	LD	D,00
	STA	E
	RET

	.TEXT
makerLOK:
	;; Name of lock to prevent a new instance
	;; of MAKER to run.
	"MKR.lock"

	.END

