	;; MMPS.inc - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>
	;; MACRO and symbols to develop MMPS executable with the lhTools assembler
	;;

	;; All symbols exported.
	.EXPORTALL

	;; defines all kernel-jump for the MMPS kernel
	;; these are related to the version 9/04/1998
	;;
krn.BAL	.EQU	&00
krn.BCH	.EQU	&01
krn.BFR	.EQU	&02
krn.BRE	.EQU	&03
krn.FSR	.EQU	&04
krn.FEN	.EQU	&05
krn.TMP	.EQU	&06
krn.LOK	.EQU	&07
krn.CON	.EQU	&08
krn.FIL	.EQU	&09
krn.QUE	.EQU	&0A
krn.DPX	.EQU	&0B
krn.SHR	.EQU	&0C
krn.MAP	.EQU	&0D
krn.DEL	.EQU	&0E
krn.ACS	.EQU	&0F
krn.RNM	.EQU	&10
krn.SIZ	.EQU	&11
krn.DSN	.EQU	&12
krn.DES	.EQU	&13
krn.OPN	.EQU	&14
krn.CLO	.EQU	&15
krn.REA	.EQU	&16
krn.WRI	.EQU	&17
krn.TEL	.EQU	&18
krn.SEK	.EQU	&19
krn.WAI	.EQU	&1A
krn.CNT	.EQU	&1B
krn.GET	.EQU	&1C
krn.PUT	.EQU	&1D
krn.GIV	.EQU	&1E
krn.TAK	.EQU	&1F
krn.SCN	.EQU	&20
krn.FLU	.EQU	&21
krn.RFR	.EQU	&22
krn.EDI	.EQU	&23
krn.HDR	.EQU	&24
krn.GBL	.EQU	&25
krn.PID	.EQU	&26
krn.FID	.EQU	&27
krn.XEQ	.EQU	&28
krn.END	.EQU	&29
krn.PRI	.EQU	&2A
krn.STA	.EQU	&2B
krn.PAU	.EQU	&2C
krn.EVT	.EQU	&2D
krn.KIL	.EQU	&2E
krn.MSK	.EQU	&2F
krn.POL	.EQU	&30
krn.RAI	.EQU	&31
krn.TIM	.EQU	&32
krn.ABT	.EQU	&33
krn.SEN	.EQU	&34
krn.SNR	.EQU	&35
krn.REC	.EQU	&36
krn.DBG	.EQU	&37
krn.TRA	.EQU	&38
krn.REG	.EQU	&39
krn.COD	.EQU	&3A
krn.DAT	.EQU	&3B
krn.RCH	.EQU	&3C
krn.DWN	.EQU	&3D


	;; MMPS <kernel-jump-code>
	;; useful macro to call the MMPS kernel
	.MACRO:	MMPS
		LDA	>__#0	;; kernel jump code to A
		SBR	(&E3)	;; go MMPS kernel
	.ENDMACRO

	;; fatal hook. Save registers as binary in the; BASIC variable E$ (&7650)
	;; and goes to the RESET vector (&FE)
	.MACRO:	FATAL
		SBR	(&93)	;; FATAL error hook
	.ENDMACRO


	;; defines all error-codes for the MMPS kernel
	;; these are related to the version 9/04/1998
	;;
err.KRN	.EQU	&A0	;; No such kernel-jump
	;;
err.MEM	.EQU	&B0	;; No block free
err.OWN	.EQU	&B1	;; Not owner
err.BAD	.EQU	&B2	;; Bad address
err.REL	.EQU	&B3	;; Can not realloc
err.MAP	.EQU	&B4	;; Can not map
	;;
err.PRO	.EQU	&C0	;; No process free
err.NOP	.EQU	&C1	;; No such process
err.DSX	.EQU	&C2	;; Not a valid DS for XEQ
err.SGN	.EQU	&C3	;; Not a valid signature
err.EVT	.EQU	&C4	;; Event already installed
err.NOV	.EQU	&C5	;; No such event
err.TIM	.EQU	&C6	;; Timer already running
err.RTR	.EQU	&C7	;; Not READY-to-RECEIVE
	;;
err.DES	.EQU	&D0	;; No descriptor free
err.XOP	.EQU	&D1	;; Exclusively open
err.NOD	.EQU	&D2	;; No such descriptor
err.EOF	.EQU	&D3	;; End of file
err.QEM	.EQU	&D4	;; Queue empty
err.QFL	.EQU	&D5	;; Queue full
err.CNT	.EQU	&D6	;; Can not connect
err.SHR	.EQU	&D7	;; Out of shared
	;;
err.TRA	.EQU	&E0	;; Not traced
err.DAT	.EQU	&E1	;; Data outside process space
	;;
err.ENT	.EQU	&F0	;; No entry free
err.NOE	.EQU	&F1	;; No such entry
err.EIS	.EQU	&F2	;; Entry already exists
err.BSY	.EQU	&F3	;; Entry busy
err.DEN	.EQU	&F4	;; Access denied
err.SIZ	.EQU	&F5	;; Can not re-size


	;; Build the MMPS executable header
	;; under MMPS, like on all OS, an executable should provide a header starting
	;; with a signature and information for the OS launcher.
	;; byte 0   := &F3 - The signature
	;; byte 1-2 := Real size of the program. XEQ checks if this size is equal to
	;;             the field byte 5-6 in the entry node structure
	;; byte 3   := number of pages to allocate for the stack and the global area
	;; byte 4-5 := offset of the global area in this block
	;;
	;; Note: The macro should belong in .BYTE fragment
	.MACRO:	BLDHEADER
		>krn.SGN
		<__#0 >__#0
		__#1
		<__#2 >__#2
	.ENDMACRO

krn.SGN	.EQU	&F3


	;; Block manager page bits
	;;
bm.MAP	.EQU	&01	;; Block is mapped
bm.FRE	.EQU	&FF	;; Block is free
	;;
bm.PSZ	.EQU	&40	;; Page size (64 bytes)

	;; File-system entry and descriptor type
	;;
fs.!LOK	.EQU	&F0	;; Not LOK
fs.LOK	.EQU	&00	;; Not LOK
fs.CON	.EQU	&10	;; Console
fs.FIL	.EQU	&20	;; File
fs.QUE	.EQU	&40	;; Queue
fs.SHR	.EQU	&80	;; Shared
fs.DPX	.EQU	&01	;; Duplex: queue | bit extension
fs.MAP	.EQU	&01	;; Map : shard | bit extension
fs.FRE	.EQU	&00	;; Node entry is free
fs.PID	.EQU	&AA	;; FS pid for owner

	;; File-system access right and descriptor mode
	;;
fs.XEQ	.EQU	&01	;; Executable
fs.WRI	.EQU	&02	;; Writable
fs.REA	.EQU	&04	;; Readable
fs.XOP	.EQU	&08	;; Exclusively open
	;;
ds.APN	.EQU	&01	;; Append mode
ds.WRI	.EQU	&02	;; Write mode
ds.REA	.EQU	&04	;; Read mode
ds.IS?	.EQU	&08	;; Descriptor is valid
ds.MAP	.EQU	&01	;; Map
ds.DPX	.EQU	&01	;; Duplex
ds.FRE	.EQU	&00	;; descriptor is free

	;; Process status-bits
	;;
ps.KRN	.EQU	&01	;; Enters kernel-jump
ps.TRA	.EQU	&02	;; Trace pending
ps.IN	.EQU	&04	;; CONSOLE input pending
ps.OUT	.EQU	&08	;; CONSOLE output pending
ps.LOK	.EQU	&10	;; LOCK waiting for GIV
ps.QUE	.EQU	&20	;; QUEUE/DUPLEX waiting
ps.PAU	.EQU	&40	;; Paused
ps.RTR	.EQU	&80	;; READY-TO-RECEIVE
ps.DBG	.EQU	&01	;; Debugger pending
ps.TIM	.EQU	&02	;; Timer pending
ps.RDY	.EQU	&F2	;; !(ps.RTR|ps.PAU|ps.QUE|ps.LOK|ps.DBG)
	;;
ps.FRE	.EQU	&FF	;; Process is free
ps.PID	.EQU	&A0	;; Kernel pid for father
ps.!PID	.EQU	&F0	;; Kernel pid has no father: orphan
	;; Events from kernel
evt.KIL	.EQU	&07	;; Break caught on console
evt.TIM	.EQU	&06	;; Timer elapsed
evt.END	.EQU	&05	;; Process ended
evt.PID	.EQU	&04	;; Son process return from execution
evt.US3	.EQU	&03	;; Not used - User event 3
evt.US2	.EQU	&02	;; Not used - User event 2
evt.US1	.EQU	&01	;; Not used - User event 1
evt.US0	.EQU	&00	;; Not used - User event 0

	;; Routines from the Run-Time (RTLIB.asm) library.
	;;

	;; highTOHEX - &1819
	;; lowTOHEX - &181A
	;;
	;; Convert the high or low 4-bits digit of A into a hexadecimal character 0123456789ABCDEF
	;; In:
	;;	A contains a 8-bits value
	;; Out:
	;;	A contains a hexadecimal character
highTOHEX	.EQU	&1819
lowTOHEX	.EQU	&181A

	;;NEXTARG - &1830
	;;
	;; Fetch the space in a string, nulls the space and return a pointer to the first argument.
	;; In:
	;;	HL contains address of the string to parse
	;; Out:
	;;	BC contains the origin address (HL in)
	;;	HL contains the address of the next character after a space.
	;;	In he string pointed by BC, the leading space are replaced by a &00.
NEXTARG		.EQU	&1830

	;;FMTSTRING - &1850
	;;
	;; Format an output buffer according to a format string and arguments.
	;; In:
	;;	BC contains the address of format string ended by &00
	;;	DE contains the address of the output buffer
	;;	HL contains the address of a binary buffer to be formatted
	;; Out:
	;;	BC, DE and HL points to the end of their respective buffers
	;;	The string is filled in the buffer pointed by DE. The string ends by a &00.
	;;
	;; The format string pointed by BC is on the form:
	;;	.<format char>.<fc>.<fc>
	;; where <format char> is
	;;	. :=	put a .
	;;	- :=	skip next byte without putting in buffer (equ. to INC HL)
	;;	H :=	put the high digit (bits7-4) in hexa
	;;	L :=	put the low digit (bits3-0) in hexa
	;;	B :=	put the 8-bits byte in hexa
	;;	W :=	put the 16-bits word in hexa
	;;	b :=	put the 8-bits byte in binary
	;;	h :=	put the high digit (bits7-4) in binary
	;;	l :=	put the low digit (bits3-0) in binary
	;;	C :=	put the character
	;;	S :=	put the string until &00 is found
	;;	N :=	put the node entry type, ie, one of 'LCFQS'
	;;	A :=	put the access mode, ie, one of 'ORWMCE'. Cleared bits are in lower case
	;;	P :=	put the process bits, ie, one of 'DOILQPR' added with a leading 'K' while
	;;		the process has enters a kernel-jump
	;;	any other character is copied directly to the formatted buffer.
	;;
	;; After each format, the HL register is incremented.
FMTSTRING	.EQU	&1850

