	.IF	INCLUDED?
	.ELSE
	;; Origin is immutable. SHOULD NEVER CHANGE
	;;
	.ORIGIN: 1800
	.ENDIF

	.CODE
	.COMMENT:	Run-time libraries - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; The RUN TIME library for MMPS executables.
	;; This library provide some usefull routines
	;; and help for writing executables.
	;; The .EXPORT entry points are immutable and
	;; the addresses should not be changed, else
	;; all executables will be impacted.
	;;
	;; MMPS kernel code is free from any call to
	;; these routines.
	;;
	;; In the future releases, entry points will
	;; be filled as JR +nn in the &1800..&1850
	;; area, and the library routines code will
	;; move after.
	;;
	.INCLUDE:	MMPS.inc

	.LOCAL

	.FILLTO: highTOHEX WITH 00

	;; highTOHEX - &1819
	;; lowTOHEX - &181A
	;;
	;; Convert the high or low 4-bits digit of A into a hexadecimal character 0123456789ABCDEF
	;; In:
	;;	A contains a 8-bits value
	;; Out:
	;;	A contains a hexadecimal character
	;;
	.IF	EXIST? highTOHEX
	.ELSE
.EXPORT: highTOHEX:
	.ENDIF

	.ORIGIN:	highTOHEX
	.CODE

	.IF	EQUAL? highTOHEX .
	.ELSE
	.FATAL	"highTOHEX differs"
	.ENDIF

		AEX

	.IF	EXIST? lowTOHEX
	.ELSE
.EXPORT: lowTOHEX:
	.ENDIF

	.IF	EQUAL? lowTOHEX .
	.ELSE
	.FATAL	"lowTOHEX differs"
	.ENDIF

	AND	0F
	OR	$0
	CPA	3A
	JR	NC,tmp_1_824
	ADC	06
tmp_1_824:
	RET
tmp_1_825:

fmt.HL:
	LDA	(HL)
	CALL	highTOHEX
	STI	(DE)
fmt.L:
	LDI	(HL)
	CALL	lowTOHEX
	STI	(DE)
	RET


	;;NEXTARG - &1830
	;;
	;; Fetch the space in a string, nulls the space and return a pointer to the first argument.
	;; In:
	;;	HL contains address of the string to parse
	;; Out:
	;;	BC contains the origin address (HL in)
	;;	HL contains the address of the next character after a space.
	;;	In he string pointed by BC, the leading space are replaced by a &00.
	;;
	.IF	EXIST? NEXTARG
	.ELSE
.EXPORT: NEXTARG:
	.ENDIF

	.IF	EQUAL? NEXTARG .
	.ELSE
	.FATAL	"NEXTARG differs"
	.ENDIF

	LD	BC,HL
tmp_1_832:
	LDI	(HL)
	INC	BC
	JR	Z,tmp_1_84b
	CPA	20	; space
	JR	Z,tmp_1_832
tmp_1_83a:
	LDI	(HL)
	JR	Z,tmp_1_84b
	CPA	20	; space
	JR	NZ,tmp_1_83a
	DEC	HL
	CLA
	STI	(HL)
tmp_1_844:
	LDI	(HL)
	JR	Z,tmp_1_84b
	CPA	20	; space
	JR	Z,tmp_1_844
tmp_1_84b:
	DEC	HL
	DEC	BC
	LDA	(BC)
	RET


tmp_1_84f:
	STI	(DE)

	;;FMTSTRING - &1850
	;;
	;; Format an output buffer according to a format string and arguments.
	;; In:
	;;	BC contains the address of format string ended by &00
	;;	DE contains the address of the output buffer
	;;	HL contains the address of a binary buffer to be formatted
	;; Out:
	;;	BC, DE and HL points to the end of their respective buffers
	;;	The string is filled in the buffer pointed by DE. The string ends by a &00.
	;;
	;; The format string pointed by BC is on the form:
	;;	.<format char>.<fc>.<fc>
	;; where <format char> is
	;;	. :=	put a .
	;;	- :=	skip next byte without putting in buffer (equ. to INC HL)
	;;	H :=	put the high digit (bits7-4) in hexa
	;;	L :=	put the low digit (bits3-0) in hexa
	;;	B :=	put the 8-bits byte in hexa
	;;	W :=	put the 16-bits word in hexa
	;;	b :=	put the 8-bits byte in binary
	;;	h :=	put the high digit (bits7-4) in binary
	;;	l :=	put the low digit (bits3-0) in binary
	;;	C :=	put the character
	;;	S :=	put the string until &00 is found
	;;	N :=	put the node entry type, ie, one of 'LCFQS'
	;;	A :=	put the access mode, ie, one of 'ORWMCE'. Cleared bits are in lower case
	;;	P :=	put the process bits, ie, one of 'DOILQPR' added with a leading 'K' while
	;;		the process has enters a kernel-jump
	;;	any other character is copied directly to the formatted buffer.
	;;
	;; After each format, the HL register is incremented.
	.IF	EXIST? FMTSTRING
	.ELSE
.EXPORT: FMTSTRING:
	.ENDIF

	.IF	EQUAL? FMTSTRING .
	.ELSE
	.FATAL	"FMTSTRING differs"
	.ENDIF

tmp_1_850:
	LDI	(BC)
tmp_1_851:
	JR	NZ,tmp_1_855
	STA	(DE)
	RET
tmp_1_855:
	CPA	$.
	JR	NZ,tmp_1_84f
	LDI	(BC)
	CPA	$.
	JR	Z,tmp_1_84f
	CPA	$-
	JR	NZ,tmp_1_865
	DEC	HL
	JR	tmp_1_850
tmp_1_865:
	CPA	$+
	JR	NZ,tmp_1_86c
	INC	HL
	JR	tmp_1_850
tmp_1_86c:
	CPA	$H
	JR	NZ,tmp_1_874
	LDI	(HL)
	AEX
	JR	tmp_1_879
tmp_1_874:
	CPA	$L
	JR	NZ,tmp_1_87e
	LDI	(HL)
tmp_1_879:
	CALL	lowTOHEX
	JR	tmp_1_84f
tmp_1_87e:
	CPA	$B
	JR	Z,tmp_1_889
	CPA	$W
	JR	NZ,tmp_1_88e
	CALL	tmp_1_825
tmp_1_889:
	CALL	tmp_1_825
	JR	tmp_1_850
tmp_1_88e:
	CPA	$C
	JR	NZ,tmp_1_895
	LDI	(HL)
	JR	tmp_1_84f
tmp_1_895:
	CPA	$S
	JR	NZ,tmp_1_89f
tmp_1_899:
	LDI	(HL)
	JR	Z,tmp_1_850
	STI	(DE)
	JR	tmp_1_899
tmp_1_89f:
	CPA	$b
	JR	NZ,tmp_1_8b6
	LDI	(HL)
	PUSH	HL
	LD	L,07
tmp_1_8a8:
	STA	H
tmp_1_8a9:
	LDA	H
	SL
	STA	H
	CLA
	ADC	$0
	STI	(DE)
	DJC	tmp_1_8a9
	POP	HL
	JR	tmp_1_850
tmp_1_8b6:
	CPA	$h
	JR	NZ,tmp_1_8c1
	LDI	(HL)
tmp_1_8bb:
	PUSH	HL
	LD	L,03
	JR	tmp_1_8a8
tmp_1_8c1:
	CPA	$l
	JR	NZ,tmp_1_8c9
	LDI	(HL)
	AEX
	JR	tmp_1_8bb
tmp_1_8c9:
	CPA	$N
	JR	NZ,tmp_1_8e7
	LDI	(HL)
	PUSH	BC
	LD	C,00
tmp_1_8d2:
	AND	F0
	JR	Z,tmp_1_8da
	SR
	INC	C
	JR	tmp_1_8d2
tmp_1_8da:
	RCF
	LDA	*_*nodeTYPstr[nodePChere]
	ADC	C
	LD	BC,PC
nodePChere:
	ADD	BC
	LDA	(BC)
	POP	BC
	JR	tmp_1_84f
tmp_1_8e7:
	CPA	$A
	JR	NZ,tmp_1_91c
	LDI	(HL)
	AEX
	PUSH	BC
	PUSH	HL
	STA	H
	LD	BC,PC
	LDA	*_*accessTYPstr
	ADD	BC
	LD	L,02
tmp_1_8fa:
	LDA	H
	SL
	STA	H
	LDI	(BC)
	JR	C,tmp_1_902
	OR	20
tmp_1_902:
	STI	(DE)
	DJC	tmp_1_8fa
	LDA	H
	CPA	C0
	JR	Z,tmp_1_915
	INC	BC
	CPA	A0
	JR	Z,tmp_1_915
	INC	BC
	CPA	90
	JR	Z,tmp_1_915
	INC	BC
tmp_1_915:
	LDI
	POP	HL
	POP	BC
	JR	tmp_1_850
tmp_1_91c:
	CPA	$P
	JR	NZ,tmp_1_84f
	LDA	$K
	BIT	(HL),01
	JR	NZ,tmp_1_928
	LDA	$-
tmp_1_928:
	STI	(DE)
	LDI	(HL)
	PUSH	BC
	LD	C,00
tmp_1_92e:
	AND	FE
	JR	Z,tmp_1_936
	SR
	INC	C
	JR	tmp_1_92e
tmp_1_936:
	RCF
	LDA	*_*procSTATstr[procstatPChere]
	ADC	C
	LD	BC,PC
procstatPChere:
	ADD	BC
	LDA	(BC)
	POP	BC
	JR	tmp_1_84f
	CPA	$*
	JR	NZ,tmp_1_84f
	LDI	(HL)
	JR	tmp_1_851

	.TEXT
nodeTYPstr:
	"LCFQS\00"

accessTYPstr:
	"ORWMCE-\00"

procSTATstr:
	"-DOILQPR\00"

	.CODE

	.END
