	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	RUNPRO - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; RUNPRO. A funny executable blinking the
	;; RUN or the PRO flags on the LCD.

	;; The program first tries to create a lock
	;; (RUN.lock),;If it is successful, the RUN
	;; flag is set or clear. Else, it tries to
	;; create a second lock (PRO.lock); If it
	;; is successful, the PRO flag is set or
	;; clear; Else the executable returns with
	;; the error code &F2.

	;; The clear/set are delayed, and so, the
	;; PRO flag is managed 2 times faster than
	;; the RUN flag.

	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

RUNPRO:
	.LOCAL

	.BYTE
	BLDHEADER	RUNPRO.asm$$._length 01 0000

	.CODE
	;; RUN bit in LOWLCDFLAG
RUNbit	.EQU	&40
	;; PRO bit in LOWLCDFLAG
PRObit	.EQU	&20

	;; The process enters here
	;;   BC contains the global address
	;;   DE the in and out descriptors
	;;   HL points to the arguments
	PUSH	DE
	LD	H,RUNbit
	LD	BC,PC
	LDA	*_*offRUN
trylock:
	ADD	BC
	;; try to create a lock (LOK) named by BC
	MMPS	krn.LOK
	STA	E
	JR	C,loop
	;; error! So test if PRObit. If yes exit with error code
	LDA	H
	CPA	PRObit
	JR	Z,outerr
	;; shift right and try again with next
	RCF
	SR
	STA	H
	LDA	0A
	JR	trylock
loop:
	LDA	H
	SL
	;; set the delay value in DE (PRO x 2 faster the RUN)
	STA	E
	LD	D,00
	LD	B,<LOWLCDFLAG
	LD	C,>LOWLCDFLAG
forever:
	LDA	H
	XOR	FF
	STA	H
	;; critical section - sure to be proof of interrupt
	DI
	CP	H,80
	JR	C,clearbit
	;; set the flag bit
	LDA	H
	OR	(BC)
	JR	loadbit
clearbit:
	;; clear the flag bit
	LDA	(BC)
	STA	L
	LDA	H
	STA	(BC)
	LDA	L
	AND	(BC)
loadbit:
	STA	(BC)
	;; critical section - restore the interrupts
	EI
	;; delay the process for the value DE
	MMPS	krn.PAU
	JR	C,forever
	STA	E
outerr:
	LD	D,00
out:
	POP	BC
	;; return to father with code in DE
	RET

	.TEXT
offRUN:
	;; lock name for RUN flag
	"RUN.lock\00\00"
	;; lock name for PRO flag
	"PRO.lock\00\00"

	.END

