	.IF	INCLUDED?
	;; Align on next 64-bytes frontier
	.ALIGN:	40
	.ELSE
	;; Origin does not matter, because code
	;; and data are fully relocatable.
	.ORIGIN: 40C5
	.ENDIF

	.CODE
	.COMMENT:	STARTER - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; STARTER - The first executable launched
	;; by the MMPS kernel boot.

	;; It creates a event queue to receive the
	;; events 4, 5 and 7 from the kernel. It
	;; polls the events queue and display the
	;; catched both event 4 (normal son process
	;; termination with the return code) and
	;; event 5 (ended son process).

	;; It reads from the console, assumes that
	;; an executable name is given followed by
	;; some arguments, opens a new console and
	;; activates it, and tries to lauch the
	;; process on this console.
	;; If this fails, it displays the error
	;; code.

	;; Note that the RCL key should be pressed
	;; to refresh the console output.

	;; If just ENTER is read, it just polls
	;; the events queue.

	;; If End-Of-File (UP-Arrow) is received
	;; and the STARTER PID is 0, a kernel
	;; shutdown (DWN) is performed.
	
	;; MMPS kernel-jump and macros
	;;
	.INCLUDE:	MMPS.inc

STARTER:
	.LOCAL

	.BYTE

	BLDHEADER	STARTER.asm$$._length 01 0040

	.CODE

	;; program enters here
	;;   BC contains the global address
	;;   DE the in and out descriptors
	;;   HL points to the arguments
	PUSH	DE
	LD	BC,PC
	LDA	*_*starterCON
	ADD	BC
	LD	DE,BC
	LD	BC,HL

	;; set event mask to 7, 5 and 4
	;; and create the event queue
	LD	L,B0
	MMPS	krn.EVT
	LD	L,02
copyseed:
	LDI	(DE)
	STI	(BC)
	DJC	copyseed

	AND	(BC),00
	LDA	0D
	ADD	BC
	POP	DE

checkevent:
	PUSH	DE

checkeventloop:
	;; check for any events in queue
	MMPS	krn.POL
	JR	NC,noevent

	;; event is present: due to mask, we expect
	;;   7 := BREAK key  -> do nothing
	;;   5 := son killed -> display '>p -5'
	;;   4 := son exited -> display '>p +4:ddee'
	LDA	*_*evt4STR[evtPChere]
	CP	L,04
	JR	Z,doevent
	CP	L,05
	JR	NZ,checkeventloop
	LDA	*_*evt5STR[evtPChere]

doevent:
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD	DE,BC
	LD	BC,SP
	INC	BC
	LD	HL,BC
	LD	BC,PC
evtPChere:
	ADD	BC
	;; Here call the FMTSRING from the run
	;; time with in
	;;    BC the format string
	;;    DE the formatted string
	;;    HL the structure to format
	CALL	FMTSTRING

	SCF
	LDA	E
	POP	BC
	POP	BC
	POP	BC
	SBC	C
	STA	L
	POP	DE

	;; write events to out descriptor E
	MMPS	krn.WRI
	JR	checkevent

noevent:
	POP	DE

	;; read 27 characters from in descriptor E
	;; we read more than 26 characters to force
	;; the user to press the ENTER key
	LD	L,#27
	MMPS	krn.REA
	;; error from REA? perhaps EOF
	JR	NC,checkforout

	;; nothing read
	CP	L,00
	JR	Z,checkevent
	PUSH	DE
	PUSH	BC
	LDA	C
	SCF
	SBC	10
	STA	C
	;; create a TMP name from seed CONp.xxxx
	MMPS	krn.TMP

	;; open a new console
	MMPS	krn.CON
	POP	HL
	JR	NC,confailed

	CALL	NEXTARG
	JR	Z,closecon

	;; select the console as active
	MMPS	krn.SCN
	LDA	E
	STA	D
	;; launch the son !
	;;   BC the name of the son
	;;   D and E the descriptor to the console
	;;   HL the remaining assumed as arguments
	MMPS	krn.XEQ

	;; oops ! launch failed
	JR	NC,closecon

	;; Display >P with P the PID of the son
	LDA	3E
	STI	(BC)
	LDA	H
	OR	30
	STD	(BC)
	POP	DE

	;; refresh the console. If a WRI is pending
	;; process will suspend
	MMPS	krn.RFR
	;;

	;; display >p with p = son pid's
	LD	L,02
	MMPS	krn.WRI
	JR	checkevent

closecon:
	STA	H
	;; Mmmm... launch failed. Close the console
	MMPS	krn.CLO

showerr:
	;; Show !ee
	POP	DE

	;; restore our console
	MMPS	krn.SCN
	LDA	21
	STI	(BC)
	LDA	H
	CALL	highTOHEX
	STI	(BC)
	LDA	H
	CALL	lowTOHEX
	STD	(BC)
	DEC	BC

	;; write !ee with ee the error code from XEQ
	LD	L,03
	MMPS	krn.WRI
	JR	checkevent

confailed:
	LD	BC,HL
	STA	H
	JR	showerr

checkforout:
	;; error code is D3 ? End-of-file
	CPA	err.EOF
	JR	NZ,fatal

	;; who am i ? is my PID 0 ?
	MMPS	krn.PID
	CP	H,00
	JR	NZ,out

	;; Yes ! User ask me to DWN the kernel
	MMPS	krn.DWN
	;; DWN should never return
	;; here we enter into the FATAL hook of MMPS
fatal:
	SBR	(93)

out:
	;; STARTER exits normally with &0000
	CLA
	STA	D
	STA	E
	RET

	SBC	C
	SBC	C
	SBC	C
	SBC	C

	.TEXT
evt4STR:
	;; string for '>p +v:eeee'
	">.L +.L:.W\00"

evt5STR:
	;; string for '>p -v'
	">.L -.L\00"

starterCON:
	;; seed for console name
	"CON\00"

	.END

