#!/bin/sh

LHASM=`which lhasm 2>/dev/null`
[ -z "$LHASM" ] && \
	echo "lhTools are not installed. Skip lhasm pass..." && LHASM=true
${LHASM} -v
echo

KERNEL=mmps-090498
FS_LIB=mmps-fs+lib
VOLATILE=mmps-volatile

for k in stddev 1500Adev stdfull 1500Afull
do
	i=Images/${k}
	[ ! -d ${i} ] && mkdir -p ${i}
	case $k in
		stddev)
			n="Standard development"
			a=4200
			b=4800
			c=4100
			d=4130
			e=4140
			f=4200 
			g=1F
			h=40
			;;
		1500Adev)
			n="1500A development"
			a=4000
			b=5000
			c=7F00
			d=7F80
			e=7E01
			f=7EFD 
			g=29
			h=7D
			;;
		stdfull)
			n="full"
			a=2000
			b=3FC0
			c=4200
			d=42FE
			e=4101
			f=41FD 
			g=29
			h=40
			;;
		1500Afull)
			n="1500A full"
			a=2000
			b=3FC0
			c=7F00
			d=7FFE
			e=7E01
			f=7EFD 
			g=29
			h=7D
			;;
	esac

	a=baseHEAP=$a
	b=topHEAP=$b
	c=tableHEAP=$c
	d=maxIndexHEAP=$d
	e=tableDS=$e
	f=maxIndexDS=$f 
	g=maxDS=$g
	h=basePROC=$h

	echo "Building $n : ${KERNEL}-${k}"
	$LHASM -E -A $a -A $b -A $c -A $d -A $e -A $f -A $g -A $h \
		-L ${i}/${KERNEL}-${k}.lst \
		-S ${i}/${KERNEL}-${k}.sym \
		-F ${i}/${KERNEL}-${k}.frag \
		-o ${i}/${KERNEL}-${k}.bin ${KERNEL}.asm
	[ $? -ne 0 ] && exit $?
	lhdump -C=16FD -F ${i}/${KERNEL}-${k}.frag -o ${i}/${KERNEL}-${k}.txt ${i}/${KERNEL}-${k}.bin

	bin2wav -p 1500 -s 5 -n "MMPS 9/04/98" -a 0x00C5 -t bin \
		${i}/${KERNEL}-${k}.bin ${i}/${KERNEL}-${k}.wav
	[ $? -ne 0 ] && exit $?
	echo

	echo "Building MMPS filesystem and run time library"
	$LHASM -E -A $a -A $b -A $c -A $d -A $e -A $f -A $g -A $h \
		-L ${i}/${FS_LIB}-${k}.lst \
		-S ${i}/${FS_LIB}-${k}.sym \
		-F ${i}/${FS_LIB}-${k}.frag \
		-o ${i}/${FS_LIB}-${k}.bin ${FS_LIB}.asm
	[ $? -ne 0 ] && exit $?
	o=`grep '\\ORIGIN:' ${FS_LIB}.asm | head -n 1 | \
		sed 's/^.*ORIGIN:[ 	]*//;s/[ 	]*$//'`
	bin2wav -p 1500 -s 5 -n "MMPS FS+LIB" -a 0x${o} -t bin \
		${i}/${FS_LIB}-${k}.bin ${i}/${FS_LIB}-${k}.wav
	[ $? -ne 0 ] && exit $?
	echo

	echo "Building MMPS volatile data..."
	o=`echo $h | sed 's/^.*=//;'`
	$LHASM -E -A $a -A $b -A $c -A $d -A $e -A $f -A $g -A $h \
		-L ${i}/${VOLATILE}-${k}.lst \
		-S ${i}/${VOLATILE}-${k}.sym \
		-F ${i}/${VOLATILE}-${k}.frag \
		-O ${o}00 -b -o ${i}/${VOLATILE}-${k}.bin ${VOLATILE}.hex
	[ $? -ne 0 ] && exit $?
	bin2wav -p 1500 -s 5 -n "MMPS VOLATILE" -a 0x${o}00 -t bin \
		${i}/${VOLATILE}-${k}.bin ${i}/${VOLATILE}-${k}.wav
	[ $? -ne 0 ] && exit $?
	echo

done

i=Images/Examples
[ ! -d ${i} ] && mkdir -p ${i}
for f in RUNPRO BDGRD SHOWKEY
do
	echo "Building ${f} executable"
	$LHASM -L ${i}/${f}.lst -E -o ${i}/${f}.bin ${f}.asm
	[ $? -ne 0 ] && exit $?
	bin2wav -p 1500 -s 5 -n "${f}" -t img ${i}/${f}.bin ${i}/${f}.wav
	[ $? -ne 0 ] && exit $?
	echo
done

exit 0
