;;=======================================================================
;;	DISCLAIMER
;;
;;	You use this program, executables and libraries at your own risks.
;;	I am mot responsible for any damage hardware or software resulting
;;	from the use of this program and from the programs, executables or
;;	libraries developed with this software.
;;=======================================================================


	;;
	;; MMPS - Mini Multi Process System
	;; A small operation system for the SHARP PC-1500/A and TANDY PC-2
	;;
	;; Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>
	;;

	;; 
	.ORIGIN: 00C5

	;; Include MMPS macros and defines
	.INCLUDE:	MMPS.macro.inc
	.INCLUDE:	MMPS.inc
	.INCLUDE:	MMPS.volatile.inc

	.CODE
	.COMMENT: 	MMPS kernel - Copyright 1994-2014 Christophe Gottheimer <cgh75015@gmail.com>

	;; Cold start point - Entry routine from BASIC or other
	JP	mmps.START

mmps.krn.conTASK:
	;; Console TASK routine
	JP	krn.conTASK

mmps.krn.jumpMAX:
	;; Code of the last kernel-jump. This is krn.DWN
	LDA	1+krn.DWN

;;***********************************************************************
;;	This address is immutable because it is hardcoded into the ROM
;;
mmps.krn.jump:
	;; Kernel-jump entry point. SBR	(&E3) jumps here
	JP	enterkrnjump
	;; If not at address &00CD, raise a FATAL error
	.IF	EQUAL?	mmps.krn.jump	&00CD
	.ELSE
	.FATAL	"mmps.krn.jump address differs from &00CD"
	.ENDIF
;;
;;***********************************************************************

	.TEXT
mmps.banner:
	;; MMPS banner with date. We happy to present ourself
	"MMPS 09Apr.1998\00"

	.CODE
mmps.krn.vect00_1F:
	;; Kernel-jumps vector: 00-1F
	JP	krn.vect00_1F.base
mmps.krn.vect20_3D:
	;; Kernel-jumps vector: 20-3D
	JP	krn.vect20_3D.base

mmps.krn.IT.scheduler:
	;; Interrupt-scheduler vector
	JP	krn.IT.scheduler

mmps.krn.IT.debugger:
	;; Interrupt-debugger vector
	JP	krn.IT.debugger

;;***********************************************************************
;;	This address is immutable because it is hardcoded into the ROM
;;
mmps.krn.FATAL:
	;; FATAL-error entry point. SBR (&93) enters here
	JP	mmps.FATAL
	;; If not at address &00EC, raise a FATAL error
	.IF	EQUAL?	mmps.krn.FATAL	&00EC
	.ELSE
	.FATAL	"mmps.krn.jump address differs from &00EC"
	.ENDIF
;;
;;***********************************************************************

mmps.krn.endproc:
	;; Process-ending vector
	JP	krn.endproc

mmps.krn.sgn:
	;; MMPS executable signature
	LDA	krn.SGN

	;; When booting, MMPS first tries to open a console
	;; named 'CONSOLE', select it as active, prints the
	;; banner, and launches the fisrt process STARTER.

	.TEXT
mmps.CONSOLE:
	;; The name of the first console
	"CONSOLE\00"

	.TEXT
mmps.STARTER:
	;; The name of the first process
	"STARTER\00"

	.TEXT
mmps.CLOSED:
	;; Printed when an active console is closed
	"Closed!\00"

	.CODE



;;=======================================================================
;;	BLOCK MANAGER : bm
;;
;;	It is the lowest level. It is totally independent from the others,
;;	but called by them each time an allocation or release is needed.
;;
;;	It is in charge of the pages allocation and release. Under MMPS,
;;	each memory unity is called 'PAGE'. A page has a size of 64 bytes
;;	and is always aligned on a 64-bytes frontier. It can manage up to
;;	128 pages. A set of 1 to N contiguous pages is called 'BLOCK'.
;;
;;	When a allocation is done, the free block may be split into 2 blocks:
;;	one allocated and one kept free. On a same way, when release is done
;;	the freed block will be concatenated with the previous if it is free,
;;	with the following if is free. So only one big free block may result
;;	of the coalescence.
;;
;;	The file-system and the user processes ask for blocks. The blocks are
;;	owned by the calling process and keep its ownership until it releases
;;	them or it is ended.
;;
;;	In the kernel, the 2 higher levels ask for blocks allocation:
;;	  + The file-system to build the entries and to reserve the space
;;	    needed by them;
;;	  + The process-scheduler to creates the stacks, the arguments pages
;;	    and the messages transmission.
;;
;;	The kernel is working with physical addresses, and these addresses
;;	will be returned to the caller. No virtual memory access is yet
;;	possible (except a similar mechanism with the SHARED entries).
;;
;;	A process has to take in charge its own memory management and to
;;	avoid to access outside of its space. The kernel has no security
;;	against arocess which will go out from its space, neither against a
;;	process which want to access to some forbidden or private areas.
;;	If a process tries to access to a non valid address, the kernel will
;;	not be warned. But accessing to the data outside the process space may
;;	cause some irrecoverable crashes. If a process wants to access to an
;;	external areas, especially while writing, it has to be sure that no
;;	other processes may access to this area.
;;
;;	The pages allocated to the file-system are kept until it asks to
;;	release them. The processes stacks, messages and blocks are also in
;;	the same physical memory. After a lot entry creations, deletions,
;;	pages allocations, messages sent and processes started, it may occur
;;	that the memory map is very fragmented, and that no big blocks
;;	(composed by a big set of contiguous pages) are possible to allocate.
;;	In this case, nothing to done except to write a code to defragment the
;;	memory map. There is also no algorithm to allocate the pages, except
;;	that the closest block (in memory pages number) is fetched. When a
;;	block preceding and following two free blocks is released, the three
;;	blocks will become one block containing the sum of the three blocks
;;	pages. In this version of MMPS, no memory usage information is
;;	available.
;;
;;	The PAGE-ALLOCATION-TABLE structure
;;	byte 0:	internal PID of the owner
;;		&00 ::= pid0
;;		&10 ::= pid1
;;		&20 ::= pid2...
;;		&70 ::= pid7
;;		&AA ::= KERNEL pid
;;		&FF ::= free
;;	byte 1: number of pages or 0 for a page inside a block
;;

;;-----------------------------------------------------------------------
;;	BAL - Block allocate
;;
;;	Call:
;;		LD	L,number-of-pages
;;		LDA	00
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Allocate the number of pages stored into L, and return the
;;		physical address of the first byte into BC. The calling
;;		process becomes owner of the allocated pages which are added
;;		to its space. All its pages will be released when the process
;;		will be terminated. A page is a group of 64 bytes always aligned
;;		on 64 bytes frontier. A block is a group of N contiguous pages.
;;
;;	Return:
;;		CARRY if success; BC contains the address of the allocated block.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	No enough page free
;;
mmps.bm.bal:
	LDA	(currentPROC)
krn.bm.allocBLOCK_pid_A:
	STA	D
krn.bm.allocBLOCK_pid_D:
	LDR	BC,tableHEAP
	LDA	L
	JR	NZ,tmp_0_118
	INC	L
tmp_0_118:
	LD	H,FF
tmp_0_11a:
	.IF	EQUAL?	<tableHEAP,<maxIndexHEAP
	CP	C,>maxIndexHEAP
	.ELSE
	CP	B,<maxIndexHEAP
	.ENDIF
	JR	C,tmp_0_135
	LDI	(BC)
	CPA	bm.FRE
	JR	Z,tmp_0_129
tmp_0_123:
	LDD	(BC)
	SL
	ADD	BC
	JR	tmp_0_11a
tmp_0_129:
	LDA	(BC)
	CPA	L
	JR	NC,tmp_0_123
	CPA	H
	JR	C,tmp_0_123
	STA	H
	LDA	C
	STA	E
	JR	tmp_0_123
tmp_0_135:
	CP	H,bm.FRE
	JR	NZ,tmp_0_13d
	LD	H,err.MEM
	RCF
	RET
tmp_0_13d:
	LDA	E
	STA	C
	DEC	BC
	LDA	D
	STI	(BC)
	LDA	(BC)
	SCF
	SBC	L
	STA	H
	LDA	L
	STI	(BC)
	DEC	L
	JR	tmp_0_14f
tmp_0_14b:
	LDA	D
	STI	(BC)
	CLA
	STI	(BC)
tmp_0_14f:
	DJC	tmp_0_14b
	LDA	H
	JR	Z,tmp_0_156
	INC	BC
	STI	(BC)
tmp_0_156:
	LDA	E
	SR
	SR
	SR
	RCF
	ADC	<baseHEAP
	STA	B
	LDA	E
	AND	&06
	SL
	AEX
	STA	C
	SCF
	RET

;;-----------------------------------------------------------------------
;;	BCH - Block validity check
;;
;;	Call:
;;		LD	BC,block-address
;;		LDA	01
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Check if block at the address stored in BC is valid. The
;;		block should be owned by the calling process.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	No owner of this resource
;;
mmps.bm.bch:
	LDA	B
	CPA	<topHEAP
	JR	C,tmp_0_195
	SCF
	SBC	<baseHEAP
	JR	NC,tmp_0_195
	SL
	SL
	SL
	STA	E
	LDA	C
	AND	C0
	SR
	AEX
	RCF
	ADC	E
	STA	C
	LD	B,<tableHEAP
	INC	BC
tmp_0_17f:
	LDA	(BC)
	JR	NZ,tmp_0_18a
	DEC	BC
	CP	C,00
	JR	Z,tmp_0_18a
	DEC	BC
	JR	tmp_0_17f
tmp_0_18a:
	DEC	BC
	LDA	C
	STA	E
	LDI	(BC)
	CPA	D
	JR	Z,tmp_0_194
tmp_0_191:
	LD	H,err.OWN
	RCF
tmp_0_194:
	RET
tmp_0_195:
	LD	H,err.BAD
	RCF
	RET

;;-----------------------------------------------------------------------
;;	BFR - Block free
;;
;;	Call:
;;		LD	BC,block-address
;;		LDA	
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Release a block allocated at the address stored in BC. The
;;		block should be owned by the calling process. The complete
;;		block is released.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	No owner of this resource
;;
mmps.bm.bfr:
	LDA	(currentPROC)
	STA	D
tmp_krn.bm.freeBLOCK_pid_D:
	CALL	mmps.bm.bch
	JR	NC,tmp_0_1d3
krn.bm.freeBLOCK:
	LDD	(BC)
	STA	H
	DEC	A
	STA	L
	CLA
tmp_0_1a7:
	DEC	A
	STI	(BC)
	INC	A
	STI	(BC)
	DJC	tmp_0_1a7
	.IF	EQUAL?	<tableHEAP,<maxIndexHEAP
	CP	C,>maxIndexHEAP
	.ELSE
	CP	B,<maxIndexHEAP
	.ENDIF
	JR	C,tmp_0_1bc
	LDI	(BC)
	CPA	FF
	JR	NZ,tmp_0_1bc
	LDA	(BC)
	RCF
	ADC	H
	STA	H
	AND	(BC),00
tmp_0_1bc:
	LDA	E
	STA	C
tmp_0_1be:
	LDA	(BC)
	CPA	FF
	JR	NZ,tmp_0_1cb
	CP	C,00
	JR	Z,tmp_0_1cd
	DEC	BC
	DEC	BC
	JR	tmp_0_1be
tmp_0_1cb:
	INC	BC
	INC	BC
tmp_0_1cd:
	INC	BC
	LDA	(BC)
	RCF
	ADC	H
	STA	(BC)
	SCF
tmp_0_1d3:
	RET

;;-----------------------------------------------------------------------
;;	BRE - Block reallocate
;;
;;	Call:
;;		LD	BC,block-address
;;		LD	L,number-of-pages
;;		LDA	03
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Reallocate (change the number of pages of) the allocated
;;		block pointed by BC with the new size given in L (the new
;;		size should be less or equal to the old one) and release
;;		the reminder.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		B2	New size is greater than the old one
;;
mmps.bm.bre:
	LDA	(currentPROC)
	STA	D
tmp_0_1d8:
	CALL	mmps.bm.bch
	JR	NC,tmp_0_1f2
	LDA	(BC)
	SBC	L
	JR	Z,tmp_0_1f2
	JR	NC,tmp_0_1f0
	STA	H
	LDA	L
	STA	(BC)
	SL
	ADD	BC
	LDA	H
	STA	(BC)
	LDA	C
	DEC	A
	STA	E
	JR	krn.bm.freeBLOCK
tmp_0_1f0:
	LD	H,err.REL
tmp_0_1f2:
	RET



;;=======================================================================
;;	FILE-SYSTEM : fs
;;
;;	It is the second level. It is in charge of the entries management
;;	and the data storage. All the entries in the file system are named.
;;	The name is a set from 1 to 9 characters. The entries have also a
;;	type; 7 types are recognized by the file-system: the 'FILE', the
;;	'CONSOLE', the 'QUEUE', the 'SHARED' and the 'LOCK'.
;;
;;	  + The FILE is a block which contains code, text or data. It is
;;	    used to store the executable code of the utilities, the sources,
;;	    the texts or simply some data.
;;	  + The CONSOLE is a logical link to the physical DISPLAY and
;;	    KEYBOARD of the computer. Several CONSOLE may be opened at the
;;	    same time, but only one is active and linked to the
;;	    DISPLAY/KEYBOARD device.
;;	  + The QUEUE is a channel between several processes; the process
;;	    which created it is owner and has the READ access. The others
;;	    have only the WRITE access. It is used for asynchronous
;;	    inter-processes communication.
;;	  + The DUPLEX is a connected channel to a peer queue. Writing to
;;	    a DUPLEX sends the message to the peer DUPLEX (read). It is an
;;	    implementation of the PIPE. A DUPLEX is only possible between
;;	    2 processes.
;;	  + The SHARED is space in the memory which may be acceded by
;;	    several processes, to shared some data. Only the action of READ
;;	    and WRITE are possible on the SHARED entries. The shared are
;;	    equivalent to a memory space with a virtual address access.
;;	  + The MAP is the export of a memory region as a file entry. The
;;	    access right to the MAP are set when creating it.
;;	  + The LOCK is just a specific entry and may be used by a process
;;	    to lock a critical resource.
;;
;;	The file-system allows also access rights control. Like the system
;;	does not support the real-ownership, these rights are just used to
;;	avoid some operations on the concerned entry. The following rights
;;	are supported: READ, WRITE and EXECUTE. An EXCLUSIVE open option is
;;	also possible to avoid to share an entry with another process. The
;;	CONSOLE and LOCK are always opened exclusively. 
;;	A link counter is also provided to avoid some operations on some
;;	case of entry, like deleting an opened entry.
;;	For the I/O actions, a process needs to get a descriptor on an entry.
;;	It is done by 'OPEN'ing the entry. When it finishes its work, the
;;	descriptor has to be 'CLOSE'd. A descriptor is allocated by the
;;	file-system which stores on it some information like the open-mode,
;;	the entry type, the owner, the node-link and some other information
;;	depending of the entry-type. The descriptors owned by a process are
;;	closed at the process termination. Opening an entry will increment
;;	its link number and closing will decrement it. To open an entry, the
;;	calling process has to give the 'OPEN MODE', which will define the
;;	actions possible. Some actions are specific to the entry type and
;;	some other are common. This is very easy to write a transparent code,
;;	independent of the descriptor type. However, the LOCK and SHARED are
;;	always managed in a separate way.
;;	An extra set of specific actions is also provides for the CONSOLE to
;;	manage the I/O.
;;
;;	The file-system can manage up to 32 descriptors (no limitation by
;;	type), and up to 64 entries (also without limitation by type).
;;	When a process wants to create an entry, the file-system will ask to
;;	the page-manager to reserve some pages for it. These pages will be
;;	released when the entry will be removed. Closing some entries like
;;	CONSOLE, LOCK, QUEUE or DUPLEX will remove them.
;;	The major constraint is that an entry has always a contiguous set of
;;	pages. So, the 'CREATE' action have to specify the number of pages to
;;	reserve for the new entry. The QUEUE, DUPLEX and the SHARED entries
;;	have a limit of 4 pages. The CONSOLE entries have always 1 page.
;;	The FILE entries may have from 1 to N pages; the number of pages will
;;	give the maximum size that the FILE may have; a real-size is given
;;	which indicates the real count of bytes stored. An action to re size
;;	the FILE entries is provided.
;;
;;	The DESCRIPTOR-TABLE structure
;;	byte 0: entry-type and open-mode (= &00 ::= free)
;;		bit7-4: entry-type
;;			0000 ::= LOCK
;;			0001 ::= CONSOLE
;;			0010 ::= FILE
;;			0100 ::= QUEUE or DUPLEX
;;			1000 ::= SHARED or MAP
;;		bit3: descriptor validity
;;		bit2-0: open-mode
;;			100 ::= READ open
;;			010 ::= WRITE open
;;			001 ::= APPEND open (FILE)
;;				Shared or Duplex bit (SHARED or DUPLEX)
;;	byte 1: descriptor owner internal-PID
;;		&00 ::= pid0
;;		&10 ::= pid1
;;		&20 ::= pid2...
;;		&70 ::= pid7
;;	byte 2,3: physical address of the entry
;;	byte 4,5: offset from the top (FILE)
;;		  connected peer DS (DUPLEX)
;;		  DS chained list (LOCK)
;;
;;
;;	The FILE-SYSTEM entry
;;	byte 0: entry-type and access-mode (= &00 ::= free)
;;		bit7-4:	entry-type
;;			0000 ::= LOCK
;;			0001 ::= CONSOLE
;;			0010 ::= FILE
;;			0100 ::= QUEUE or DUPLEX
;;			1000 ::= SHARED or MAP
;;		bit3:	exclusive open (bit set when the entry is open with
;;			the EXCLUSIVE flag; this bit is set for the CONSOLE
;;			 and LOCK).
;;		bit2-0: access-mode
;;			100 ::= READ allowed
;;			010 ::= WRITE allowed
;;			001 ::= EXECUTE allowed (FILE); if set, the KERNEL
;;				will permit the run it.)
;;				Duplex or Map bit (DUPLEX or MAP)
;;	byte 1: link counter (= &00 ::= no link)
;;		(to be deleted or exclusively open, an entry should have a
;;		link counter equal to &00.)
;;	byte 2,3: physical address of the entry-space
;;		  DS owner (LOCK)
;;	byte 4: number of pages reserved for the entry-space
;;	byte 5,6: real-size (FILE)
;;		  head of DS chained list (LOCK)
;;	byte 7-15: entry-name (terminated by &00 if the name is less than
;;		   9 characters.)
;;

krn.fs.findFIRSTFSENTRY:
	LD	L,>tableFS
tmp_0_1f5:
	LD	H,<tableFS
tmp_0_1f7:
	CP	L,>maxIndexFS
	JR	C,tmp_0_21b
	LDI	(HL)
	CPA	FF
	JR	NZ,tmp_0_203
tmp_0_200:
	INC	HL
	JR	tmp_0_1f7
tmp_0_203:
	STA	D
	LDA	(HL)
	STA	E
	LDA	L
	STA	H
tmp_0_208:
	LDA	(DE)
	JR	NZ,tmp_0_219
krn.fs.findNEXTFSENTRY:
	LDA	10
	ADD	DE
	LDA	E
	BIT	maskFS
	JR	NZ,tmp_0_208
	LDA	H
	STA	L
	INC	L
	JR	tmp_0_1f5
tmp_0_219:
	SCF
	RET
tmp_0_21b:
	LD	H,err.NOE
	RCF
	RET

	;; Find the entry pointed by BC in the
	;; whole file-system
krn.fs.findENTRY:
	CALL	krn.fs.findFIRSTFSENTRY
tmp_0_222:
	JR	NC,tmp_0_21b
	PUSH	DE
	LDA	07
	ADD	DE
	LD	L,08
	PUSH	BC
tmp_0_22e:
	LDI	(DE)
	CPI
	JR	NZ,tmp_0_23c
	CPA	N
	JR	Z,tmp_0_237
	DJC	tmp_0_22e
tmp_0_237:
	POP	BC
tmp_0_239:
	POP	DE
	RET
tmp_0_23c:
	POP	BC
	POP	DE
	CALL	krn.fs.findNEXTFSENTRY
	JR	tmp_0_222

krn.fs.findFSENTRY:
	LD	H,<tableFS
	CPA	FF
	JR	Z,krn.fs.findFIRSTFSENTRY
	STA	E
	AND	0F
	SL
	OR	>tableFS
	STA	L
	LDI	(HL)
	CPA	FF
	JR	Z,tmp_0_200
	STA	D
	LDA	E
	AND	maskFS
	OR	(HL)
	STA	E
	LDA	L
	STA	H
	JR	krn.fs.findNEXTFSENTRY

;;-----------------------------------------------------------------------
;;	FSR - File System entry read
;;
;;	Call:
;;		LD	BC,name
;;		LD	D,fsds (&FF at first call)
;;		LD	HL,entry-struct
;;		LDA	04
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Read the file system entry. At the beginning, the fsds D
;;		should be initialized with &FF. The filesystem entry
;;		information are stored into the entry structure pointed by
;;		HL, the name is stored in the name buffer pointed by BC.
;;		When no more entry are read, the error code F1 is returned.
;;		The entry structure is described in the FEN kernel jump.
;;		The output of FSR is the same as FEN.
;;
;;	Return:
;;		CARRY if success; D contains the updated fsds for a next call.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such node
;;
mmps.fs.fsr:
	PUSH	HL
	PUSH	BC
	LDA	D
	CALL	krn.fs.findFSENTRY
	POP	BC
	JR	NC,tmp_0_239
	LDA	L
	AND	1F
	SR
	RCF
	ADC	E
	STA	H
	PUSH	DE
	LDA	07
	ADD	DE
	LD	L,08
tmp_0_27c:
	LDI	(DE)
	STI	(BC)
	DJC	tmp_0_27c
	AND	(BC),00
	POP	DE
	POP	BC
	JR	tmp_0_291

;;-----------------------------------------------------------------------
;;	FEN - File system entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	HL,entry-struct
;;		LDA	05
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Read the node information of a named entry in BC and stores
;;		them into the entry structure pointed by HL. The entry
;;		structure contains the following information:
;;		byte 0, bits 7 to 4	Entry type
;;					0000	LOCK entry
;;					0001	CONSOLE entry
;;					0010	FILE entry
;;					0100	QUEUE or DUPLEX entry
;;					1000	SHARED or MAP entry
;;		byte 0, bits 3 to 0	Access rights
;;					1000	Exclusively opened
;;					0100	Read access
;;					0010	Write access
;;					0001	Executable access (FILE)
;;						Duplex or Map bit (DUPLEX or MAP)
;;		byte 1			Number of links
;;		byte 2			Number of pages reserved
;;		byte 3 & 4		Real length
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such entry
;;
mmps.fs.fen:
	PUSH	HL
	CALL	krn.fs.findENTRY
	POP	BC
	JR	NC,tmp_0_2a0
tmp_0_291:
	LDI	(DE)
	STI	(BC)
	LDI	(DE)
	STI	(BC)
	INC	DE
	INC	DE
	LDI	(DE)
	STI	(BC)
	LDI	(DE)
	STI	(BC)
	LDI	(DE)
	STI	(BC)
	LDA	H
	STA	D
	SCF
tmp_0_2a0:
	RET

tmp.byteTOASCII:
	STA	D
	AEX
	CALL	tmp.digitTOASCII
	LDA	D
tmp.digitTOASCII:
	AND	0F
	RCF
	ADC	$a
	STI	(BC)
	RET

;;-----------------------------------------------------------------------
;;	TMP - Create a unique temporary entry name
;;
;;	Call:
;;		LD	BC,generic-name
;;		LDA	06
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create a temporary unique name using the first 3 characters
;;		of the buffer pointed by BC and returns the full name in
;;		this buffer, terminated by a &00 byte.
;;
;;	Return:
;;		CARRY and BC is filled with the temporary name.
;;
;;	Error Code:
;;		None.
;;
mmps.fs.tmp:
	PUSH	BC
	LD	L,02
tmp_0_2b2:
	LDA	(BC)
	JR	Z,tmp_0_2b8
	INC	BC
	DJC	tmp_0_2b2
tmp_0_2b8:
	LDA	(currentPROC)
	AEX
	OR	$0
	STI	(BC)
	LDA	$.
	STI	(BC)
	LDR	HL,tmpSEED
	LDI	(HL)
	STA	D
	LDA	(HL)
	STA	E
	INC	DE
	LDA	E
	STD	(HL)
	LDA	D
	STA	(HL)
	CALL	tmp.byteTOASCII
	LDA	E
	CALL	tmp.byteTOASCII
	AND	(BC),00
	POP	BC
	PUSH	BC
	CALL	krn.fs.findENTRY
	POP	BC
	JR	C,mmps.fs.tmp
	SCF
	RET

;;-----------------------------------------------------------------------
;;	LOK - Create and open a LOCK entry
;;
;;	Call:
;;		LD	BC,name
;;		LDA	07
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a LOCK entry named by BC, open
;;		it and return the descriptor in E. When a LOCK descriptor is
;;		closed, the LOCK entry is deleted from the file system.
;;		A LOCK is opened with the EXCLUSIVE mode. No page are needed
;;		to create a LOCK entry.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F0	No entry free
;;		F2	Entry already exists
;;		D0	No DS free
;;
mmps.fs.lok:
	LD	D,fs.LOK
	LD	E,02
tmp_0_2e9:
	LD	L,00
	JR	tmp_0_318

;;-----------------------------------------------------------------------
;;	CON - Create and open a CONSOLE entry
;;
;;	Call:
;;		LD	BC,name
;;		LDA	08
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a CONSOLE entry named by BC,
;;		open it and return the descriptor in E. A CONSOLE is always
;;		opened with the modes READ, WRITE and EXCLUSIVE. When a
;;		CONSOLE is opened, it is reset. When a CONSOLE is closed,
;;		it is deleted from the file system. Creating a CONSOLE entry
;;		needs one page.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	Not enough page free
;;		F0	No entry free
;;		F2	Entry already exists
;;		D0	No DS free
;;
mmps.fs.con:
	LD	D,fs.CON
	LD	L,01
	JR	tmp_0_311

;;-----------------------------------------------------------------------
;;	FIL - Create and open a FILE entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	E,access
;;		LD	L,number-of-pages
;;		LDA	09
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a FILE entry named by BC, with
;;		the access rights given by E, reserve the number of pages
;;		given by L for it, open it and return the descriptor in E.
;;		If not enough page are available to create the FILE entry,
;;		an error is returned. The access rights are given by E; E is
;;		an OR of the following:
;;			0100 := READ access
;;			0010 := WRITE access
;;			0001 := EXECUTABLE access
;;		The real length of a FILE entry is set to 0, even the number
;;		of pages is reserved. The length will be set after each WRI
;;		to this FILE.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	Not enough page free
;;		F0	No entry free
;;		F2	Entry already exists
;;		D0	No DS free
;;
mmps.fs.fil:
	LD	D,fs.FIL
	LDA	E
	AND	07
	STA	E
	JR	tmp_0_313

;;-----------------------------------------------------------------------
;;	QUE - Create and open a QUEUE entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	L,number-of-pages
;;		LDA	0A
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a QUEUE entry name by BC,
;;		reserve the number of pages given in L for it, open it,
;;		and return the descriptor in E. Only the process which
;;		created the QUEUE entry has the READ access. When the QUEUE
;;		descriptor with the READ access is closed, the QUEUE entry
;;		is deleted from the file system and all the descriptors
;;		linked to it are invalidated. It is not possible to create a
;;		QUEUE with more than 4 pages. If there is not enough page
;;		available, the QUEUE entry is not created.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	Not enough page free
;;		F0	No entry free
;;		F2	Entry already exists
;;		D0	No DS free
;;
mmps.fs.que:
	LD	D,fs.QUE
	JR	tmp_0_301

;;-----------------------------------------------------------------------
;;	DPX - Create and open a DUPLEX entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	L,number-of-pages
;;		LDA	0B
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a DUPLEX entry name by BC,
;;		reserve the number of pages given in L for it, open it,
;;		and return the descriptor in E. Writing to a DUPLEX entry
;;		will give data available to the peer DUPLEX. In a same way,
;;		to be usable, the DUPLEX pair should be connected by a call
;;		to the CNT kernel-jump. It is not possible to create a
;;		DUPLEX with more than 4 pages. If there is not enough page
;;		available, the DUPLEX entry is not created.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	Not enough page free
;;		F0	No entry free
;;		F2	Entry already exists
;;		D0	No DS free
;;
mmps.fs.dpx:
	LD	D,41	; (fs.QUE|fs.DPX)
tmp_0_301:
	CP	L,05
	JR	NC,tmp_0_307
	LD	L,04
tmp_0_307:
	CP	D,41	; (fs.QUE|fs.DPX)
	JR	NZ,tmp_0_311
	LD	E,0F
	JR	tmp_0_313

;;-----------------------------------------------------------------------
;;	SHR - Create and open a SHARED entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	L,number-of-pages
;;		LDA	0C
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a SHARED entry name by BC,
;;		reserve the number of pages given in L for it, open it, and
;;		return the descriptor in E. The SHARED entry has no owner.
;;		When a SHARED descriptor is close, the SHARED entry will
;;		remind in the file system until the link number is 0; at
;;		this time, it will be deleted. It is not possible to create
;;		a SHARED entry with more than 4 pages. If not enough page are
;;		available, the SHARED entry is not created and an error is
;;		returned.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	Not enough page free
;;		F0	No entry free
;;		F2	Entry already exists
;;		D0	No DS free
;;
mmps.fs.shr:
	LD	D,fs.SHR
tmp_0_311:
	LD	E,06
tmp_0_313:
	CP	L,01
	JR	C,tmp_0_318
	INC	L
tmp_0_318:
	PUSH	BC
	PUSH	DE
	PUSH	HL
	CALL	krn.fs.findENTRY
	POP	HL
	JR	NC,tmp_0_329
	LD	H,err.EIS
	JR	tmp_0_37f
tmp_0_329:
	LDA	L
	JR	Z,tmp_0_33c
	PUSH	A
	LD	D,fs.PID
	CALL	krn.bm.allocBLOCK_pid_D
	JR	C,tmp_0_340
	POP	A
tmp_0_337:
	POP	DE
	POP	BC
	RET
tmp_0_33c:
	STA	B
	STA	C
	PUSH	A
tmp_0_340:
	PUSH	BC
	LDR	HL,tableFS
	LD	D,FF
tmp_0_348:
	CP	L,>maxIndexFS
	JR	Z,tmp_0_36d
	LDI	(HL)
	CPA	FF
	JR	NZ,tmp_0_35c
	INC	HL
	CP	D,FF
	JR	NZ,tmp_0_348
	LDA	L
	DEC	A
	DEC	A
	STA	D
	JR	tmp_0_348
tmp_0_35c:
	STA	B
	LDI	(HL)
	STA	C
tmp_0_35f:
	LDA	(BC)
	JR	Z,tmp_0_3a0
	LDA	10
	ADD	BC
	LDA	C
	BIT	maskFS
	JR	NZ,tmp_0_35f
	JR	tmp_0_348
tmp_0_36d:
	LDA	D
	CPA	FF
	JR	NZ,tmp_0_382
tmp_0_372:
	LD	D,fs.PID
	POP	BC
	POP	A
	JR	Z,tmp_0_37d
	CALL	tmp_krn.bm.freeBLOCK_pid_D
tmp_0_37d:
	LD	H,err.ENT
tmp_0_37f:
	RCF
	JR	tmp_0_337
tmp_0_382:
	STA	L
	PUSH	HL
	LD	L,01
	LD	D,fs.PID
	CALL	krn.bm.allocBLOCK_pid_D
	POP	HL
	JR	NC,tmp_0_372
	LDA	B
	STI	(HL)
	LDA	C
	STD	(HL)
	LD	DE,BC
	LDA	10
	LD	L,03
tmp_0_39a:
	AND	(DE),00
	ADD	DE
	DJC	tmp_0_39a
tmp_0_3a0:
	POP	HL
	POP	A
	POP	DE
	PUSH	A
	LDA	E
	STA	(BC)
	LDA	D
	OR	(BC)
	STI	(BC)
	CLA
	STI	(BC)
	LDA	H
	STI	(BC)
	LDA	L
	STI	(BC)
	POP	A
	STI	(BC)
	STA	L
	LDA	D
	STA	H
	AND	D0
	JR	Z,tmp_0_3c7
	LDA	L
	SR
	SR
	STI	(BC)
	LDA	L
	AEX
	SL
	SL
	JR	tmp_0_3c8
tmp_0_3c7:
	STI	(BC)
tmp_0_3c8:
	STI	(BC)
	POP	DE
tmp_0_3cb:
	LD	L,08
tmp_0_3cd:
	LDI	(DE)
	JR	Z,tmp_0_3d5
	STI	(BC)
	DJC	tmp_0_3cd
	JR	tmp_0_3d8
tmp_0_3d5:
	STI	(BC)
	DJC	tmp_0_3d5
tmp_0_3d8:
	LDA	H
	DEC	BC
	LDA	C
	AND	F0
	STA	E
	LDA	B
	STA	D
	LDA	(DE)
	CPA	4F
	JR	Z,tmp_0_3eb
	AND	06
	STA	H
tmp_0_3e8:
	JP	tmp_0_593
tmp_0_3eb:
	LD	H,0D
	JR	tmp_0_3e8

;;-----------------------------------------------------------------------
;;	MAP - Create and open a MAP entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	E,access-right
;;		LD	HL,block-address
;;		LDA	0D
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create, if not already done, a MAP entry name by BC, fixes
;;		the access rights given by E and set the base address to the
;;		block pointed by HL for it, open it, and return the descriptor
;;		in E. The access rights are given by E; E is an OR of the
;;		following:
;;			0100 := READ access
;;			0010 := WRITE access
;;		The block should belong in the ownership of the calling
;;		process. If the MAP is closed, all descriptors related to this
;;		MAP are invalidated. No page are needed to create a MAP entry.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	Not enough page free
;;		B1	Not owner
;;		F0	No entry free
;;		F2	Entry already exists
;;		F3	Can not map
;;		D0	No DS free
;;
mmps.fs.map:
	PUSH	DE
	PUSH	BC
	LDA	(currentPROC)
	STA	D
	LD	BC,HL
	CALL	mmps.bm.bch
	LD	HL,BC
	POP	BC
	POP	DE
	JR	NC,tmp_0_447
	PUSH	HL
	LD	D,fs.SHR
	LDA	E
	AND	06
	STA	E
	CALL	tmp_0_2e9
	JR	NC,tmp_0_448
	STA	C
	OR	(BC),ds.MAP
	INC	BC
	INC	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	STA	L
	OR	(HL),fs.MAP
	INC	HL
	INC	HL
	POP	BC
	LDA	C
	SR
	SR
	SR
	RCF
	ADC	<baseHEAP
	STI	(HL)
	LDA	C
	AND	06
	SL
	AEX
	STI	(HL)
	LDA	(BC)
	STI	(HL)
	SR
	SR
	STI	(HL)
	LDA	(BC)
	AEX
	SL
	SL
	AND	F0
	STA	(HL)
	LDD	(BC)
tmp_0_43b:
	ADD	(BC),bm.MAP
	INC	BC
	INC	BC
	DEC	A
	JR	NZ,tmp_0_43b
	SCF
	RET
	LD	H,err.MAP
	RCF
tmp_0_447:
	RET
tmp_0_448:
	POP	BC
	RET

;;-----------------------------------------------------------------------
;;	DEL - Delete an entry
;;
;;	Call:
;;		LD	BC,name
;;		LDA	0E
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Delete the entry named by BC from the file system. To be
;;		deleted, an entry should be free (no link) and the WRITE
;;		access should be allowed. The pages reserved are released,
;;		and the entry is deleted from the file system. It is not
;;		possible to delete an entry opened or linked to a running
;;		process.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such entry
;;		F3	Entry is busy
;;		F4	Access denied
;;		D1	Exclusively opened
;;
mmps.fs.del:
	CALL	krn.fs.findENTRY
	JR	NC,tmp_0_457
	INC	DE
	LDD	(DE)
	JR	Z,tmp_0_458
tmp_0_454:
	LD	H,err.BSY
	RCF
tmp_0_457:
	RET
tmp_0_458:
	LDA	(DE)
	BIT	02
;; >>> Patch cgh - Delete #0003 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;;	JR	NZ,tmp_0_461
;;	LD	H,F4
;;	RCF
;;	RET
	JR	NZ,tmp_0_460
tmp_0_45d:
	JP	tmp_0_8c9
tmp_0_460:
	STA	H
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
tmp_0_461:
	AND	(DE),00
	BIT	F0
	JR	Z,tmp_0_493
	INC	DE
	INC	DE
	LDI	(DE)
	STA	B
	LDI	(DE)
	STA	C
	PUSH	DE
	LDA	H
	BIT	80
	JR	Z,tmp_0_478
	BIT	01
	JR	NZ,tmp_0_47f
tmp_0_478:
	LD	D,fs.PID
	CALL	tmp_krn.bm.freeBLOCK_pid_D
	JR	tmp_0_491
tmp_0_47f:
	LDA	(currentPROC)
	INC	A
	STA	D
	CALL	mmps.bm.bch
	JR	NC,tmp_0_491
	LDD	(BC)
tmp_0_48a:
	ADD	(BC),FF
	INC	BC
	INC	BC
	DEC	A
	JR	NZ,tmp_0_48a
tmp_0_491:
	POP	DE
tmp_0_493:
	LDA	E
	AND	C0
	STA	E
	LD	L,03
	LDA	10
tmp_0_49b:
	BIT	(DE),FF
	JR	NZ,tmp_0_4c5
	ADD	DE
	DJC	tmp_0_49b
	DEC	DE
	LDA	E
	AND	C0
	STA	C
	LDA	D
	STA	B
	LDR	HL,tableFS
tmp_0_4ae:
	CP	L,>maxIndexFS
	SBR	C,(93)
	LDI	(HL)
	CPA	B
	JR	NZ,tmp_0_4ba
	LDA	(HL)
	CPA	C
	JR	Z,tmp_0_4bd
tmp_0_4ba:
	INC	HL
	JR	tmp_0_4ae
tmp_0_4bd:
	DEC	HL
	OR	(HL),FF
	LD	D,fs.PID
	CALL	tmp_krn.bm.freeBLOCK_pid_D
tmp_0_4c5:
	RET

;;-----------------------------------------------------------------------
;;	ACS - Change the access right of an entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	E,access
;;		LDA	0F
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Change the access right of the entry named by BC with the
;;		new access given by E; E is an OR of the following values:
;;			0100 := READ access
;;			0010 := WRITE access
;;			0001 := EXECUTE access (FILE)
;;		If this entry is opened, the descriptors will keep the old
;;		rights. The new rights will be taken at the next open (OPN).
;;		The access rights of the LOCK and the CONSOLES entry have
;;		no meaning, because the access rights are set while creating
;;		the entry, and these are exclusively opened.
;;		It is not possible to change the access rights if the entry
;;		is exclusively opened.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such entry
;;		D1	Exclusively opened
;;
mmps.fs.acs:
	PUSH	DE
	CALL	krn.fs.findENTRY
	POP	HL
	JR	NC,tmp_0_4d6
	AND	(DE),F8	; ~(fs.REA|fs.WRI|fs.XEQ)
	LDA	L
	AND	07	; (fs.REA|fs.WRI|fs.XEQ)
	OR	(DE)
	STA	(DE)
tmp_0_4d6:
	RET

;;-----------------------------------------------------------------------
;;	RNM - Rename an entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	DE,new-name
;;		LDA	10
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Rename the entry named by BC with the new name given by DE.
;;		The rename is dynamic, all the processes linked to it will
;;		take the new name, because the process table keeps only the
;;		address of the entry. It is not possible to rename an entry
;;		with an already existing name or if the entry is exclusively
;;		opened.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such entry
;;		F2	Entry already exists
;;		D1	Exclusively opened
;;
mmps.fs.rnm:
	PUSH	DE
	PUSH	BC
	LD	BC,DE
	CALL	krn.fs.findENTRY
	POP	BC
	JR	C,tmp_0_4f8
	CALL	krn.fs.findENTRY
	LD	BC,DE
	POP	DE
	JR	NC,tmp_0_4d6
	BIT	(BC),08
	JR	NZ,tmp_0_5c6
	LDA	07
	ADD	BC
	JP	tmp_0_3cb
tmp_0_4f8:
	LD	H,err.EIS
	POP	DE
	RCF
	RET

;;-----------------------------------------------------------------------
;;	SIZ - Re-size an FILE entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	L,number-of-pages
;;		LDA	11
;;		SBR	E3,0D
;;
;;	Action:
;;		Re-size the FILE entry named by BC with the new number of
;;		pages given in L. To be re-sized, the WRITE access should
;;		be allowed, the new size should be less than the old one,
;;		but greater than the real size and the number of link should
;;		be 0. It is not possible to re size while a descriptor to
;;		the FILE entry exists. It is not possible to re-size a FILE
;;		entry with 0 page.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such entry
;;		F3	Entry is busy
;;		F4	Access denied
;;		F5	Can not re-size
;;
mmps.fs.siz:
	PUSH	HL
	CALL	krn.fs.findENTRY
	POP	HL
	JR	NC,tmp_0_53e
	BIT	(DE),20
	JR	Z,tmp_0_53b
	LDI	(DE)
	AND	06
	CPA	06
tmp_0_510:
	JR	NZ,tmp_0_45d
	LDI	(DE)
	JR	NZ,tmp_0_454
	LDI	(DE)
	STA	B
	LDI	(DE)
	STA	C
	LDA	L
	CPA	(DE)
	JR	Z,tmp_0_53a
	JR	C,tmp_0_53b
	INC	DE
	SR
	SR
	CPA	(DE)
	INC	DE
	JR	NC,tmp_0_53b
	JR	NZ,tmp_0_531
	LDA	L
	AND	03
	AEX
	SL
	SL
	CPA	(DE)
	JR	NC,tmp_0_53b
tmp_0_531:
	DEC	DE
	DEC	DE
	LDA	L
	STA	(DE)
	LD	D,fs.PID
	CALL	tmp_0_1d8
tmp_0_53a:
	RET
tmp_0_53b:
	LD	H,err.SIZ
	RCF
tmp_0_53e:
	RET

krn.fs.allocDS:
	PUSH	A
	LDR	BC,tableDS
	LD	L,maxDS
	LDA	06
	SCF
tmp_0_54a:
	BIT	(BC),FF
	JR	Z,tmp_0_555
	ADD	BC
	DJC	tmp_0_54a
	LD	H,err.DES
	RCF
tmp_0_555:
	POP	A
	RET

;;-----------------------------------------------------------------------
;;	DSN - Get entry name from a opened descriptor
;;
;;	Call:
;;		LD	BC,name
;;		LDA	12
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the entry name of the descriptor E into BC.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.dsn:
	PUSH	BC
;; >>> Patch cgh - DSN #0001 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;;	CALL	063B
	CALL	krn.fs.getDS
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	POP	DE
	JR	NC,tmp_0_575
	INC	BC
	INC	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	ADC	06
	STA	L
	LD	C,09
tmp_0_56b:
	LDI	(HL)
	STI	(DE)
	JR	Z,tmp_0_574
	DEC	C
	JR	NZ,tmp_0_56b
	AND	(DE),00
tmp_0_574:
	SCF
tmp_0_575:
	RET

;;-----------------------------------------------------------------------
;;	DES - Get entry type from a opened descriptor
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	13
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the entry type (entry type and open mode) of the
;;		descriptor E into H. The following bits are used:
;;		bits7-4 Descriptor type
;;			0000 := LOCK entry
;;			0001 := CONSOLE entry
;;			0010 := FILE entry
;;			0100 := QUEUE or DUPLEX (bit0 = 1) entry
;;			1000 := SHARED or MAP (bit0 = 1) entry
;;		bits3-0 Open mode
;;			1000 := VALIDITY bit
;;			0100 := READ mode
;;			0010 := WRITE mode
;;			0001 := APPEND mode (FILE)
;;		       	        DUPLEX (bits7-4 = 0100)
;;		                MAP (bits7-4 = 1000)
;;		If the VALIDITY bit (1000) is not set, no operation are
;;		possible on this descriptor except to close it.
;;
;;	Return:
;;		CARRY if success; H contains the entry type and the open mode.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.des:
;; >>> Patch cgh - DES #0001 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;;	CALL	063B
	CALL	krn.fs.getDS
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	JR	NC,tmp_0_57c
	STA	H
tmp_0_57c:
	RET

;;-----------------------------------------------------------------------
;;	OPN - Open an entry
;;
;;	Call:
;;		LD	BC,name
;;		LD	D,open-mode
;;		LDA	14
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Open the entry named by BC with the open mode given by D
;;		(only for FILE, SHARED and MAP entries, the QUEUE entries
;;		are always opened with only the WRITE mode). To be opened,
;;		an entry should have the access rights asked by the open
;;		mode and should not be already exclusively opened. The open
;;		mode is an OR of the following values:
;;			1000 := EXCLUSIVE mode
;;			0100 := READ mode
;;			0010 := WRITE mode
;;			0001 := APPEND mode (FILE only)
;;		The descriptor is returned in E. Opening with the APPEND
;;		mode will sets the position at the end. Opening with only
;;		the WRITE mode will rewind the descriptor and resets the
;;		FILE. APPEND mode should be given with READ or WRITE mode.
;;		Opening with WRITE mode will set the EXCLUSIVE mode. It is
;;		not possible to open with the EXCLUSIVE mode if at least one
;;		link exists. Opening an entry will increase by 1 its number
;;		of links.
;;
;;	Return:
;;		CARRY if success; E contains the opened descriptor.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F1	No such entry
;;		F3	Entry is busy
;;		F4	Access denied
;;		D0	No DS free
;;		D1	Exclusively opened
;;
mmps.fs.opn:
	PUSH	DE
	CALL	krn.fs.findENTRY
	JR	C,tmp_0_587
	POP	DE
	RET
tmp_0_587:
	POP	HL
	BIT	(DE),fs.QUE
	JR	NZ,tmp_0_591
	BIT	(DE),fs.!LOK
	JR	NZ,tmp_0_593
tmp_0_591:
	LD	H,ds.WRI
tmp_0_593:
	BIT	(DE),fs.FIL
	JR	Z,tmp_0_59f
	LDA	H
	BIT	ds.WRI
	JR	Z,tmp_0_59e
	OR	fs.XOP
tmp_0_59e:
	STA	H
tmp_0_59f:
	BIT	(DE),fs.CON
	JR	Z,tmp_0_5a5
	LD	H,0E	; (fs.XOP|ds.REA|ds.WRI)
tmp_0_5a5:
	BIT	(DE),fs.!LOK
	JR	NZ,tmp_0_5ab
	LD	H,fs.WRI
tmp_0_5ab:
	LDA	H
	AND	0F	; (fs.XOP|ds.REA|ds.WRI|ds.APN)
	JR	NZ,tmp_0_5b2
	LDA	ds.REA
tmp_0_5b2:
	CALL	krn.fs.allocDS
	JR	NC,tmp_0_5c9
	STA	H
	BIT	fs.XOP
	JR	NZ,tmp_0_5c0
	BIT	(DE),fs.XOP
	JR	Z,tmp_0_5cc
tmp_0_5c0:
	INC	DE
	BIT	(DE),FF
	DEC	DE
	JR	Z,tmp_0_5ca
tmp_0_5c6:
	LD	H,err.XOP
	RCF
tmp_0_5c9:
	RET
tmp_0_5ca:
	OR	(DE),fs.XOP
tmp_0_5cc:
	AND	07	; (fs.REA|fs.WRI|fs.XEQ)
	STA	H
	AND	06	; (ds.REA|ds.WRI)
	STA	L
	AND	(DE)
	CPA	L
	JR	Z,tmp_0_5da
	AND	(DE),F7	; ~(fs.XOP)
	JR	tmp_0_510
tmp_0_5da:
	LDA	(DE)
	AND	F0
	RCF
	ADC	H
	OR	ds.IS?
	STA	H
	STI	(BC)
	LDA	(currentPROC)
	STI	(BC)
	LDA	D
	STI	(BC)
	LDA	E
	STI	(BC)
	INC	DE
	ADD	(DE),01
	LDA	H
	PUSH	A
	INC	DE
	LDI	(DE)
	STA	H
	LDI	(DE)
	STA	L
	POP	A
	BIT	fs.QUE
	JR	Z,tmp_0_612
	BIT	fs.REA
	JR	Z,tmp_0_62a
	LDI	(DE)
	AEX
	SL
	SL
	SCF
	SBC	05
	STI	(HL)
	LDA	(currentPROC)
	STI	(HL)
	CLA
	STI	(HL)
	STI	(HL)
	STI	(HL)
	JR	tmp_0_62a
tmp_0_612:
	BIT	fs.FIL
	JR	Z,tmp_0_620
	BIT	ds.APN
	JR	Z,tmp_0_62a
	INC	DE
	LDI	(BC)
	STI	(BC)
	LDI	(BC)
	JR	tmp_0_62c
tmp_0_620:
	BIT	fs.CON
	JR	Z,tmp_0_62a
	CALL	tmp_0_633
	CALL	tmp_0_633
tmp_0_62a:
	CLA
	STI	(BC)
tmp_0_62c:
	STA	(BC)
	SCF
	LDA	C
	SBC	05
	STA	E
	RET

tmp_0_633:
	LD	E,19
	LDA	0D
	CALL	tmp_0_63d
	LD	E,05
	CLA
tmp_0_63d:
	STI	(HL)
	DEC	E
	JR	C,tmp_0_63d
	RET

krn.fs.getDS:
	LD	B,<tableDS
	LDA	E
	STA	C
	INC	BC
	LDD	(BC)
	CPA	(currentPROC)
	JR	Z,tmp_0_651
	LD	H,err.OWN
tmp_0_64f:
	RCF
	RET
tmp_0_651:
	LDA	(BC)
	BIT	ds.IS?
	JR	NZ,tmp_0_659
	LD	H,err.NOD
	RCF
tmp_0_659:
	RET

;;-----------------------------------------------------------------------
;;	CLO - Close an opened entry
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	15
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Close the entry linked to the descriptor given by E. If the
;;		entry is a MAP, a DUPLEX, a CONSOLE or a LOCK, the entry is
;;		deleted from the file system.
;;		  + If the entry is a QUEUE and the descriptor has the READ
;;		    access, the entry is deleted, and all the other descriptors
;;		    linked to this entry are invalidated; else the close work
;;		    normally.
;;		  + If the entry is a DUPLEX, the connection is broken and the
;;		    entry is deleted.
;;		  + If the entry is a SHARED the entry will remind till the
;;		    last descriptor is closed.
;;		  + If the entry is a MAP, the block is unmapped and all the
;;		    other descriptors linked to this entry are invalidated.
;;		  + If the entry is a FILE, it closes the descriptor and the
;;		    entry reminds.
;;		  + If the entry is a CONSOLE and the current active CONSOLE is
;;		    linked to it, the kernel will display Closed! and no action
;;		    will be possible except to change the active CONSOLE by the
;;		    SELECT key.
;;		Closing an entry will decrease by 1 its number of links. 
;;
;;	Return:
;;		CARRY if success. The descriptor in E is invalidated and is no
;;		more usable.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.clo:
	CALL	krn.fs.getDS
	JR	C,krn.fs.closeDS
	CP	H,err.NOD
	JR	NZ,tmp_0_64f
krn.fs.closeDS:
	STA	H
	CLA
	STI	(BC)
	INC	BC
	LDI	(BC)
	STA	D
	LDI	(BC)
	STA	E
	LDA	H
	BIT	fs.!LOK
	JR	Z,tmp_0_6da
	BIT	fs.CON
	JR	Z,tmp_0_68c
	LDA	C
	SCF
	SBC	04
	CPA	(currentCON)
	JR	NZ,tmp_0_6d6
	AND	(currentCON),00
	BIT	(currentPROC),80
	JR	NZ,tmp_0_6d6
	CALL	tmp_1_2c8
	JR	tmp_0_6d6
tmp_0_68c:
	BIT	fs.QUE
	JR	Z,tmp_0_6da
	BIT	fs.DPX
	JR	Z,tmp_0_6b4
	LDA	(BC)
	JR	Z,tmp_0_6d6
	STA	C
	AND	(BC),FD
	INC	BC
	INC	BC
	LDI	(BC)
	STA	L
	LDI	(BC)
	AND	(BC),00
	STA	C
	LDA	L
	STA	B
	INC	BC
	INC	BC
	LDI	(BC)
	STA	L
	LDA	(BC)
	STA	C
	LDA	L
	STA	B
	INC	BC
	INC	BC
	CLA
	STI	(BC)
	STI	(BC)
	STA	(BC)
	JR	tmp_0_6d6
tmp_0_6b4:
	BIT	ds.REA
	JR	Z,tmp_0_6da
tmp_0_6b8:
	LDR	BC,tableDS
	LD	L,maxDS
tmp_0_6be:
	LDA	(BC)
	JR	Z,tmp_0_6d0
	INC	BC
	INC	BC
	LDA	D
	CPI
	JR	NZ,tmp_0_6c9
	LDA	E
	CPA	(BC)
tmp_0_6c9:
	DEC	BC
	DEC	BC
	DEC	BC
	JR	NZ,tmp_0_6d0
	AND	(BC),!ds.IS?
tmp_0_6d0:
	LDA	06
	ADD	BC
	DJC	tmp_0_6be
tmp_0_6d6:
	LDA	H
tmp_0_6d7:
	JP	tmp_0_461
tmp_0_6da:
	AND	(DE),!fs.XOP
	INC	DE
	ADD	(DE),FF
	BIT	70	; (fs.QUE|fs.FIL|fs.CON)
	JR	NZ,tmp_0_71c
	BIT	(DE),FF
	DEC	DE
	JR	Z,tmp_0_6d7
;; >>> Patch cgh - Close #0007 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;;	BIT	fs.SHR
;;	JR	Z,tmp_0_6f0
;;	BIT	fs.MAP
;;	JR	NZ,tmp_0_6b8
	BIT	fs.MAP
	JR	NZ,tmp_0_6b8
	BIT	fs.SHR
	JR	NZ,tmp_0_70b
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
tmp_0_6f0:
	LDA	C
	SCF
	SBC	04
	INC	DE
	INC	DE
	CPA	(DE)
	JR	Z,tmp_0_70d
	STA	H
	LDA	(DE)
tmp_0_6fb:
	RCF
	ADC	04
	STA	C
	LD	DE,BC
	LDA	(DE)
	JR	Z,tmp_0_70b
	CPA	H
	JR	NZ,tmp_0_6fb
	ADC	03
	STA	C
	LDI
tmp_0_70b:
	SCF
	RET

tmp_0_70d:
	LDA	(BC)
	STA	(DE)
	AND	(BC),00
	STA	C
	LDA	C
	JR	Z,tmp_0_71c
	INC	BC
	LDA	(BC)
	STA	C
	LD	B,<tablePROC
	AND	(BC),EF	; ps.!LOK
tmp_0_71c:
	SCF
	RET

tmp_0_71e:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_730
tmp_0_723:
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	E
	INC	DE
	INC	DE
	LDI	(DE)
	STA	B
	LDI	(DE)
	STA	C
	SCF
tmp_0_730:
	RET

tmp_0_731:
	PUSH	BC
	CALL	krn.fs.getDS
	JR	NC,tmp_0_76a
	BIT	fs.SHR
	JR	NZ,tmp_0_740
	LD	H,err.NOD
	JR	tmp_0_76a
tmp_0_740:
	LDA	D
	PUSH	A
	PUSH	HL
	RCF
	ADC	L
	STA	L
	JR	NC,tmp_0_74e
	INC	H
	JR	C,tmp_0_764
tmp_0_74e:
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	OR	05
	STA	E
	LDA	H
	CPA	(DE)
	JR	NC,tmp_0_76e
	JR	NZ,tmp_0_764
	LDA	L
	INC	DE
	CPA	(DE)
	DEC	DE
	JR	NC,tmp_0_76e
	JR	Z,tmp_0_76e
tmp_0_764:
	LD	H,err.SHR
	POP	A
	POP	HL
tmp_0_76a:
	POP	BC
	RCF
	RET
tmp_0_76e:
	POP	HL
	DEC	DE
	DEC	DE
	RCF
	LDD	(DE)
	ADC	L
	STA	C
	LDA	(DE)
	ADC	H
	STA	B
	POP	A
	STA	H
	POP	DE
	SCF
	RET
tmp_0_780:
	LDI
tmp_0_781:
	DEC	H
	JR	NC,tmp_0_788
	DJC	tmp_0_780
	INC	L
tmp_0_788:
	RET

tmp_0_789:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_7a7
	PUSH	A
	INC	BC
	INC	BC
	LDI	(BC)
	STA	H
	LDI	(BC)
	STA	L
	INC	BC
	INC	HL
	INC	HL
	INC	HL
	RCF
	LDD	(BC)
	ADC	(HL)
	DEC	HL
	STA	E
	LDI	(BC)
	ADC	(HL)
	INC	HL
	INC	HL
	STA	D
	POP	A
	SCF
tmp_0_7a7:
	RET

tmp_0_7a8:
	LDA	E
	CPA	(currentCON)
	JR	NZ,tmp_0_7b2
	OR	(HIGHLCDFLAG),conLCD.PCUR

tmp_0_7b2:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_7c7
	PUSH	HL
	LD	HL,BC
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDI	(BC)
	BIT	(HL),fs.FIL
	JR	Z,tmp_0_7c6
	OR	05
tmp_0_7c6:
	STA	E
tmp_0_7c7:
	BIT	(HL),fs.CON
	JR	Z,tmp_0_7d7
	INC	DE
	INC	DE
	LDI	(DE)
	PUSH	A
	LDA	(DE)
	OR	3A
	STA	E
	POP	A
	STA	D
tmp_0_7d7:
	LDA	(HL)
	POP	HL
	RET

;;-----------------------------------------------------------------------
;;	REA - Read data from a descriptor
;;
;;	Call:
;;		LD	BC,data-buffer
;;		LD	E,descriptor
;;		LD	L,number-of-bytes
;;		LDA	16
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Read the number of bytes given by L from the opened descriptor
;;		E and store the data to the buffer pointed by BC.
;;		While reading from a QUEUE or a DUPLEX entry, the number of
;;		bytes specified is discarded, and a complete message is
;;		stored. Only the process which created the QUEUE can read
;;		from it. The real number of bytes read is returned in L.
;;		Reading from a CONSOLE will suspend the process until the
;;		number of bytes is reached or the ENTER or the DOWN-ARROW
;;		is pressed; If the UP-ARROW is pressed, the error D3 (End
;;		of file) is returned; If only 1 byte is asked from a CONSOLE
;;		descriptor, REA will return even ENTER, DOWN-ARROW or UP-ARROW
;;		are not pressed.
;;		It is not possible to read from a LOCK nor a SHARED nor a MAP.
;;		To read, the descriptor should be opened with the READ mode.
;;
;;	Return:
;;		CARRY if success; L contains the number of bytes read.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		F4	Access denied
;;		D2	Not a valid DS
;;		D3	End of file
;;		D4	Queue is empty
;;
mmps.fs.rea:
	PUSH	BC
	CALL	krn.fs.getDS
	POP	BC
	JR	NC,tmp_0_8be
	BIT	ds.REA
	JR	Z,tmp_0_8c9
	BIT	fs.QUE
	JR	Z,tmp_0_839
	PUSH	BC
	CALL	tmp_0_71e
	POP	DE
	JR	NC,tmp_0_8be
tmp_0_7f5:
	PUSH	BC
	PUSH	DE
	LDI	(BC)
	STA	H
	INC	BC
	LDA	(BC)
	JR	NZ,tmp_0_803
	LD	H,err.QEM
tmp_0_801:
	JR	tmp_0_8b9
tmp_0_803:
	LD	DE,BC
	INC	BC
	RCF
	LDA	H
	SBC	(BC)
	STA	H
	LDI	(BC)
	INC	BC
	ADD	BC
	LDA	(BC)
	STA	L
	LDA	(DE)
	SCF
	SBC	L
	STI	(DE)
	POP	DE
	DEC	L
	LDA	L
	INC	BC
	JR	tmp_0_81c
tmp_0_81b:
	LDI
tmp_0_81c:
	DEC	H
	JR	C,tmp_0_829
	POP	BC
	PUSH	BC
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	INC	BC
tmp_0_829:
	DJC	tmp_0_81b
	POP	DE
	STA	L
	INC	DE
	INC	DE
	INC	DE
	SCF
	LDA	C
	SBC	E
	DEC	A
	DEC	A
	STA	(DE)
	SCF
	RET
tmp_0_839:
	BIT	A0
	JR	Z,tmp_0_880
	LDA	E
	STA	H
	PUSH	A
	LDA	L
	PUSH	A
	PUSH	BC
	LDA	H
	CALL	tmp_0_789
	JR	NC,tmp_0_8b9
	INC	HL
	INC	HL
	LDD	(HL)
	SBC	(BC)
	DEC	BC
	PUSH	A
	LDA	(HL)
	SBC	(BC)
	JR	Z,tmp_0_85d
	POP	A
	LDA	FF
	JR	tmp_0_85f
tmp_0_85d:
	POP	A
tmp_0_85f:
	STA	H
	LD	BC,DE
	POP	DE
	POP	A
	STA	L
	CALL	tmp_0_781
	SCF
	SBC	L
	STA	L
	POP	A
tmp_0_86f:
	STA	C
	LD	B,<tableDS
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	RCF
	LDA	(BC)
	ADC	L
	STD	(BC)
	LDA	(BC)
	ADC	N
	STD	(BC)
	SCF
	RET
tmp_0_880:
	BIT	10
	JR	Z,tmp_0_8c5
	PUSH	BC
	CALL	tmp_0_7a8
	POP	BC
	JR	NC,tmp_0_8bf
	LDA	E
	AND	C0
	STA	E
	LD	H,19
	LDA	0D
tmp_0_895:
	STI	(DE)
	DEC	H
	JR	C,tmp_0_895
	CLA
tmp_0_89b:
	STI	(DE)
	LD	H,08
tmp_0_89e:
	LDA	L
	STI	(DE)
	LDA	B
	STI	(DE)
	LDA	C
	STI	(DE)
	LDA	(currentPROC)
	STA	(DE)
	STA	C
	LD	B,<tablePROC
	LDA	H
	OR	(BC)
	STA	(BC)
tmp_0_8ae:
	BIT	(HIGHLCDFLAG),conLCD.PCUR
	JR	Z,tmp_0_8b7
	CALL	tmp_1_2c8
tmp_0_8b7:
	SCF
	RET
tmp_0_8b9:
	POP	BC
tmp_0_8bb:
	POP	DE
	RCF
tmp_0_8be:
	RET
tmp_0_8bf:
	AND	(HIGHLCDFLAG),!conLCD.PCUR
	RCF
	RET
tmp_0_8c5:
	LD	H,err.NOD
	RCF
	RET
tmp_0_8c9:
	LD	H,F4
	RCF
	RET

;;-----------------------------------------------------------------------
;;	WRI - Write data to a descriptor
;;
;;	Call:
;;		LD	BC,data-buffer
;;		LD	E,descriptor
;;		LD	L,number-of-bytes
;;		LDA	17
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Write the number of bytes given by L from the buffer pointed
;;		by BC to the descriptor given by E. The real number of bytes
;;		written is returned into L.
;;		Writing to a CONSOLE entry will suspend the process if a
;;		WRI request is already pending. If a process tries to write
;;		more bytes than possible, the system will write all the bytes
;;		until the end of file, discards the remaining and returns the
;;		real number of bytes written.
;;		When writing into a FILE descriptor, the real length is
;;		updated.
;;		When writing to a FILE more bytes than the amount free, the
;;		system will write all the bytes until the complete FILE space
;;		is full, and return the real number of bytes written.
;;		If a process tries to write more bytes than free into a
;;		QUEUE or a DUPLEX entry, an error is returned but nothing will
;;		be written.
;;		It is not possible to write to a LOCK nor to a SHARED or a
;;		MAP entry.
;;		To write, the descriptor should be opened with the WRITE
;;		mode.
;;
;;	Return:
;;		CARRY if success; L contains the real number of bytes written.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		F4	Access denied
;;		D2	Not a valid DS
;;		D5	Queue is full
;;
mmps.fs.wri:
	PUSH	BC
	CALL	krn.fs.getDS
	POP	BC
	JR	NC,tmp_0_940
	BIT	02
	JR	Z,tmp_0_8c9
	BIT	40
	JR	Z,tmp_0_941
	PUSH	BC
	CALL	krn.fs.getDS
	JR	NC,tmp_0_8bb
	BIT	01
	JR	Z,tmp_0_8f1
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	LDA	(BC)
	STA	C
	STA	E
	LDA	(BC)
tmp_0_8f1:
	CALL	tmp_0_723
	POP	DE
	JR	NC,tmp_0_8be
tmp_0_8f8:
	PUSH	BC
	PUSH	DE
	LD	DE,BC
	LDI	(DE)
	STA	H
	INC	DE
	RCF
	LDA	(DE)
	ADC	L
	JR	C,tmp_0_909
	CPA	H
	JR	NC,tmp_0_90e
tmp_0_909:
	LD	H,err.QFL
	JP	tmp_0_801
tmp_0_90e:
	INC	A
	STI	(DE)
	INC	DE
	LDA	H
	RCF
	SBC	(DE)
	STA	H
	LDI	(DE)
	ADD	DE
	LDA	L
	INC	A
	STI	(DE)
	POP	BC
	JR	tmp_0_920
tmp_0_91f:
	LDI
tmp_0_920:
	DEC	H
	JR	C,tmp_0_92d
	POP	DE
	PUSH	DE
	INC	DE
	INC	DE
	INC	DE
	INC	DE
	INC	DE
tmp_0_92d:
	DJC	tmp_0_91f
	POP	BC
	INC	BC
	LDI	(BC)
	STA	L
	LD	H,<tablePROC
	AND	(HL),DF
	INC	BC
	INC	BC
	SCF
	LDA	E
	SBC	C
	DEC	A
	STA	(BC)
	SCF
tmp_0_940:
	RET
tmp_0_941:
	BIT	A0
	JR	Z,tmp_0_99b
	LDA	E
	STA	H
	PUSH	A
	LDA	L
	PUSH	A
	PUSH	BC
	LDA	H
	CALL	tmp_0_789
	JR	NC,tmp_0_8b9
	LDA	(HL)
	STA	L
	SR
	SR
	STA	H
	LDA	L
	AND	03
	AEX
	SL
	SL
	STA	L
	SCF
	SBC	(BC)
	DEC	BC
	PUSH	A
	LDA	H
	SBC	(BC)
	JR	Z,tmp_0_96f
	POP	A
	LDA	FF
	JR	tmp_0_971
tmp_0_96f:
	POP	A
tmp_0_971:
	STA	H
	POP	BC
	POP	A
	STA	L
	CALL	tmp_0_781
	SCF
	SBC	L
	STA	L
	POP	A
	CALL	tmp_0_86f
	LDD	(BC)
	OR	05
	STA	E
	LDI	(BC)
	STA	D
	INC	BC
	LDI	(BC)
	CPA	(DE)
	JR	Z,tmp_0_993
	JR	NC,tmp_0_999
	STI	(DE)
	LDA	(BC)
	JR	tmp_0_998
tmp_0_993:
	INC	DE
	LDA	(BC)
	CPA	(DE)
	JR	NC,tmp_0_999
tmp_0_998:
	STA	(DE)
tmp_0_999:
	SCF
	RET
tmp_0_99b:
	BIT	10
	JR	Z,tmp_0_8c5
	PUSH	BC
	CALL	tmp_0_7a8
	POP	BC
	JR	NC,tmp_0_8bf
	CP	L,1B
	JR	NC,tmp_0_9ae
	LD	L,1A
tmp_0_9ae:
	LDA	(DE)
	JR	Z,tmp_0_9b7
	INC	DE
	LD	H,04
	JP	tmp_0_89e
tmp_0_9b7:
	LDA	L
	STI	(DE)
	AND	(DE),00
	LDA	E
	AND	E0
	STA	E
	LDA	L
	JR	tmp_0_9c3
tmp_0_9c2:
	LDI
tmp_0_9c3:
	DJC	tmp_0_9c2
	STA	L
	JP	tmp_0_8ae

;;-----------------------------------------------------------------------
;;	TEL - Get the current position of a descriptor
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	18
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the current position of the descriptor E into HL.
;;		For a CONSOLE, a QUEUE or a DUPLEX descriptor returns
;;		always 0. It is not possible to get the position from a
;;		LOCK nor a SHARED nor a MAP descriptor.
;;
;;	Return:
;;		CARRY if success; HL contains the current position.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.tel:
	CALL	tmp_0_7b2
	JR	NC,tmp_0_9da
	LD	H,00
	LD	L,00
	BIT	A0
	JR	Z,tmp_0_9db
	LDI	(BC)
	STA	H
	LDA	(BC)
	STA	L
tmp_0_9da:
	RET
tmp_0_9db:
	BIT	10
	JR	Z,tmp_0_9e1
	LDA	(DE)
	STA	L
tmp_0_9e1:
	RET

;;-----------------------------------------------------------------------
;;	SEK - Set the current position of a descriptor
;;
;;	Call:
;;		LD	E,descriptor
;;		LD	HL,offset
;;		LDA	19
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Set the current position of the descriptor E to the offset
;;		given by HL, and return the real offset set into HL. The
;;		offset is always specified from the beginning of the
;;		descriptor. It is not possible to set the position after the
;;		end. SEK is without effect for the CONSOLE or QUEUE or
;;		DUPLEX descriptors.
;;		It is not possible to set the position for a LOCK nor a
;;		SHARED nor a MAP descriptor.
;;
;;	Return:
;;		CARRY if success; HL contains the real offset set.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.sek:
	CALL	tmp_0_7b2
	JR	NC,tmp_0_9e7
tmp_0_9e7:
	BIT	A0
	JR	Z,tmp_0_9fe
	LDA	H
	CPA	(DE)
	JR	NC,tmp_0_9f8
	JR	NZ,tmp_0_a0d
	INC	DE
	LDA	L
	CPA	(DE)
	JR	NC,tmp_0_9f8
	JR	NZ,tmp_0_a0d
tmp_0_9f8:
	LDA	H
	STI	(BC)
	LDA	L
	STI	(BC)
	SCF
	RET
tmp_0_9fe:
	BIT	10
	JR	Z,tmp_0_a0c
	CP	H,00
	JR	NZ,tmp_0_a0d
	CP	L,3A
	JR	C,tmp_0_a0d
	LDA	L
	STA	(DE)
tmp_0_a0c:
	RET
tmp_0_a0d:
	LD	H,err.EOF
	RCF
	RET

;;-----------------------------------------------------------------------
;;	WAI - Wait from a descriptor
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	1A
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Suspend the calling process until a message is available
;;		from the QUEUE or the DUPLEX descriptor given by E, and
;;		return the number of bytes present in the QUEUE or the
;;		DUPLEX into L. Waiting from a CONSOLE or from a FILE always
;;		return, but is not suspensive.
;;		It is not possible to wait from a LOCK nor a SHARED nor a
;;		MAP entry.
;;		Only the process which have the READ access can wait from it.
;;
;;	Return:
;;		CARRY if success; L contains the number of bytes present
;;		in the QUEUE.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid descriptor
;;
mmps.fs.wai:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_a28
	BIT	fs.QUE
	JR	Z,tmp_0_a27
	CALL	tmp_0_723
	INC	BC
	LDI	(BC)
	STA	E
	LD	D,<tablePROC
	LDA	(BC)
	JR	NZ,tmp_0_a27
	OR	(DE),21
tmp_0_a27:
	SCF
tmp_0_a28:
	RET

;;-----------------------------------------------------------------------
;;	CNT - Connect a DUPLEX to a peer DUPLEX
;;
;;	Call:
;;		LD	BC,name
;;		LD	E,descriptor
;;		LDA	A,1B
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Realize a connection between the DUPLEX descriptor given in
;;		E and the peer DUPLEX named by BC. To be successful, both
;;		DUPLEX should be free from any connection.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid descriptor
;;		D6	Can not connect
;;
mmps.fs.cnt:
	PUSH	BC
	CALL	krn.fs.getDS
	LD	DE,BC
	POP	BC
	JR	NC,tmp_0_a77
	CPA	4D
	JR	NZ,tmp_0_a5d
	PUSH	DE
	CALL	krn.fs.findENTRY
	POP	BC
	JR	NC,tmp_0_a77
	LDA	C
	STA	H
	LD	C,>tableDS
	LD	L,maxDS
tmp_0_a47:
	LDA	(BC)
	JR	Z,tmp_0_a57
	INC	BC
	INC	BC
	LDA	D
	CPI
	JR	NZ,tmp_0_a52
	LDA	E
	CPA	(BC)
tmp_0_a52:
	DEC	BC
	DEC	BC
	DEC	BC
	JR	Z,tmp_0_a61
tmp_0_a57:
	LDA	06
	ADD	BC
	DJC	tmp_0_a47
tmp_0_a5d:
	LD	H,err.CNT
	RCF
	RET
tmp_0_a61:
	LDA	(BC)
	CPA	4D
	JR	NZ,tmp_0_a5d
	LDA	C
	STA	L
	LDA	H
	CALL	tmp_0_a6f
	LDA	H
	STA	C
	LDA	L
tmp_0_a6f:
	OR	(BC),02
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	STA	(BC)
	SCF
tmp_0_a77:
	RET

;;-----------------------------------------------------------------------
;;	GET - Get data from a SHARED/MAP descriptor
;;
;;	Call:
;;		LD	BC,data-buffer
;;		LD	E,descriptor
;;		LD	H,offset
;;		LD	L,number-of-bytes
;;		LDA	1C
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Get the number of bytes L located at the offset H from the
;;		SHARED or MAP descriptor E and put then into the buffer
;;		pointed by BC. If the offset H + the number of bytes L are
;;		greater than the SHARED or MAP, an error is returned and no
;;		get is done.
;;		It is not possible to do this action on a non SHARED or MAP
;;		entry.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;		D7	Out of shared
;;
mmps.fs.get:
	CALL	tmp_0_731
	JR	NC,tmp_0_a90
	PUSH	DE
	LD	DE,BC
	POP	BC
	JR	tmp_0_a8d

;;-----------------------------------------------------------------------
;;	PUT - Put data to a SHARED/MAP descriptor
;;
;;	Call:
;;		LD	BC,data-buffer
;;		LD	E,descriptor
;;		LD	H,offset
;;		LD	L,number-of-bytes
;;		LDA	1D
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Put the number of bytes L located at the offset H from the
;;		buffer pointed by BC to the SHARED or MAP descriptor E. If
;;		the offset H + the number of bytes L are greater than the
;;		SHARED or MAP, an error is returned and no put is done.
;;		It is not possible to do this action on a non SHARED or MAP
;;		entry.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;		D7	Out of shared
;;
mmps.fs.put:
	CALL	tmp_0_731
	JR	NC,tmp_0_a90
	JR	tmp_0_a8d
tmp_0_a8c:
	LDI
tmp_0_a8d:
	DJC	tmp_0_a8c
	SCF
tmp_0_a90:
	RET

tmp_0_a91:
	CALL	tmp_0_7b2
	JR	NC,tmp_0_aa2
	BIT	F0
	JR	NZ,tmp_0_b30
	LDA	C
	SCF
	SBC	04
	STA	L
	INC	DE
	INC	DE
	LDA	(DE)
tmp_0_aa2:
	RET

;;-----------------------------------------------------------------------
;;	GIV - Give a LOCK
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	1E
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Release the LOCK pointed by the descriptor in E. If a
;;		process is waiting for the LOCK (TAK), it will be
;;		rescheduled and will become owner of the descriptor.
;;		It is not possible to do this action on a non LOCK entry.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.giv:
	CALL	tmp_0_a91
	JR	NC,tmp_0_aa2
	JR	Z,tmp_0_ab0
	CPA	L
	JR	NZ,tmp_0_ab0
	JP	tmp_0_70d
tmp_0_ab0:
	JP	tmp_0_191

;;-----------------------------------------------------------------------
;;	TAK - Take a LOCK
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	1F
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Take the LOCK pointed by the descriptor in E. If a process
;;		is already owner of the LOCK, the calling process will be
;;		suspended until the LOCK is released (GIV).
;;		It is not possible to do this action on a non LOCK entry.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.tak:
	CALL	tmp_0_a91
	JR	NC,tmp_0_ad3
	JR	Z,tmp_0_ad0
	CPA	L
	JR	Z,tmp_0_ad2
tmp_0_abd:
	STA	E
	CPA	L
	JR	Z,tmp_0_ad2
	INC	DE
	INC	DE
	INC	DE
	INC	DE
	LDA	(DE)
	JR	NZ,tmp_0_abd
	LDA	(currentPROC)
	STA	C
	LD	B,<tablePROC
	OR	(BC),ps.LOK
tmp_0_ad0:
	LDA	L
	STA	(DE)
tmp_0_ad2:
	SCF
tmp_0_ad3:
	RET

;;-----------------------------------------------------------------------
;;	SCN - Set a CONSOLE active
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	20
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Activate the CONSOLE linked to the descriptor E. It is
;;		possible only if the current active CONSOLE is closed, or
;;		if the current active  CONSOLE is owned by the calling
;;		process, and has no request pending.
;;		Activate a CONSOLE will start the CONSOLE-IO process on
;;		this CONSOLE.
;;		The current CONSOLE, the process-id PID and the process
;;		name linked to are displayed by pressing the OFF key.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.scn:
	LDA	(currentCON)
	JR	Z,tmp_0_ae6
	CPA	E
	JR	Z,tmp_0_b02
	PUSH	DE
	STA	E
	CALL	krn.fs.getDS
	POP	DE
	JR	NC,tmp_0_b03
tmp_0_ae6:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_b03
	BIT	10
	JR	Z,tmp_0_bd2
	LDA	E
	STA	(currentCON)
	INC	BC
	INC	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	STA	L
	INC	HL
	INC	HL
	LDI	(HL)
	STA	D
	LDA	(HL)
	STA	E
	CALL	tmp_1_2c8
tmp_0_b02:
	SCF
tmp_0_b03:
	RET

;;-----------------------------------------------------------------------
;;	FLU - Flush a CONSOLE output
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	21
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Flush the output request of the CONSOLE descriptor E, if
;;		there is one, and suspend the calling process like a WRI
;;		request were pending. This action can be used to activate
;;		the output before to end a process.
;;		The pending message is displayed until the RCL is pressed
;;		to refresh the console.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.flu:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_b28
	BIT	10
	JR	Z,tmp_0_c0a
	INC	BC
	INC	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	STA	L
	INC	HL
	INC	HL
	LDI	(HL)
	STA	D
	LDA	(HL)
	OR	3A
	STA	E
	LDI	(DE)
	INC	DE
	INC	DE
	INC	DE
	JR	Z,tmp_0_b27
	CALL	tmp_0_bda
	OR	(BC),04
	STA	(DE)
tmp_0_b27:
	SCF
tmp_0_b28:
	RET

;;-----------------------------------------------------------------------
;;	RFR - Refresh a CONSOLE output
;;
;;	Call:
;;		LD	E,descriptor
;;		LDA	22
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Refresh the output request of the CONSOLE descriptor E, if
;;		there is one, kill all pending message. This avoid the
;;		calling process to be suspended if a WRI is pending.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.rfr:
	CALL	tmp_0_7b2
	JR	NC,tmp_0_b34
	BIT	10
tmp_0_b30:
	JR	Z,tmp_0_c0a
	AND	(DE),00
tmp_0_b34:
	RET

;;-----------------------------------------------------------------------
;;	EDI - Start a editable REA on a CONSOLE
;;
;;	Call:
;;		LD	BC,buffer
;;		LD	E,descriptor
;;		LD	H,cursor
;;		LD	L,number-of-bytes
;;		LDA	23
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Fill the input buffer of a CONSOLE descriptor given in E with
;;		the string in the buffer pointed by BC, start editing with
;;		the cursor at the position H, make a READ request and return
;;		the real number of bytes read in L.
;;		This action looks like REA but the input buffer is filled
;;		before instead of to be cleaned, and the process can specify
;;		the initial position of the cursor.
;;		The buffer is filled by the string present in the buffer,
;;		until a non-printable character is found (character < &20),
;;		after fill the remaining buffer with the CR character (&0D).
;;		If the cursor is specified after the string end, it will be
;;		put on last character before CR.
;;		If the cursor is after the number of bytes read, it will
;;		return at the first character read.
;;		It is also possible to give an empty string.
;;		The input buffer will be filled up to 26 characters.
;;		At the return, the buffer pointed by BC will be filled with the
;;		characters read and the number of characters will be return
;;		into L.
;;		It is not possible to use this action if a REA request is
;;		already running.
;;
;;	Return:
;;		CARRY if success; L contains the number of bytes read.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D2	Not a valid DS
;;
mmps.fs.edi:
	PUSH	HL
	PUSH	BC
	CALL	tmp_0_7a8
	JR	NC,tmp_0_b6b
	BIT	fs.CON
	JR	Z,tmp_0_b69
	LDA	E
	AND	C0
	STA	E
	POP	BC
	PUSH	BC
	LD	L,19
tmp_0_b4c:
	LDI	(BC)
	CPA	20
	JR	NC,tmp_0_b54
	STI	(DE)
	DJC	tmp_0_b4c
tmp_0_b54:
	LDA	E
	AND	1F
	STA	H
	LDA	0D
tmp_0_b5a:
	STI	(DE)
	DJC	tmp_0_b5a
	LDA	H
	POP	BC
	POP	HL
	CPA	H
	JR	NC,tmp_0_b66
	LDA	H
tmp_0_b66:
	JP	tmp_0_89b
tmp_0_b69:
	LD	H,err.NOD
tmp_0_b6b:
	POP	BC
	POP	BC
	JP	tmp_0_8bf

;;-----------------------------------------------------------------------
;;	HDR - Build an executable header
;;
;;	Call:
;;		LD	D,number-of-page
;;		LD	E,descriptor
;;		LD	HL,global-address-offset
;;		LDA	24
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Build an executable header in the FILE entry pointed by the
;;		descriptor given by E. The number of pages to be allocated
;;		when launching a process (XEQ) is given by D and the offset
;;		of the global address in HL.
;;		The executable header as the following format:
;;		byte 0:	  executable signature &F3
;;		byte 1,2: real-size (copy of the bytes 5,6 of the entry)
;;		byte 3:   number of pages to allocate for the stack and the
;;			  global area excluding the arguments-page.
;;		byte 4,5: offset of the global area in the global block.
;;		The descriptor should be opened in WRITE mode.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		F4	Access denied
;;		D2	Not a valid DS
;;
mmps.fs.hdr:
	LDA	D
	AND	03
	AEX
	SL
	SL
	STA	C
	LDA	D
	AND	FC
	SL
	SL
	AEX
	STA	B
	CPA	H
	JR	NC,tmp_0_bce
	JR	NZ,tmp_0_b89
	LDA	C
	CPA	L
	JR	NC,tmp_0_bce
tmp_0_b89:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_bd1
	BIT	20
	JR	Z,tmp_0_bd2
;; >>> Patch cgh - Header #0006 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;;	BIT	04
;;	JR	Z,tmp_0_bd2
	BIT	02
	JR	Z,tmp_0_bd6
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	LDA	05
	ADD	BC
	CLA
	STD	(BC)
	STD	(BC)
	LDD	(BC)
	PUSH	A
	LDA	(BC)
	STA	B
	POP	A
	STA	C
	BIT	(BC),01
	JR	Z,tmp_0_bd6
	LDA	40
	ADD	HL
	LDA	06
	ADD	BC
	PUSH	HL
	STA	L
	LDA	D
	PUSH	A
	LDD	(BC)
	PUSH	A
	LDA	(BC)
	PUSH	A
	LD	BC,SP
	LDA	(1+mmps.krn.sgn)
	PUSH	A
	CALL	mmps.fs.wri
	POP	MN
	POP	MN
	POP	MN
	RET
tmp_0_bce:
	LD	H,err.MEM
	RCF
tmp_0_bd1:
	RET
tmp_0_bd2:
	LD	H,err.NOD
	RCF
	RET
tmp_0_bd6:
	LD	H,err.DEN
	RCF
	RET
tmp_0_bda:
	LDA	(currentPROC)
	STA	C
	LD	B,<tablePROC
	RET



;;=======================================================================
;;	PROCESS-SCHEDULER : ps
;;
;;	It is the third and the highest level. It manages the process start,
;;	the process time slot allocation, the inter-process messages and the
;;	events control. It can create up to 8 processes, with priority and
;;	father return. When a process starts a new one, it becomes the
;;	'FATHER' and the started process the 'SON'. A process is an entry
;;	in the process table, marked as active and recognized by the scheduler
;;	which will restart it. The scheduler allows some time slots to the
;;	processes; this number of slots given depends of its priority. The
;;	priority is a number from 0 (highest priority) to 15 (lowest priority).
;;	The priority of a process is the same as this of its father, but it
;;	can change it. A process of priority 0 will get 16 consecutive slots
;;	for its execution, and only 1 if its priority is 15. All the slots
;;	have the same period. The default priority is 8, so 8 slots.
;;
;;	All the processes started receive an unique number called process-id
;;	(PID); this process-id permits to another process to access to the
;;	named process.
;;	The information recorded into the process entry are the process status
;;	(PAUSED, WAITING, ...), its priority, its stack pointer SP, the delay
;;	counters (PAUSE and TIME), the global address, its father process-id
;;	(FID) and a link to a FILE entry in the file-system.
;;
;;	When the scheduler starts a new process, it first asks to the file-system
;;	to find the named entry, checks if the type (FILE) and the access
;;	rights are good (READ and EXECUTE). After, it reads the 6 first bytes
;;	of the FILE to find some specific information like the executable
;;	signature, the number of pages to allocate for the stack and the
;;	global area. The father gives to its son the in and out descriptors
;;	and an address to a page called arguments page which will be copied
;;	into the son process space; the father can pass up to 63 bytes in the
;;	arguments page, the last is always 0. The FILE header contains how many
;;	pages should be asked for the stack and the offset of the global area.
;;	When the scheduler will start a process, it will create its stack. This
;;	stack is built in the new process space. This space is allocated and
;;	has the number of pages needed by the process (excluding the arguments
;;	page which is allocated in same block. In fact, the scheduler add 1 to
;;	the number of pages), add the offset of the global area. The global
;;	area is the memory area allocated to the process after the top of the
;;	stack.
;;	This space is very important, because its address is passed to the son
;;	process at the beginning. It is used to store some global data which
;;	may be used later. The address GLOBAL - 2 contains always the address
;;	of the arguments page. If the space area is allocated and the in and out
;;	descriptors are valid, the son process is created and started. The father
;;	gets as return the process-id of its son.
;;	When its son will terminate, the scheduler will warn the father by using
;;	an 'EVENT'. An EVENT is in fact an action which will ask to the scheduler
;;	to send an event message into the event queue of a process. The receiver
;;	process has to poll its event queue to get the oldest event. A process
;;	may enable or disable some events by changing its EVENT mask. On MMPS,
;;	8 events are supported: 4 are reserved for the scheduler and the 4 others
;;	are free for the other processes.
;;
;;	The 4 events reserved by the scheduler are: 
;;	  + 4	The son has returned to scheduler. In this case the value DE of
;;		the son is returned to the father. The scheduler 'PUSH'es a return
;;		address into the stack where the process has to return when it is
;;		finished. A simple RET is needed to return from a process.
;;	  + 5	The son was terminated by an external event.
;;	  + 6	The time alarm is reached.
;;	  + 7	The BREAK key was pressed on the CONSOLE linked to this process.
;;		This event is particular, because if it is not handled, the
;;		process will be terminated. It is a case to raise an event 5 to
;;		its father.
;;
;;	The 4 other events (0 to 3) are without specific meaning, and are free
;;	for usage. When raising an event 0 to 3, the DE value of the calling
;;	process will be passed to the triggered one.
;;	While polling its event queue, a process will get some information from
;;	the kernel, the number of the event and the process-id of the process
;;	which raised it, and eventually the value of DE from the raising process.
;;	A process can raise an event to itself. The event queue is filled even
;;	the process is suspended.
;;
;;	Another feature for process communication provided by the scheduler is
;;	the messages. To use them, a process is the receiver and another the
;;	sender. When a process wants to receive some message, it asks to the
;;	scheduler to give to it the state 'READY-TO-RECEIVE' and will be suspended
;;	until a message will be sent. When a process wants to send a message to
;;	another, it asks to scheduler to send it. This is possible only if the
;;	target process is ready. A message is a set of bytes which will be copied
;;	from the sender space to the receiver space. The block needed is allocated
;;	by scheduler and returned to the receiver which will be restarted. The
;;	messages mechanism provides tools for synchronous inter-process-communication.
;;	A process can also SEND-and-RECEIVE a message; it can be useful for a
;;	message with acknowledgment. A message is a set from 0 to 255 bytes.
;;
;;	When a process is finished (returned or terminated), all its resources,
;;	descriptors and pages will be released, and its father will be warned. If
;;	a process has no more father, the FID will be set to the value FF. When a
;;	father is terminated, its son are still running and are not warned.
;;
;;	The PROCESS-TABLE structure
;;	byte 0: process status (= &FF ::= free)
;;		bit7 ::= READY-TO-RECEIVE
;;		bit6 ::= PAUSED
;;		bit5 ::= QUEUE/DUPLEX WAITING
;;		bit4 ::= LOCK WAITING
;;		bit3 ::= CONSOLE-INPUT
;;		bit2 ::= CONSOLE-OUTPUT
;;		bit1 ::= TRACED
;;		bit0 ::= KERNEL-JUMP RUNNING
;;	byte 1: priority and slot-counter
;;		bit7-4 ::= priority (0 .. 15)
;;		bit3-0 ::= slot counter
;;	byte 2,3: DELAY counter
;;	byte 4,5: stack pointer SP
;;	byte 6: events mask
;;		bit7 ::= event7
;;		bit0 ::= event0
;;	byte 7,8: event queue address
;;	byte 9,10: TIMER counter
;;	byte 11,12: global address
;;	byte 13: father PID and extended process status
;;		bit7-4 ::= father internal-PID (internal-FID)
;;			&0 ::= pid0
;;			&1 ::= pid1
;;			&2 ::= pid2...
;;			&7 ::= pid7
;;			&A ::= KERNEL
;;			&F ::= no father
;;		bit3 ::= not used
;;		bit2 ::= not used
;;		bit1 ::= TIMER running
;;		bit1 ::= DEBUG RUNNING
;;	byte 14,15: physical address of the process entry
;;

;;-----------------------------------------------------------------------
;;	GBL - Get the global address
;;
;;	Call:
;;		LDA	25
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the global address of the calling process into BC.
;;		The global address of a process is located on the top of
;;		the stack at the offset fixed by building the executable
;;		header (HDR).
;;		BC - 2 is a pointer to the arguments page initialized by
;;		XEQ when starting. The global address is the same value
;;		set in BC when a process is started and (BC - 2) is the
;;		value of HL.
;;
;;	Return:
;;		CARRY and BC contains the global address.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.gbl:
	CALL	tmp_0_bda
	LDA	0B
	ADD	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	STA	L
	LD	BC,HL
	SCF
	RET

;;-----------------------------------------------------------------------
;;	PID - Get the process-id
;;
;;	Call:
;;		LDA	26
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the process-id PID of the calling process into H.
;;
;;	Return:
;;		CARRY and H contains the process PID.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.pid:
	LDA	(currentPROC)
tmp_0_bf3:
	AEX
	STA	H
	SCF
	RET

;;-----------------------------------------------------------------------
;;	FID - GET the father process-id
;;
;;	Call:
;;		LDA	27
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the father process-id FID of the calling process
;;		into H.
;;
;;	Return:
;;		CARRY and H contains the father process FID.
;;
;;	Error Code:
;;		None.
;;
mmps.fs.fid:
	CALL	tmp_0_bda
	LDA	0D
	ADD	BC
	LDA	(BC)
	JR	tmp_0_bf3

tmp_0_c01:
	CALL	krn.fs.getDS
	JR	NC,tmp_0_c0d
	BIT	fs.!LOK
	JR	NZ,tmp_0_c0d
tmp_0_c0a:
	LD	H,err.DSX
	RCF
tmp_0_c0d:
	RET

tmp_0_c0e:
	PUSH	BC
	PUSH	DE
	INC	A
	INC	A
	INC	A
	STA	E
	LD	D,<tableDS
	LDD	(DE)
	STA	C
	LDD	(DE)
	STA	B
	LDA	L
	STD	(DE)
	BIT	(DE),40
	JR	Z,tmp_0_c2d
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	C
	LDA	D
	STA	B
	INC	BC
	LDA	L
	STA	(BC)
tmp_0_c2d:
	LDA	E
	POP	DE
	POP	BC
	RET

;;-----------------------------------------------------------------------
;;	XEQ - Execute a process
;;
;;	Call:
;;		LD	BC,name
;;		LD	D,input-descriptor
;;		LD	E,output-descriptor
;;		LD	HL,arguments
;;		LDA	28
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Start the process named by BC; pass to it the descriptors in
;;		and out given by D and E and the arguments page pointed by
;;		HL. The son process-id is returned to the father into H. The
;;		son process becomes owner of the descriptors D and E. The
;;		arguments page is copied into the son space with a limit of
;;		63 bytes; the last is always 0. The descriptors in and out
;;		given to the son should be owned by the father and be valid.
;;		It is possible to give only descriptors to CONSOLE, FILE,
;;		DUPLEX or QUEUE. The descriptor D and E may be the same.
;;
;;		The son process starts with the global address into BC, the
;;		descriptors in and out into D and E and the arguments page
;;		pointer into HL. The arguments page address is copied into
;;		(BC - 2).
;;		The son process gets the same priority as its father, its
;;		event queue is removed and the event mask is set to 0.
;;		A return address is pushed into the stack to handle the son
;;		termination.
;;		When the top-level RET is executed, the process is terminated,
;;		all its resources are released, all its descriptors closed
;;		and the event 4 is raised to its father with the son returned
;;		value of DE.
;;		When a father is returned, all its sons are still running, but
;;		their father process-id is set to &FF.
;;		The executable FILE should have the access READ and EXECUTE.
;;		Executing a process will increase by 1 the number of links of
;;		the entry; when the process returns, the number of links will
;;		be decrease by 1.
;;		It is not possible to start a process if the FILE entry is
;;		exclusively opened.
;;		The number of pages allocated to the son while starting are set
;;		into the executable FILE header. See the kernel-jump HDR for
;;		a description of executable header.
;;
;;	Return:
;;		CARRY if success; H contains the son process-id.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D1	Exclusively opened
;;		D2	Not a valid DS
;;		F1	No such entry
;;		F4	Access denied
;;		C0	No process free
;;		C2	Not a valid DS for XEQ
;;		C3	Not a valid signature
;;
mmps.ps.xeq:
	PUSH	HL
	PUSH	DE
	PUSH	BC
	LDA	E
	CALL	tmp_0_c01
	JR	NC,tmp_0_d37
	LDA	D
	CALL	tmp_0_c01
	JR	NC,tmp_0_d37
	POP	BC
	CALL	krn.fs.findENTRY
	JR	NC,tmp_0_d39
	LDI	(DE)
	AND	2D
	CPA	25
	JR	NZ,tmp_0_d2f
	INC	DE
	LDI	(DE)
	STA	H
	LDI	(DE)
	STA	L
	LDI	(HL)
	CPA	(1+mmps.krn.sgn)
	JR	NZ,tmp_0_d2f
	INC	DE
	LDI	(DE)
	CPA	(HL)
	INC	HL
	JR	NZ,tmp_0_d2f
	LDD	(DE)
	CPA	(HL)
	JR	NZ,tmp_0_d2f
	INC	HL
	LDA	E
	AND	F0
	STA	E
	LDR	BC,tablePROC
tmp_0_c71:
	LDA	(BC)
	CPA	FF
	JR	Z,tmp_0_c82
	LDA	10
	ADD	BC
	CP	C,>maxIndexPROC
	JR	NC,tmp_0_c71
	LD	H,err.PRO
	JR	tmp_0_d39
tmp_0_c82:
	PUSH	BC
	PUSH	DE
	LDA	0F
	ADD	BC
	LDA	E
	STD	(BC)
	LDA	D
	STD	(BC)
	LDA	(currentPROC)
	STA	E
	STD	(BC)
	LDA	FF
	STD	(BC)
	STD	(BC)
	STD	(BC)
	STD	(BC)
	STD	(BC)
	STD	(BC)
	CLA
	STD	(BC)
	STD	(BC)
	STD	(BC)
	STD	(BC)
	STD	(BC)
	LD	D,<tablePROC
	LDA	80
	CP	E,>maxIndexPROC
	JR	C,tmp_0_cab
	INC	DE
	LDA	(DE)
tmp_0_cab:
	AND	F0
	STA	E
	AEX
	RCF
	ADC	E
	STD	(BC)
	LDI	(HL)
	JR	NZ,tmp_0_cb6
	INC	A
tmp_0_cb6:
	INC	A
	PUSH	HL
	STA	L
	LDA	C
	STA	D
	CALL	krn.bm.allocBLOCK_pid_D
	JR	NC,tmp_0_d33
	POP	HL
	POP	DE
	INC	DE
	ADD	(DE),01
	LD	DE,BC
	LDA	2-bm.PSZ	; page size - 2 byte for argument page @
	ADD	BC
	INC	HL
	LDD	(HL)
	RCF
	ADC	C
	STA	C
	LDI	(HL)
	ADC	B
	STA	B
	INC	HL
	PUSH	BC
	DEC	BC
	LDA	(2+mmps.krn.endproc)
	STD	(BC)
	LDA	(1+mmps.krn.endproc)
	STD	(BC)
	LDA	03
	STD	(BC)
	LDA	L
	STD	(BC)
	LDA	H
	STD	(BC)
	CLA
	STD	(BC)
	POP	HL
	LDA	L
	STD	(BC)
	LDA	H
	STD	(BC)
	POP	HL
	DEC	BC
	POP	A
	STI	(BC)
	CALL	tmp_0_c0e
	POP	A
	STD	(BC)
	CALL	tmp_0_c0e
	DEC	BC
	LDA	E
	STD	(BC)
	LDA	D
	STD	(BC)
	CLA
	STI	(HL)
	INC	HL
	INC	HL
	INC	HL
	LDA	B
	STI	(HL)
	LDA	C
	STI	(HL)
	LDA	05
	ADD	HL
	LDA	0D
	ADD	BC
	LDA	D
	STI	(BC)
	LDA	E
	STI	(BC)
	LDA	B
	STI	(HL)
	LDA	C
	STA	(HL)
	LDA	L
	AND	F0
	AEX
	STA	H
	POP	BC
	LD	L,3F
tmp_0_d27:
	LDI	(BC)
	STI	(DE)
	JR	Z,tmp_0_d2d
	DJC	tmp_0_d27
tmp_0_d2d:
	SCF
	RET
tmp_0_d2f:
	LD	H,err.SGN
	JR	tmp_0_d39
tmp_0_d33:
	POP	BC
	POP	BC
tmp_0_d37:
	POP	BC
tmp_0_d39:
	POP	BC
	POP	BC
	RCF
	RET

krn.ps.getPROC:
	LDA	H
	AEX
	CPA	FF
	JR	NZ,tmp_0_d48
	LDA	(currentPROC)
tmp_0_d48:
	BIT	8F
	JR	NZ,tmp_0_d54
tmp_0_d4c:
	STA	C
	LD	B,<tablePROC
	LDA	(BC)
	CPA	FF
	JR	NZ,tmp_0_d2d
tmp_0_d54:
	LD	H,err.NOP
	RCF
	RET

;;-----------------------------------------------------------------------
;;	END - Terminate a process
;;
;;	Call:
;;		LD	H,process-id or &FF
;;		LDA	29
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Terminate the process pointed by H or itself if &FF. When
;;		a process is terminated, all its resources are released,
;;		all its descriptors are closed and the event 5 is raised
;;		to its father.
;;		All its sons are still running, but their father process-id
;;		FID is set to &FF.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;
mmps.fs.end:
	CALL	krn.ps.getPROC
	JR	NC,tmp_0_df7
tmp_0_d5d:
	LD	L,evt.END

	;; End a process:
	;;  - the father (if any) receive the evt.END or evt.PID event;
	;;  - if the process is a debugger leader, all his sons process
	;;    are ended (no orphan debugged process)
	;;  - if the ending process has some sons process, these become
	;;    orphans (&F0)
	;;  - close all descriptors open by the ending process
	;;  - release all blocks open by the ending process
	;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	;; cgh: I forgot here to wake-up the debugger leader if a son
	;;	is ended.... Yes ! But it is not so simple :-(
	;;	Because the debugger leader should be awaken only if
	;;	the ending process is CURRENTLY this under TRA control
	;;	else, debugger leader think the instruction is executed
	;;	and I do not know how to avoid that...
	;;
	;; #0004 - Debug leader not awaken if traced process ends
	;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
krn.ps.endME:
	OR	(BC),ps.FRE
	LDA	C
	STA	(endingPROC)
	LDA	0D
	ADD	BC
	LDA	C
	AND	F0
	STA	H
	LDI	(BC)
	AND	F0
	CPA	80
	JR	C,tmp_0_d7b
	PUSH	BC
	CALL	tmp_0_ed6
	POP	BC
tmp_0_d7b:
	LDI	(BC)
	STA	H
	LDD	(BC)
	STA	L
	INC	HL
	ADD	(HL),FF
	DEC	BC
	LD	L,07
tmp_0_d85:
	LDA	(BC)
	AND	F0
	CPA	(endingPROC)
	JR	NZ,tmp_0_daa
	BIT	(BC),ps.DBG
	JR	Z,tmp_0_da8
	LDA	(endingPROC)
	STA	H
	PUSH	HL
	PUSH	BC
	LDA	C
	AND	F0
	STA	C
	CALL	krn.ps.endME
	POP	BC
	POP	HL
	LDA	H
	STA	(endingPROC)
tmp_0_da8:
	OR	(BC),ps.!PID
tmp_0_daa:
	LDA	10
	ADD	BC
	DJC	tmp_0_d85

	;; Release all descriptors open by the ending process
tmp_0_db0:
	LDR	BC,tableDS
tmp_0_db4:
	LDI	(BC)
	JR	Z,tmp_0_dcc
	LDA	(endingPROC)
	CPA	80
	JR	C,tmp_0_dc1
	CPA	(BC)
	JR	NZ,tmp_0_dcc
tmp_0_dc1:
	DEC	BC
	LDA	(BC)
	PUSH	BC
	LDA	(BC)
	CALL	krn.fs.closeDS
	POP	BC
	INC	BC
tmp_0_dcc:
	LDA	05
	ADD	BC
	.IF	EQUAL?	<tableDS,<maxIndexDS
	CP	C,>maxIndexDS
	.ELSE
	CP	B,<maxIndexDS
	.ENDIF
	JR	NC,tmp_0_db4

	;; Release all blocks allocated to the ending process
	LDR	BC,tableHEAP
tmp_0_dd8:
	LDA	(endingPROC)
	CPA	80
	JR	NC,tmp_0_de3
	BIT	(BC),80
	JR	Z,tmp_0_de6
tmp_0_de3:
	CPA	(BC)
	JR	NZ,tmp_0_ded
tmp_0_de6:
	LDA	C
	STA	E
	INC	BC
	CALL	krn.bm.freeBLOCK
	DEC	BC
tmp_0_ded:
	INC	BC
	LDD	(BC)
	SL
	ADD	BC
	.IF	EQUAL?	<tableHEAP,<maxIndexHEAP
	CP	C,>maxIndexHEAP
	.ELSE
	CP	B,<maxIndexHEAP
	.ENDIF
	JR	NC,tmp_0_dd8
	SCF
tmp_0_df7:
	RET

;;-----------------------------------------------------------------------
;;	PRI - Change the priority of a process
;;
;;	Call:
;;		LD	H,process-id
;;		LD	L,new-priority
;;		LDA	2A
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Change the priority of the process pointed by H (or itself
;;		if &FF) with the priority in L. The priority is a number
;;		from &0 to &F. &0 is the highest and &F is the lowest. The
;;		priority is the number of consecutive slots allocated by
;;		the scheduler to a process until 16 is reached. A priority
;;		0 will allow 16 slots and priority 15 will allow 1 slot.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;
mmps.ps.pri:
	CALL	krn.ps.getPROC
	JR	NC,tmp_0_e04
	INC	BC
	AND	(BC),0F
	LDA	L
	AEX
	OR	(BC)
	STA	(BC)
tmp_0_e04:
	RET

;;-----------------------------------------------------------------------
;;	STA - Process status
;;
;;	Call:
;;		LD	BC,process-structure
;;		LD	H,process-id
;;		LDA	2B
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Return the process status of the process pointed by H (&FF
;;		for itself) into the process structure pointed by BC. The
;;		information are:
;;		byte 0	Status bits
;;			80 := READY-to-RECEIVE
;;			40 := PAUSED
;;			20 := QUEUE/DUPLEX Waiting
;;			10 := LOCK Waiting
;;			08 := CONSOLE input pending
;;			04 := CONSOLE output pending
;;			02 := TRACED
;;			01 := KERNEL jump pending
;;		byte 1	Priority (4 lower bits)
;;		byte 2	Event mask
;;		byte 3
;;			bits7-4	Extended status bits
;;				02 := TIMER running
;;				01 := DEBUG
;;			bits3-0	Father process-id (4 lower bits)
;;		bytes 4-12	Process name
;;
;;	Return:
;;		CARRY if success; BC is filled with the process-structure.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;
mmps.ps.sta:
	LD	DE,BC
	CALL	krn.ps.getPROC
	JR	NC,tmp_0_e04
	LDI	(BC)
	STI	(DE)
	LDI	(BC)
	AND	F0
	AEX
	STI	(DE)
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	LDI	(BC)
	STI	(DE)
	LDA	06
	ADD	BC
	LDI	(BC)
	AEX
	STI	(DE)
	LDI	(BC)
	STA	H
	LDA	(BC)
	OR	07
	STA	L
	LD	BC,HL
	LD	L,08
	JP	tmp_0_d27

;;-----------------------------------------------------------------------
;;	PAU - Pause the calling process for a delay
;;
;;	Call:
;;		LD	DE,delay
;;		LDA	2C
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Pause for the delay DE the calling process. The process is
;;		suspended until the delay is elapsed. The delay is in
;;		scheduler slots, and its decremented each time the scheduler
;;		is running. When the delay reaches &FFFF, the process is
;;		restarted.
;;
;;	Return:
;;		CARRY.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.pau:
	CALL	tmp_0_bda
	OR	(BC),ps.PAU
	INC	BC
	INC	BC
tmp_0_e34:
	LDA	D
	STI	(BC)
	LDA	E
	STA	(BC)
	SCF
	RET

tmp_0_e3a:
	CALL	tmp_0_bda
krn.ps.evtQUE?:
	LD	B,<tablePROC
	LDA	07
	ADD	BC
	LDD	(BC)
	CPA	FF
	RET

;;-----------------------------------------------------------------------
;;	EVT - Create the event queue for asynchronous events
;;
;;	Call:
;;		LD	L,enabled-event-mask
;;		LDA	2D
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Create the event queue for the calling process to receive
;;		the asynchronous events and set the event mask to L. The
;;		mask is an OR of the events from 0 to 7. It is not possible
;;		to create an event queue if one is already created.
;;		When an asynchronous event is raised to the process, and if
;;		the event is enabled, the event is queued and register the
;;		raising process-id, the event number, and the eventual value
;;		DE from the raising process (events 0-4).
;;		The events 4 to 7 are reserved to the kernel:
;;			4 := Son returned to the scheduler with the value DE
;;			5 := Son terminated
;;			6 := TIME is reached
;;			7 := BREAK
;;		The events 0 to 3 are available for user purpose. The value
;;		DE is meaningful.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C4	Event queue already created
;;
mmps.ps.evt:
	CALL	tmp_0_e3a
	JR	NC,tmp_0_e6e
	PUSH	HL
	PUSH	BC
	LD	L,01
	CALL	mmps.bm.bal
	POP	DE
	POP	HL
	JR	NC,tmp_0_e6d
	LDA	L
	STI	(DE)
	LDA	B
	STI	(DE)
	LDA	C
	STA	(DE)
	LDA	7B
	STI	(BC)
	LDA	(currentPROC)
	STI	(BC)
	CLA
	STI	(BC)
	STI	(BC)
	STA	(BC)
	SCF
tmp_0_e6d:
	RET
tmp_0_e6e:
	LD	H,err.EVT
	RCF
	RET

;;-----------------------------------------------------------------------
;;	KIL - Destroy the events queue
;;
;;	Call:
;;		LDA	2E
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Destroy the events queue of the calling process, and reset
;;		the event-enable-mask to &00. All pending events are lost.
;;
;;	Return:
;;		CARRY.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.kil:
	CALL	tmp_0_e3a
	JR	C,tmp_0_e6d
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	E
	LDA	FF
	STD	(BC)
	STD	(BC)
	STA	(BC)
	LD	BC,DE
	JP	mmps.bm.bfr

;;-----------------------------------------------------------------------
;;	MSK - Set the events mask
;;
;;	Call:
;;		LD	H,disabled-event-mask
;;		LD	L,enabled-event-mask
;;		LDA	2F
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Change the events mask by clearing the events given by H
;;		and setting the events given by L. If H and L are &00, the
;;		mask is unchanged and is useful to get the current event
;;		mask in L.
;;		This action has no effect if the events queue is not
;;		created.
;;
;;	Return:
;;		CARRY and L contains the current event mask.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.msk:
	CALL	tmp_0_e3a
	JR	C,tmp_0_e98
	LDA	(BC)
	XOR	FF
	STA	(BC)
	LDA	H
	OR	(BC)
	XOR	FF
	STA	(BC)
	LDA	L
	OR	(BC)
	STA	(BC)
	STA	L
tmp_0_e98:
	SCF
	RET

;;-----------------------------------------------------------------------
;;	POL - Poll the event queue and get the oldest event
;;
;;	Call:
;;		LDA	30
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Poll the events queue and if an event is pending, return
;;		the raising process-id into H, the event number into L,
;;		and the sent value of DE into DE for the event 0 to 4. If
;;		no event is pending an error is returned.
;;
;;	Return:
;;		CARRY if success; H contains the process-id, L the event
;;		number and DE the sent value for events 0 to 4.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C4	No event queue
;;		D4	Queue empty
;;
mmps.ps.pol:
	CALL	tmp_0_e3a
	JR	C,tmp_0_e6e
	INC	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	STA	L
	PUSH	DE
	PUSH	A
	LD	BC,SP
	INC	BC
	LD	DE,BC
	LD	BC,HL
	CALL	tmp_0_7f5
	POP	A
	POP	DE
	JR	NC,tmp_0_ec2
	STA	L
	AND	F0
	AEX
	STA	H
	LDA	L
	AND	0F
	STA	L
	SCF
tmp_0_ec2:
	RET

;;-----------------------------------------------------------------------
;;	RAI - Raise an event to process
;;
;;	Call:
;;		LD	DE,value
;;		LD	H,process-id
;;		LD	L,event-number
;;		LDA	31
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Raise the event number in L with the value from DE to the
;;		process pointed by H. Only the events from 0 to 3 may be
;;		raised by a process. This action has no effect if the
;;		receiver process has no events queue.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;		C5	No such event number
;;
mmps.ps.rai:
	CALL	krn.ps.getPROC
	JR	NC,tmp_0_ed5
	CP	L,04
	JR	C,tmp_0_ecc
tmp_0_ecc:
	LDA	(currentPROC)
	STA	H
	JR	krn.ps.evtRAI
	LD	H,err.NOV
	RCF
tmp_0_ed5:
	RET
tmp_0_ed6:
	STA	C
krn.ps.evtRAI:
	;; Check if event queue is installed
	CALL	krn.ps.evtQUE?
	JR	C,tmp_0_f13
	LDA	L
	STA	B
	LDA	01
	JR	tmp_0_ee3
tmp_0_ee2:
	SL
tmp_0_ee3:
	DEC	B
	JR	C,tmp_0_ee2
	LD	B,<tablePROC
	AND	(BC)
	JR	Z,tmp_0_f13
	;; Event is enabled in mask so put it into the
	;; event queue
	PUSH	DE
	LDA	H
	RCF
	ADC	L
	PUSH	A
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	E
	LD	BC,SP
	INC	BC
	PUSH	BC
	LD	BC,DE
	POP	DE
	LDA	01
	CP	L,evt.END
	JR	C,tmp_0_f09
	;; Events from 0..4 have significant value into DE
	OR	02
tmp_0_f09:
	STA	L
	CALL	tmp_0_8f8
	POP	A
	POP	MN
tmp_0_f11:
	SCF
	RET
tmp_0_f13:
	CP	L,evt.KIL
	JR	NZ,tmp_0_f11
	;; No event queue and event is KIL, so terminate the current
	;; process. Only this event has a specific treatment. Other
	;; events are discarded if disabled or no event queue.
	LDA	H
	STA	C
	JP	tmp_0_d5d

;;-----------------------------------------------------------------------
;;	TIM - Initialized the TIME event
;;
;;	Call:
;;		LD	DE,delay
;;		LDA	32
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Initialize a TIME event with the delay in DE. The scheduler
;;		will decrement the TIME delay and when &FFFF is reached, it
;;		will push the event 6 to the process. Only one TIME is allowed
;;		by process and it is not possible to initialize a new one
;;		until the first is not reached. This action is without effect
;;		if the calling process does not handled the event 6.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C6	TIME is running
;;
mmps.ps.tim:
	CALL	tmp_0_bda
	LDA	0D
	ADD	BC
	BIT	(BC),ps.TIM
	JR	Z,tmp_0_f2b
	LD	H,err.TIM
	RCF
	RET
tmp_0_f2b:
	OR	(BC),ps.TIM
	LDA	C
	XOR	0F
	STA	C
	JR	tmp_0_e34

;;-----------------------------------------------------------------------
;;	ABT - Abort the TIME delay
;;
;;	Call:
;;		LDA	33
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Abort the TIME delay if one is running.
;;
;;	Return:
;;		CARRY.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.abt:
	CALL	tmp_0_bda
	LDA	0D
	ADD	BC
	AND	(BC),!ps.TIM
	SCF
	RET

;;-----------------------------------------------------------------------
;;	SEN - Send a message to a process
;;
;;	Call:
;;		LD	BC,message-address
;;		LD	H,process-id
;;		LD	L,length
;;		LDA	34
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Send the message pointed by BC with the length L to the
;;		process pointed by H. The destination process should be in
;;		the state READY-TO-RECEIVE. The message will be copied into
;;		the destination process space after the needed number of
;;		pages is allocated. If not enough pages are available to
;;		send the message, an error is returned.
;;		The destination process will restart when a message is
;;		received with the address of the message in BC, its length
;;		in L and the sender process-id in H. The receiver process
;;		need to free the block pointed by BC after the message has
;;		been treated.
;;		It is not possible to send a message to itself.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	No enough page free
;;		C1	No such process
;;		C7	Process not READY-TO-RECEIVE
;;
mmps.ps.sen:
	LDA	H
	AEX
	STA	H
	PUSH	BC
	CALL	tmp_0_d4c
	JR	NC,tmp_0_f4e
	BIT	80
	JR	NZ,tmp_0_f52
	LD	H,err.RTR
tmp_0_f4e:
	POP	BC
	RCF
	RET
tmp_0_f52:
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	E
	PUSH	DE
	LDA	L
	JR	NZ,tmp_0_f60
	INC	A
tmp_0_f60:
	PUSH	A
	DEC	A
	AND	C0
	AEX
	SR
	SR
	INC	A
	STA	L
	LDA	H
	CALL	krn.bm.allocBLOCK_pid_A
	POP	A
	STA	L
	LD	DE,BC
	POP	BC
	JR	NC,tmp_0_f4e
	INC	BC
	LDA	(currentPROC)
	AEX
	STI	(BC)
	LDA	L
	STI	(BC)
	INC	BC
	INC	BC
	LDA	D
	STI	(BC)
	LDA	E
	STA	(BC)
	POP	BC
	JR	tmp_0_f8a
tmp_0_f89:
	LDI
tmp_0_f8a:
	DJC	tmp_0_f89
	LDA	H
	STA	L
	LD	H,<tablePROC
	AND	(HL),7F
	SCF
	RET

;;-----------------------------------------------------------------------
;;	SNR - Send and receive a message
;;
;;	Call:
;;		LD	BC,message-address
;;		LD	H,process-id
;;		LD	L,length
;;		LDA	35
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Like SEN, send a message to another process, and wait for a
;;		reception of a message. When the message is sent, the message
;;		block is released and the process goes into the state
;;		READY-TO-RECEIVE. Like REC, the calling process is restarted
;;		when a message is received, and BC will contain the address
;;		of the message, L its length and H the sender process-id.
;;		The receiver process need to free the block pointed by BC
;;		after the message has been treated.
;;		It is not possible to send a message to itself.
;;
;;	Return:
;;		CARRY if success; BC contains the message address, L its
;;		length and H the sender process-id.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B0	No enough page free
;;		C1	No such process
;;		C7	Process not READY-TO-RECEIVE
;;
mmps.ps.snr:
	PUSH	BC
	CALL	mmps.ps.sen
	POP	BC
	JR	NC,tmp_0_fa6
	CALL	mmps.bm.bfr

;;-----------------------------------------------------------------------
;;	REC - Receive a message from a process
;;
;;	Call:
;;		LDA	36
;;		SBR	E3	;; kernel jump	
;;
;;	Action:
;;		The calling process is suspended until a message is received;
;;		when the process execution restarts, BC will contain the
;;		address of the message, L its length and H the sender
;;		process-id.
;;		The message is copied into the receiver space after the pages
;;		are allocated. If there is not enough page available to send
;;		the message, the receiver process will be not restarted.
;;		The receiver process need to free the block pointed by BC
;;		after the message has been treated.
;;
;;	Return:
;;		CARRY and BC contains the message address, L its length
;;		and H the sender process-id.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.rec:
	CALL	tmp_0_bda
	OR	(BC),80
	SCF
tmp_0_fa6:
	RET

;;-----------------------------------------------------------------------
;;	DBG - Execute a process under debug
;;
;;	Call:
;;		LD	BC,name
;;		LD	D,input-descriptor
;;		LD	E,output-descriptor
;;		LD	HL,arguments
;;		LDA	37
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Start the process named by BC; pass to it the descriptors in
;;		and out given by D and E and the arguments page pointed by HL.
;;		The son process-id is returned to the father into H. The son
;;		process becomes owner of the descriptors D and E. The arguments
;;		page is copied into the son space with a limit of 63 bytes; the
;;		last is always 0. The descriptors in and out given to the son
;;		should be owned by the father and be valid. It is possible to
;;		give only descriptors to CONSOLE, FILE, DUPLEX or QUEUE. The
;;		descriptor D and E may be the same.
;;
;;		Under debug, the son process is stopped. It is fully controlled
;;		by its father, through the kernel jumps TRA, REG, COD and DAT.
;;
;;		Like XEQ, the son process starts with the global address into
;;		BC, the descriptors in and out into D and E and the arguments
;;		page pointer into HL. The arguments page address is copied into
;;		(BC - 2).
;;		The son process gets the same priority as its father, its event
;;		queue is removed and the event mask is set to 0.
;;		A return address is pushed into the stack to handle the son
;;		termination.
;;		When the top-level RET is executed, the process is terminated,
;;		all its resources are released, all its descriptors closed and
;;		the event 4 is raised to its father with the son returned value
;;		of DE.
;;		If the father process is ended, the debugged process is also
;;		ended.
;;		The executable FILE should have the access READ and EXECUTE.
;;		Executing a process will increase by 1 the number of links of
;;		the entry; when the process returns, the number of links will
;;		be decrease by 1.
;;		It is not possible to start a process if the FILE entry is
;;		exclusively opened.
;;		The number of pages allocated to the son while starting are set
;;		into the executable FILE header. See the kernel-jump HDR for
;;
;;	Return:
;;		CARRY if success; H contains the son process-id.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		B1	Not owner of this resource
;;		D1	Exclusively opened
;;		D2	Not a valid DS
;;		F1	No such entry
;;		F4	Access denied
;;		C0	No process free
;;		C2	Not a valid DS for XEQ
;;		C3	Not a valid signature
;;
mmps.ps.dbg:
	CALL	mmps.ps.xeq
	JR	NC,tmp_0_fb9
	LDA	H
	AEX
	STA	C
	LD	B,<tablePROC
	OR	(BC),02
	LDA	0D
	ADD	BC
	OR	(BC),01
tmp_0_fb9:
	RET
tmp_0_fba:
	CALL	krn.ps.getPROC
	JR	NC,tmp_0_fd4
	BIT	(BC),02
	JR	Z,tmp_0_fd1
	LD	DE,BC
	LDA	0D
	ADD	DE
	LDA	(currentPROC)
	OR	01
	CPA	(DE)
	JR	Z,tmp_0_fd4
tmp_0_fd1:
	LD	H,err.TRA
	RCF
tmp_0_fd4:
	RET

;;-----------------------------------------------------------------------
;;	TRA - Trace the next instruction of a debugged process
;;
;;	Call:
;;		LD	H,process-id
;;		LDA	38
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Trace the next instruction of the process pointed by H. This
;;		process should have been launched by DBG. The calling process
;;		is suspended until the debugged process has executed its
;;		instruction.
;;
;;	Return:
;;		CARRY if success.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;		E0	Process not debugged
;;
mmps.ps.tra:
	CALL	tmp_0_fba
	JR	NC,tmp_0_fe1
	AND	(BC),FD
	CALL	tmp_0_bda
	OR	(BC),02
tmp_0_fe1:
	RET
tmp_0_fe2:
	CALL	tmp_0_fba
	JR	NC,tmp_0_ff1
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	E
	LD	BC,DE
tmp_0_ff1:
	RET

;;-----------------------------------------------------------------------
;;	REG - Get registers of a debugged process
;;
;;	Call:
;;		LD	BC,register-structure
;;		LD	H,process-id
;;		LDA	39
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Get the registers into BC of the process pointed by H. This
;;		process should have been launched by DBG.
;;		The register-structure is:
;;			byte 0-1  := SP
;;			byte 2-3  := HL
;;			byte 4-5  := DE
;;			byte 6-7  := BC
;;			byte 8    := A
;;			byte 9-10 := PC
;;			byte 11   := F
;;
;;	Return:
;;		CARRY if success; BC is filled with the registers.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;		E0	Process not debugged
;;
mmps.ps.reg:
	PUSH	BC
	CALL	tmp_0_fe2
	JR	NC,tmp_1_008
	LDA	E
	STD	(BC)
	LDA	D
	STA	(BC)
	LD	L,0C
tmp_0_fff:
	POP	DE
	JR	tmp_1_004
tmp_1_003:
	LDI
tmp_1_004:
	DJC	tmp_1_003
	SCF
	RET
tmp_1_008:
	POP	BC
	RCF
	RET

;;-----------------------------------------------------------------------
;;	COD - Get code from a debugged process
;;
;;	Call:
;;		LD	BC,code-buffer
;;		LD	H,process-id
;;		LD	L,number-of-bytes
;;		LDA	3A
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Get the number of bytes given by L to the code buffer BC of
;;		the process pointed by H, starting at the current PC address
;;		of the debugged process. This process should have been
;;		launched by DBG.
;;
;;	Return:
;;		CARRY if success; BC is filled with the code.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;		E0	Process not debugged
;;
mmps.ps.cod:
	PUSH	BC
	CALL	tmp_0_fe2
	JR	NC,tmp_1_008
	LDA	08
	ADD	DE
	LDI	(DE)
	STA	B
	LDA	(DE)
	STA	C
	JR	tmp_0_fff
tmp_1_01d:
	PUSH	BC
	CALL	mmps.bm.bch
	LDA	C
	STA	E
	POP	BC
	RET

;;-----------------------------------------------------------------------
;;	DAT - Get data from a debugged process
;;
;;	Call:
;;		LD	BC,data-buffer
;;		LD	DE,data-address
;;		LD	H,process-id
;;		LD	L,number-of-bytes
;;		LDA	3B
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Get the number of bytes given by L to the data buffer BC of
;;		the process pointed by H, starting at the address specified
;;		by DE. The data space, from DE to (DE + L) should be located
;;		in the same block and this block should be owned by the
;;		debugged process. This process should have been launched by
;;		DBG.
;;
;;	Return:
;;		CARRY if success; BC is filled with the data.
;;		No CARRY on error, error code into A.
;;
;;	Error Code:
;;		C1	No such process
;;		E0	Process not debugged
;;		E1	Data outside process space
;;
mmps.ps.dat:
	PUSH	BC
	PUSH	DE
	CALL	tmp_0_fba
	POP	BC
	JR	NC,tmp_1_008
	LDA	H
	AEX
	STA	D
	PUSH	BC
	LDA	L
	ADD	BC
	CALL	tmp_1_01d
	POP	BC
	JR	NC,tmp_1_008
	PUSH	A
	CALL	tmp_1_01d
	POP	A
	JR	NC,tmp_1_008
	CPA	E
	JR	Z,tmp_0_fff
	LD	H,err.DAT
	JR	tmp_1_008

;;-----------------------------------------------------------------------
;;	RCH - Reschedule
;;
;;	Call:
;;		LDA	3C
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Reschedule the process activity; Suspend the calling process
;;		and start the next process ready. If no other process may be
;;		started, the calling process is restarted.
;;		When the kernel-jump return, the process is rescheduled.
;;
;;	Return:
;;		CARRY.
;;
;;	Error Code:
;;		None.
mmps.ps.rch:
	CALL	tmp_0_bda
	INC	BC
	OR	(BC),0F
	SCF
	RET



;;=======================================================================
;;	CONSOLE TASK :
;;
;;	The console TASK is in charge of the management of the keyboard
;;	and the screen devices.
;;	It runs when a process calling a console request, INput (REA) or
;;	OUTput (WRI). It has the privilege to run "in place" of the
;;	process with its PID.
;;
;;	In MMPS, the time allocated to the console TASK is the sum of all
;;	the slots of all process requesting on a console. It is not a very
;;	nice implementation choice, and it may be evolute later.
;;
;;	The LCD indicators are set or cleared depending of the pending
;;	requests:
;;	  I	: INput request
;;	  II	: OUTput request
;;
;;	The console TASK handles the following keys:
;;	  SELECT	: Select the next console as active*
;;	  OFF		: Display while the OFF is pressed the PID of the
;;			  process owning the console, the process name
;;			  and the console name as follow
;;			    P NNNNNNNNN CCCCCCCCC
;;	  ON/BREAK	: Send an event KIL to the process owning the
;;			  console, or kill it if the process do not handle
;;			  the event
;;	If an OUTput (II is set) request is pending:
;;	  RCL		: Clear the pending output (II is cleared)
;;	If an INput (I is set) request is pending:
;;	  left-arrow	: Go back one character
;;	  right-arrow	: Go forward one character
;;	  shift-DEL	: Go back to begin of the line
;;	  shift-INS	: Go forward to end of the line
;;	  CL		: Clear from cursor to the end of the line
;;	  shift-CA	: Clear all
;;	  ENTER		: Terminate the INput request, all characters in
;;			  the line are returned to the calling process
;;	  down-arrow	: Terminate the INput request, all characters
;;			  from the begin of the line to the cursor are
;;			  returned to the calling process
;;	  up-arrow	: Return end-of-file error (CARRY is cleared and
;;			  D3 is returned in A
;;
;;	Note*	: The 'SELECT' key is the 'double up-down arrow' key on the
;;		  left of the RCL key.
;;
tmp_1_059:
	LDI	(HL)
	STA	B
	LDA	(HL)
	STA	C
	LDA	07
	ADD	BC
	LD	L,08
tmp_1_063:
	LDI	(BC)
	JR	Z,tmp_1_069
	STI	(DE)
	DJC	tmp_1_063
tmp_1_069:
	LDA	20
	INC	L
tmp_1_06c:
	STI	(DE)
	DJC	tmp_1_06c
	RET
	;; When pressing OFF, the PID, name of PROCESS and name of CONSOLE
	;; are displayed:
	;;	"P NNNNNNNNN CCCCCCCCC "
tmp_1_070:
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LDA	(currentCON)
	STA	C
	LD	B,<tableDS
	LDR	DE,kernelBUF
	INC	BC
	LDI	(BC)
	PUSH	BC
	STA	L
	LD	H,<tablePROC
	AEX
	OR	$0
	STI	(DE)
	LDA	20
	STI	(DE)
	LDA	0E
	ADD	HL
	CALL	tmp_1_059
	POP	HL
	CALL	tmp_1_059
	CLS
	LDR	HL,kernelBUF
	LD	C,16	; 9+9+1 + 3 spaces
	CALL	NCHARDISP
tmp_1_0a4:
	EI
	CALL	INKEY
	DI
	CPA	conKEY.OFF	
	JR	Z,tmp_1_0a4
	POP	HL
	POP	DE
	POP	BC

	;; The console TASK entry point
	;; This task "take" the slot of any process
	;; pending on a console kernel-jump. The
	;; task manage the keyboard (REA) and the
	;; screen (WRI).
krn.conTASK:
	DI
	BIT	(currentCON),FF
	JR	Z,tmp_1_0c8
	CALL	tmp_1_287
tmp_1_0c0:

	;; Close the current active console.
	;; When a console is closed and is selected
	;; as active, Closed! is displayed.
krn.conClosed:
	DI
	BIT	(currentCON),FF
	JR	NZ,tmp_1_0d4
tmp_1_0c8:
	CLS
	LDR	HL,mmps.CLOSED
	LD	C,07
	CALL	NCHARDISP
	JR	tmp_1_0ef
tmp_1_0d4:
	BIT	(HIGHLCDFLAG),conLCD.POUT
	JR	NZ,tmp_1_0ef
	PUSH	DE
	LDA	D
	STA	H
	LDA	E
	STA	(787B)
	AND	E0
	STA	L
	LD	C,1A
	CALL	NCHARDISP
	CALL	tmp_1_2c1
	POP	DE
tmp_1_0ef:
	EI
	OR	(7B0E),40
	CALL	GETKEY
	DI
	CPA	conKEY.SLC	; Select another console as active
	JR	NZ,tmp_1_101
	JP	tmp_1_253
tmp_1_101:
	BIT	(currentCON),FF
	JR	Z,tmp_1_0c0
	CPA	conKEY.OFF	; Display current procces and current console
	JR	Z,tmp_1_070
	BIT	(HIGHLCDFLAG),conLCD.POUT
	JR	Z,tmp_1_156
	CPA	conKEY.RCL	; Flush the pending OUTPUT if any
	JR	NZ,tmp_1_156
	LDA	D
	STA	B
	LDA	E
	AND	C0
	OR	3A
	STA	C
	AND	(BC),00
	INC	BC
	LDD	(BC)
	JR	Z,tmp_1_143
	STA	L
	LD	DE,BC
	STI	(DE)
	CLA
	STI	(DE)
	LDI	(DE)
	STA	B
	LDI	(DE)
	STA	C
	LDA	(DE)
	STA	H
	LDA	E
	AND	E0
	STA	E
	JR	tmp_1_136
tmp_1_135:
	LDI
tmp_1_136:
	DJC	tmp_1_135
	LDA	H
	STA	C
	LD	B,<tablePROC
	AND	(BC),FB
	CALL	tmp_1_287
	JR	tmp_1_0c0
tmp_1_143:
	LDA	(currentCON)
	STA	L
	LD	H,<tableDS
	INC	HL
	LDA	(HL)
	STA	C
	LD	B,<tablePROC
	AND	(BC),FB
	AND	(HIGHLCDFLAG),DF	; !conLCD.POUT
tmp_1_154:
	JR	tmp_1_0c0
tmp_1_156:
	BIT	(HIGHLCDFLAG),conLCD.PIN
	JR	Z,tmp_1_0c0
	STA	B
	LDA	E
	AND	C0
	OR	1B
	STA	L
	LDA	D
	STA	H
	LDA	01
	CPA	(HL)
	JR	Z,tmp_1_171
	LDA	E
	AND	1F
	INC	A
	CPA	(HL)
	JR	NC,tmp_1_1cc
tmp_1_171:
	LDA	B
	STI	(DE)
tmp_1_173:
	PUSH	DE
	PUSH	A
	LD	BC,DE
	CLA
	STI	(HL)
	LDI	(HL)
	STA	D
	LDI	(HL)
	STA	E
	LDA	(HL)
	STA	H
	POP	A
	PUSH	A
	CPA	FF
	JR	Z,tmp_1_198
	LDA	C
	AND	1F
	STA	L
	LDA	C
	AND	C0
	STA	C
	LDA	L
	JR	tmp_1_195
tmp_1_194:
	LDI
tmp_1_195:
	DJC	tmp_1_194
	STA	L
tmp_1_198:
	CLA
	STA	(DE)
	LDA	H
	STA	C
	LD	B,<tablePROC
	AND	(BC),!ps.OUT
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	LDI	(BC)
	STA	D
	LDA	(BC)
	STA	E
	LD	BC,DE
	POP	A
	POP	DE
	INC	BC
	CPA	FF
	JR	NZ,tmp_1_1c0
	LDA	06
	ADD	BC
	LDA	D3
	STI	(BC)
	INC	BC
	INC	BC
	AND	(BC),1E
	JR	tmp_1_1c3
tmp_1_1c0:
	INC	BC
	LDA	L
	STA	(BC)
tmp_1_1c3:
	AND	(HIGHLCDFLAG),0F	; !(conLCD.PIN|conLCD.POUT|conLCD.PCUR)
	CALL	tmp_1_287
tmp_1_1ca:
	JR	tmp_1_154
tmp_1_1cc:
	LDA	B
	CPA	conKEY.RGT	; Goto next character right
	JR	NZ,tmp_1_1df
	BIT	(DE),E0
	JR	Z,tmp_1_1e9
	INC	E
tmp_1_1d6:
	LDA	E
	AND	1F
	CPA	1A
	JR	NC,tmp_1_1e9
	JR	tmp_1_1e8
tmp_1_1df:
	CPA	conKEY.LFT	; Goto next character left
	JR	NZ,tmp_1_1f0
	LDA	E
	AND	3F
	JR	Z,tmp_1_1e9
tmp_1_1e8:
	DEC	E
tmp_1_1e9:
	DEC	HL
	LDA	E
	AND	1F
	STA	(HL)
tmp_1_1ee:
	JR	tmp_1_1ca
tmp_1_1f0:
	CPA	conKEY.DEL	; Goto begin-of-line
	JR	NZ,tmp_1_1fa
	LDA	E
	AND	C0
	STA	E
	JR	tmp_1_1e9
tmp_1_1fa:
	CPA	conKEY.INS	; Goto end-of-line
	JR	NZ,tmp_1_215
	CALL	tmp_1_203
	JR	tmp_1_1d6
tmp_1_203:
	LDA	E
	AND	C0
	OR	19
	STA	E
	LDA	19
tmp_1_20b:
	BIT	(DE),E0
	JR	NZ,tmp_1_213
	DEC	DE
	DEC	A
	JR	C,tmp_1_20b
tmp_1_213:
	INC	DE
	RET
tmp_1_215:
	CPA	conKEY.CLR	; Clear from cursor to end-of-line
	JR	NZ,tmp_1_22b
tmp_1_219:
	LD	BC,DE
	LDA	C
	AND	1F
	STA	L
	SCF
	LDA	19
	SBC	L
	STA	L
	LDA	0D
tmp_1_226:
	STI	(BC)
	DJC	tmp_1_226
	JR	tmp_1_1e9
tmp_1_22b:
	CPA	conKEY.CA	; Clear all line
	JR	NZ,tmp_1_235
	LDA	E
	AND	C0
	STA	E
	JR	tmp_1_219
tmp_1_235:
	CPA	conKEY.ENTER	; Finish entry input
	JR	NZ,tmp_1_240
	CALL	tmp_1_203
	LDA	0D
	JR	tmp_1_173
tmp_1_240:
	CPA	conKEY.DWN	; Finish entry input at current position
	JR	Z,tmp_1_173
	CPA	conKEY.UP	; Return end-of-file
	JR	NZ,tmp_1_24c
	LDA	FF
	JR	tmp_1_173
tmp_1_24c:
	CPA	20
	JR	NC,tmp_1_1ee
	STI	(DE)
	JR	tmp_1_1d6
tmp_1_253:
	LDA	(currentCON)
	JR	NZ,tmp_1_25a
	LDA	>tableDS
tmp_1_25a:
	STA	C
	LD	B,<tableDS
	LD	L,maxDS
tmp_1_25f:
	RCF
	LDA	06
	ADC	C
	CPA	>maxIndexDS
	JR	NZ,tmp_1_269
	LDA	>tableDS
tmp_1_269:
	STA	C
	LDA	(BC)
	BIT	10
	JR	NZ,tmp_1_273
	DJC	tmp_1_25f
	JR	tmp_1_1ee
tmp_1_273:
	LDA	C
	STA	(currentCON)
	INC	BC
	INC	BC
	LDI	(BC)
	STA	H
	LDA	(BC)
	INC	A
	INC	A
	STA	L
	LDI	(HL)
	STA	D
	LDA	(HL)
	CALL	tmp_1_28a
	JR	tmp_1_1ee
tmp_1_287:
	LDA	E
	AND	C0
tmp_1_28a:
	OR	1A
	STA	E
	AND	(HIGHLCDFLAG),9F	; !(conLCD.PIN|conLCD.POUT)
	CLS
	CALL	ECAE
	AND	(787C),00
	AND	(7880),00
	INC	DE
	LDD	(DE)
	JR	Z,tmp_1_2a5
	OR	(HIGHLCDFLAG),conLCD.PIN
tmp_1_2a5:
	LDA	E
	OR	20
	STA	L
	LDA	E
	AND	C0
	RCF
	ADC	(DE)
	STA	E
	LDA	D
	STA	H
	LDA	(HL)
	JR	Z,tmp_1_2c0
	STA	C
	OR	(HIGHLCDFLAG),conLCD.POUT
	LDA	L
	AND	E0
	STA	L
	CALL	NCHARDISP
tmp_1_2c0:
	RET
tmp_1_2c1:
	PUSH	DE
	PUSH	A
	JP	E9FF
tmp_1_2c8:
	LDR	BC,conTASKSTACK
	LDA	03
	STD	(BC)
	LDA	(2+mmps.krn.conTASK)
	STD	(BC)
	LDA	(1+mmps.krn.conTASK)
	STD	(BC)
	LDA	E
	STD	(BC)
	STD	(BC)
	STD	(BC)
	STD	(BC)
	LDA	D
	STD	(BC)
	LDR	BC,conPROC.SP
	LDA	<conTASKSP
	STI	(BC)
	LDA	>conTASKSP
	STA	(BC)
	AND	(HIGHLCDFLAG),!conLCD.PCUR
	RET

;;=======================================================================
;;	SCHEDULER :
;;
;;	The MMPS scheduler will be triggered at each interruption, when a
;;	process call the RCH kernel-jump, or at each kernel-jump entry.
;;
;;	The scheduler is in charge to decide what process is ready to run
;;	and so to give the hand to it, or what process need to sleep
;;	according its state.
;;
;;	To run a process should be the next "eligible" process, that is
;;	the first process ready in the scheduler queue. The following
;;	status bits set the process sleeping until the "sleeping" event
;;	is completed:
;;	  READY-TO-RECEIVE : The process is waiting for a message (REC or
;;			     SNR); It will rescheduled when a message is
;;			     received from another process (SEN or SNR)
;;	  PAUSED	   : The process is paused for a delay (PAU); It
;;			     will be rescheduled when the delay has
;;			     elapsed
;;	  QUEUE-WAITING	   : Waiting for a message in a queue (WAI); It
;;			     will be rescheduled when a message is written
;;			     in the queue (WRI)
;;	  LOCK-WAITING	   : Waiting for owning a lock (TAK); It will be
;;			     rescheduled when the lock becomes given (GIV)
;;			     is taken by the process
;;	  TRACE		   : Waiting for trace control (DBG); it will be
;;			     rescheduled when the debugger leader will
;;			     request the execution of the next instruction
;;			     (TRA)
;;
;;	Each time the scheduler runs a process, it looses one slot. When
;;	all slots are consumed, the process is put at the tail of the
;;	scheduler queue. Under MMPS, a kernel-jump consumes one slot. The
;;	number of slots allocated to a process may be changed by the
;;	kernel-jump PRI.
;;	The RCH kernel-jump put directly the calling process to the tail
;;	of the scheduler queue.
;;

krn.ps.decTIMER:
	INC	DE
	LDA	(DE)
	DEC	A
	STD	(DE)
	JR	C,tmp_1_2f6
	LDA	(DE)
	DEC	A
	STA	(DE)
tmp_1_2f6:
	RET

	;; Entry point for all kernel-jumps
	;; called by a process: SBR (&E3)
enterkrnjump:
	;; Disable the interruptions
	DI
	PUSH	BC
	PUSH	A
	;; Retrieve our process structure
	CALL	tmp_0_bda
	SCF
	POP	A

	;; Disable the interruptions
	DI
	;; Kernel-jump bit set. when scheduled
	;; we will re-enter into the kernel
	;; to execute the kernel-jump inside the
	;; kernel space
	OR	(BC),ps.KRN
	POP	BC

	;; Disable the interruptions
	DI

	;; Save all registers. Because the CPU
	;; register from a proces should be safe
	;; when executing a kernel-jump
	PUSH	A
	PUSH	A
	PUSH	BC
	LD	BC,SP
	PUSH	DE
	PUSH	HL

	;; Set the interrupt-enable flag for the
	;; before "pushing" the status F into our
	;; stack
	LDA	F
	OR	cpuF.EI
	INC	BC
	INC	BC
	INC	BC
	INC	BC
	LD	DE,BC
	INC	BC
	LDI
	LDI
	STA	(DE)
	JR	tmp_1_34c

	;; End-process entry point
	;; A process doing a RET from top of its
	;; stack will jump here. This address is
	;; pushed onto the stack when XEQ builds
	;; the process stack.
krn.endproc:
	;; Disable the interruptions
	DI
	CALL	tmp_0_bda
	LD	L,evt.PID
	CALL	krn.ps.endME
	JR	tmp_1_34c



;;=======================================================================
;;	INTERRUPTION HANDLER - IT
;;
;;	We set our own handler by setting the IT handler address into the
;;	&79DB/&79DC address and enable it with a &55 in &79DA.
;;
;;	When an external interrupt is received, the BASIC ROM branch here
;;	with the A and HL pushed more onto the stack. Of course, all the
;;	CPU registers are safely pushed into the stack by the BASIC ROM
;;	internal handler.
;;
;;	:-D Thank-you SHARP to have implemented this facility !!!
;;

	;; Debugger interrupt entry point
	;; The BASIC branch here when an external
	;; interrupt is raised and a user-handler
	;; is installed.
	;; Great thing from SHARP !!!
	;; Specific handler for a process being
	;; traced. Because it is executed step-by-
	;; step. Father process need to be awaken
	;; when son debugged process a complete
	;; its instruction
krn.IT.debugger:
	;; Disable the interruptions
	DI

	;; Get our process-structure and set the
	;; debug bit. This will prevent the scheduler
	;; to run it.
	CALL	tmp_0_bda
	OR	(BC),ps.TRA
	LDA	0D
	ADD	BC
	LDA	(BC)
	AND	F0
	;; Get our father pid and clear the trace bit
	STA	C
	AND	(BC),!ps.TRA

	;; Scheduler interrupt entry point
	;; The BASIC branch here when an external
	;; interrupt is raised and a user-handler
	;; is installed.
	;; Great thing from SHARP !!!
krn.IT.scheduler:
	;; Disable the interruptions
	DI
	;; BASIC branch here but, HL and A are
	;; pushed into the stack. But all CPU
	;; registers are saved by the BASIC itself
	POP	HL
	POP	A

	;; Clear the interrupt source and
	;; acknowledge the interruption
	SBR	(A0)
tmp_1_34c:
	AND#	(F00A),00
	AND#	(F00B),02

	;; Current effective running process
	LDA	(effectivePROC)
	STA	E
	LD	D,<tablePROC

	;; Register SP into the process-structure
	LD	BC,SP
	INC	DE
	INC	DE
	INC	DE
	INC	DE
	LDA	B
	STI	(DE)
	LDA	C
	STA	(DE)

krn.schedule:
	LD	SP,kernelSTACK

	;; Check if ON/BREAK pressed
	SBR	(A6)
	JR	Z,tmp_1_383

	;; Yes. If a CONSOLE is active and has
 	;; a process is attached, generate the
	;; event 7 or kill the process !
	AND#	(F00B),FD

	LDA	(currentCON)
	JR	Z,tmp_1_383

	;; A console is active, look for the
	;; process attached on it
	STA	E
	LD	D,<tableDS
	INC	DE
	LDA	(DE)
	STA	C
	STA	H
	;; Send a KIL event to the process
	LD	L,evt.KIL
	CALL	krn.ps.evtRAI
tmp_1_383:
	LDR	BC,tablePROC
	LD	L,07
tmp_1_389:
	LDA	(BC)

	;; No process
	CPA	ps.FRE
	JR	Z,tmp_1_3c2
	LD	DE,BC
	INC	DE
	INC	DE

	;; Process is paused until a delay
	BIT	(BC),ps.PAU
	JR	Z,tmp_1_39d
	CALL	krn.ps.decTIMER
	JR	C,tmp_1_39d
	AND	(BC),!ps.PAU
tmp_1_39d:
	LDA	0D
	ADD	BC
	BIT	(BC),ps.TIM
	JR	Z,tmp_1_3c2
	LDA	07
	ADD	DE
	CALL	krn.ps.decTIMER
	JR	C,tmp_1_3c2
	AND	(BC),!ps.TIM
	PUSH	BC
	PUSH	HL
	LDA	C
	AND	F0
	STA	C
	STA	H
	;; Send a TIM event to the process
	LD	L,evt.TIM
	CALL	krn.ps.evtRAI
	POP	HL
	POP	BC
tmp_1_3c2:
	LDA	10
	ADD	BC
	LDA	C
	AND	F0
	STA	C
	DJC	tmp_1_389
	LDA	(currentPROC)

krn.ps.SCHEDULE:
	STA	C
	LD	L,08

	;; Schedule the process if the process status is not one
	;;   ps.RTR - READY-TO-RECEIVE : waiting for a message
	;;   ps.PAU - Paused up to delay elapses
	;;   ps.QUE - Waiting message from queue
	;;   ps.LOK - Waiting until taking the lock
	;;   ps.DBG - Under debugger session
	BIT	(BC),ps.RDY
	JR	NZ,tmp_1_3eb

	;; Consume one time slot. When all slots are consummed
	;; the process in linked to the tail of the scheduler
	;; queue. If no proces are ready, so it it is scheduled
	;; again...
	INC	BC
	LDA	(BC)
	STA	H
	INC	A
	STA	(BC)
	JR	NH,krn.ps.REscheduleME
	LDA	H
	AND	F0
	STA	H
	AEX
	RCF
	ADC	H
	STD	(BC)
	JR	tmp_1_3eb

	;; Scan the scheduler queue to fetch the next process to
	;; be run
tmp_1_3e7:
	BIT	(BC),ps.RDY
	JR	Z,krn.ps.scheduleME
tmp_1_3eb:
	LDA	C
	RCF
	ADC	10
	AND	70
	STA	C
	DJC	tmp_1_3e7
	LD	L,07
	LD	C,00
tmp_1_3f8:
	LDA	(BC)
	CPA	FF
	JR	NZ,tmp_1_40f
	LDA	10
	ADD	BC
	DJC	tmp_1_3f8

	;; No process to schedule... Why to go DWN ?
	;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	;; cgh:	Better to implemant an IDLE here, with the interrupts
	;;	enabled, scheduler will be entered at the next interrupt
	;;	caught
	;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	JP	mmps.ps.dwn

krn.ps.REscheduleME:
	DEC	BC
krn.ps.scheduleME:
	LDA	C
	STA	(currentPROC)

	;; INput or OUTput pending on this process, so
	;; the console TASK take the hand with the PID
	;; of the process
	BIT	(BC),0C	; (ps.IN|ps.OUT)
	JR	Z,tmp_1_412

tmp_1_40f:
	LDA	conPID
	STA	C
tmp_1_412:
	STA	(effectivePROC)
	LD	DE,BC
	INC	DE
	INC	DE
	INC	DE
	INC	DE
	LDI	(DE)
	STA	B
	LDD	(DE)
	STA	C
	DEC	DE
	DEC	DE
	DEC	DE
	DEC	DE

	;; Restore process stack SP
	LD	SP,BC
	;; Check for kernel-jump
	BIT	(DE),01
	JR	Z,tmp_1_46e

	;; Process is requesting a kernel-jump
	;; So clear the kernel-jump bit
	AND	(DE),FE

	;; Retrieve the kernel-jump code from A
	POP	HL
	POP	DE
	POP	BC
	POP	A
	PUSH	A
	PUSH	BC
	PUSH	DE
	PUSH	HL

	;; Is kernel-jump code valid ?
	CPA	(1+mmps.krn.jumpMAX)
	JR	C,tmp_1_4b2

	;; Because of a stupid(!) operation done, I
	;; split the kernel-jump table in 2 parts:
	;; 00-1f, and 20-3d
	PUSH	BC

	;; kernel-jump-code >= 32, use the second
	;; table
	CPA	32
	JR	C,krn.vect20_3D

	;; Address in table: (A*5) + mmps.krn.vect00_1F
	STA	C
	SL
	SL
	RCF
	ADC	C
	ADC	(2+mmps.krn.vect00_1F)
	;; Restore BC
	POP	BC
	;; Push the low-byte kernel-jump-address
	PUSH	A
	LDA	(1+mmps.krn.vect00_1F)
	ADC	N
	;; Push the high-byte kernel-jump-address
	PUSH	A
	;; Execute kernel-jump
	RET

krn.vect20_3D:
	SBC	32

	;; Address in table: (A*5) + mmps.krn.vect00_1F
	STA	C
	SL
	SL
	RCF
	ADC	C
	ADC	(2+mmps.krn.vect20_3D)
	;; Restore BC
	POP	BC
	;; Push the low-byte kernel-jump-address
	PUSH	A
	LDA	(1+mmps.krn.vect20_3D)
	ADC	N
	;; Address in table: (A*5) + mmps.krn.vect00_1F
	PUSH	A
	;; Execute kernel-jump
	RET

tmp_1_46e:
	LD	C,>1+mmps.krn.IT.scheduler
	LD	L,17
	CP	E,>conPROC
	JR	Z,tmp_1_482
	LDA	0D
	ADD	DE
	BIT	(DE),01
	JR	Z,tmp_1_482
	LD	L,10
	LD	C,>1+mmps.krn.IT.debugger

tmp_1_482:
	LD	B,<1+mmps.krn.IT.scheduler

	;; Prepare the interruption.
	LDR	DE,F007
	CLA
	STA#	(DE)
	INC	DE
	AND#	(DE),BF
	INC	DE
	LDA	L
	STA#	(DE)
	INC	DE
	LDA	0B
	STA#	(DE)
	INC	DE
	AND#	(DE),FE

	;; Install the external interrupt vector
	;; to BASIC and activate it
	LDR	DE,&79DA
	LDA	55
	STI	(DE)
	LDI
	LDI

	;; Restore the registers of the process
	;; to schedule
	POP	HL
	POP	DE
	POP	BC
	POP	A
	;; Arm the interrupt
	STA#	(F006)

	;; Return to scheduled process. The RTI
	;; return to caller, but it restore the
	;; status F register from the stack.
	;; We set the IE flag.
	RTI
tmp_1_4b2:

nokrnjump:
	LD	H,err.KRN
	RCF

	;; Exit from a kernel-jump
	;; If CARRY is cleared, the error code
	;; is retruned into A, and CARRY is cleared
	;; in F. No other CPU register is modified.
	;; If CARRY is set, kernel-jmup complete
	;; and some values may be returned into
	;; CPU registers

	;; No values returned
outkrnjump.NONE:
	;; kernel-jump complete :-D
	JR	C,outkrnjump

	;; No. A will contains the error code
	;; and we clear the CARRY
	LD	BC,SP
	LDA	07
	ADD	BC
	LDA	H
	STI	(BC)
	INC	BC
	INC	BC
	AND	(BC),1E	;; Be sure to clear the CARRY
tmp_1_4c3:

outkrnjump:
	;; To exit, just schedule the next slot or next
	;; process...
	JP	krn.schedule

	;; Values returned into BC
outkrnjump.BC:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	POP	HL
	POP	DE
	POP	A
	POP	A

outkrnjump.PUSH.BCDEHL:
	PUSH	BC

outkrnjump.PUSH.DEHL:
	PUSH	DE

outkrnjump.PUSH.HL:
	PUSH	HL
	JR	tmp_1_4c3

	;; Values returned to E
outkrnjump.E:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	POP	HL
	LDA	E
	POP	DE
	STA	E
	JR	outkrnjump.PUSH.DEHL

	;; Values returned to D
outkrnjump.D:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	POP	HL
	LDA	D
	POP	DE
	STA	D
	JR	outkrnjump.PUSH.DEHL

	;; Values returned to L
outkrnjump.L:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	LDA	L
	POP	HL
	STA	L
	JR	outkrnjump.PUSH.HL

	;; Values returned to H
outkrnjump.H:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	LDA	H
	POP	HL
	STA	H
	JR	outkrnjump.PUSH.HL

	;; Values returned to E
outkrnjump.HL:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	POP	BC
	JR	outkrnjump.PUSH.HL

	;; Values returned to DE and HL
outkrnjump.DEHL:
	;; kernel-jump failed :-(
	JR	NC,outkrnjump.NONE
	POP	BC
	POP	BC
	JR	outkrnjump.PUSH.DEHL


	;; Kernel-jump matrix: Call the internal
	;; kernel-jump, and go to return routine
	;; depending on the values expected to
	;; sent back to the calling process.

	;; **************************************
	;; ALL CALLS HERE SHOULD BE DONE WITH THE
	;; INTERRUPTIONS DISABLED, because the
	;; kernel does not provide an internal
	;; lock mechanism, and these routines
	;; touch the internal structure of the
	;; kernel.
	;; **************************************
	;; NEVER CALL THEM FROM OUTSIDE THE KERNEL
	;; **************************************

krn.vect00_1F.base:
	CALL	mmps.bm.bal
outkrnjump._BC:		;; For far kernel-jump
	JR	outkrnjump.BC
	CALL	mmps.bm.bch
	JR	outkrnjump.NONE
	CALL	mmps.bm.bfr
	JR	outkrnjump.NONE
	CALL	mmps.bm.bre
	JR	outkrnjump.NONE
	CALL	mmps.fs.fsr
	JR	outkrnjump.D
	CALL	mmps.fs.fen
	JR	outkrnjump.E
	CALL	mmps.fs.tmp
	JR	outkrnjump.NONE
	CALL	mmps.fs.lok
	JR	outkrnjump.E
	CALL	mmps.fs.con
	JR	outkrnjump.E
	CALL	mmps.fs.fil
	JR	outkrnjump.E
	CALL	mmps.fs.que
	JR	outkrnjump.E
	CALL	mmps.fs.dpx
	JR	outkrnjump.E
	CALL	mmps.fs.shr
	JR	outkrnjump.E
	CALL	mmps.fs.map
	JR	outkrnjump.E
	CALL	mmps.fs.del
	JR	outkrnjump.NONE
	CALL	mmps.fs.acs
	JR	outkrnjump.NONE
	CALL	mmps.fs.rnm
	JR	outkrnjump.NONE
	CALL	mmps.fs.siz
	JR	outkrnjump.NONE
	CALL	mmps.fs.dsn
	JR	outkrnjump.NONE
	CALL	mmps.fs.des
	JR	outkrnjump.H
	CALL	mmps.fs.opn
;; >>> Patch cgh - Open should return E #0005 >>>>>>>>>>>>>>>>>>>>>>>>>>>
;;	JR	outkrnjump.NONE
	JR	outkrnjump.E
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	CALL	mmps.fs.clo
outkrnjump._NONE:	;; For far kernel-jump
	JR	outkrnjump.NONE
	CALL	mmps.fs.rea
	JR	outkrnjump.L
	CALL	mmps.fs.wri
outkrnjump._L:		;; For far kernel-jump
	JR	outkrnjump.L
	CALL	mmps.fs.tel
	JR	outkrnjump.HL
	CALL	mmps.fs.sek
outkrnjump.__NONE:	;; For far kernel-jump
	JR	outkrnjump.NONE
	CALL	mmps.fs.wai
	JR	outkrnjump.NONE
	CALL	mmps.fs.cnt
	JR	outkrnjump.NONE
	CALL	mmps.fs.get
	JR	outkrnjump._NONE
	CALL	mmps.fs.put
	JR	outkrnjump._NONE
	CALL	mmps.fs.giv
	JR	outkrnjump._NONE
	CALL	mmps.fs.tak
	JR	outkrnjump._NONE
	CALL	mmps.fs.scn
	JR	outkrnjump._NONE
	CALL	mmps.fs.flu
	JR	outkrnjump._NONE
	CALL	mmps.fs.rfr
	JR	outkrnjump._NONE
	CALL	mmps.fs.edi
	JR	outkrnjump.L
	CALL	mmps.fs.hdr
	JR	outkrnjump._NONE
	CALL	mmps.ps.gbl
	JR	outkrnjump._BC
	CALL	mmps.ps.pid
	JR	outkrnjump.H
	CALL	mmps.fs.fid
	JR	outkrnjump.H
	CALL	mmps.ps.xeq
outkrnjump._H:		;; For far kernel-jump
	JR	outkrnjump.H
	CALL	mmps.fs.end
	JR	outkrnjump.__NONE
	CALL	mmps.ps.pri
	JR	outkrnjump.__NONE
	CALL	mmps.ps.sta
	JR	outkrnjump.__NONE
	CALL	mmps.ps.pau
	JR	outkrnjump.__NONE
	CALL	mmps.ps.evt
	JR	outkrnjump.__NONE
	CALL	mmps.ps.kil
	JR	outkrnjump.__NONE
	CALL	mmps.ps.msk
	JR	outkrnjump._L
	CALL	mmps.ps.pol
	JR	outkrnjump.DEHL
	CALL	mmps.ps.rai
	JR	outkrnjump.__NONE

krn.vect20_3D.base:
	CALL	mmps.ps.tim
	JR	outkrnjump.__NONE
	CALL	mmps.ps.abt
	JR	outkrnjump.__NONE
	CALL	mmps.ps.sen
	JR	outkrnjump.__NONE
	CALL	mmps.ps.snr
	JR	outkrnjump.__NONE
	CALL	mmps.ps.rec
	JR	outkrnjump.__NONE
	CALL	mmps.ps.dbg
	JR	outkrnjump._H
	CALL	mmps.ps.tra
	JR	outkrnjump.__NONE
	CALL	mmps.ps.reg
	JR	outkrnjump.__NONE
	CALL	mmps.ps.cod
	JR	outkrnjump.__NONE
	CALL	mmps.ps.dat
	JR	outkrnjump.__NONE
	CALL	mmps.ps.rch
	JR	outkrnjump.__NONE

;;-----------------------------------------------------------------------
;;	DWN - Down the system and return to BASIC
;;
;;	Call:
;;		LDA	3D
;;		SBR	E3	;; kernel jump
;;
;;	Action:
;;		Down the system: end all the running processes without warning
;;		the fathers, close all the descriptors, release all the
;;		resources and return to BASIC. This is the only way to exit
;;		properly from the system. No privilege are asked to execute
;;		it. The system is always calling DWN before starting. This
;;		kernel jump should never return.
;;
;;	Return:
;;		DWN never returns.
;;
;;	Error Code:
;;		None.
;;
mmps.ps.dwn:
	LD	SP,kernelSTACK

mmps.ps.dwn_noSP:
	LDR	BC,tablePROC
tmp_1_642:
	LDA	(BC)
	CPA	ps.FRE
	JR	Z,tmp_1_657
	PUSH	BC
	OR	(BC),ps.FRE
	LDA	C
	STA	(endingPROC)
	OR	0E
	STA	C
	CALL	tmp_0_d7b
	POP	BC
tmp_1_657:
	LDA	10
	ADD	BC
	CP	C,>maxIndexPROC
	JR	NC,tmp_1_642
	OR	(endingPROC),99
	CALL	tmp_0_db0
	LDR	BC,conPROC.SP
	CLA
	STI	(BC)
	STI	(BC)
	STI	(BC)
	STI	(BC)
	STI	(BC)
	RET



;;=======================================================================
;;	MMPS - Mini Multi-Process System
;;
;;	When the system loaded into the memory, it may be called directly
;;	by the BASIC instruction CALL &C5 under the editor or a program.
;;
;;	The file-system and the volatile data are kept when exiting MMPS.
;;	To start MMPS again, just do CALL &C5 under the BASIC editor, or
;;	in a program or in a LM routine.
;;
;;	When starting, the system will first perform a 'DOWN'; it means that
;;	it will terminate all the processes, close all the descriptors and
;;	release all the blocks and pages not reserved for the file-system.
;;	After it will create a CONSOLE entry named 'CONSOLE', put it as the
;;	active one and try to start the process 'STARTER' on this console.
;;	If one of these actions failed, the kernel will provoke a computer
;;	soft reset and give the hand back to the BASIC. The message
;;		'NEW0? :CHECK'
;;	will be displayed. In this case, a trace-back is written in binary
;;	into the variable BASIC E$ (at address &7650). The binary data
;;	should be retrieved using the PEEK function or under a monitor. The
;;	E$ content is the following:
;;		FF AA BB CC DD EE HH LL PCPC SPSP
;;	
;;	If all has worked, STARTER will display the MMPS banner (see
;;	'MMPS 09Apr.1998') and be waiting for a command. To clear a message
;;	pending on a CONSOLE, use the 'RCL' key.
;;	STARTER assumes that all the commands given are the name of a
;;	process to start followed by its arguments. If the process is
;;	started, STARTER will display '>N' where N is PID of the process.
;;	If the process can not be launched correctly, STARTER will display
;;	'!EE' where EE is error code returned by MMPS.
;;	When a new process starts, STARTER activates and selects the
;;	console to it. To change the active console, use the 'DOUBLE-ARROW'
;;	key located on the left of 'RCL'.
;;	When a process is normally finished, STARTER catches the event and
;;	displays '>N +4:RRRR' with N the PID of the ended process and RRRR
;;	the returned value.
;;	When a process is killed, STARTER catches the event and display
;;	'>N -5' with N the PID of the ended process.
;;	Another STARTER process may be started. It is finished by pressing
;;	the 'UP-ARROW' key. In this case, the process just returns to the
;;	scheduler except if STARTER has the PID 0 (that means that it is
;;	the son of the kernel) and in this case it will down the system.
;;	When the system is 'DOWN'ing, it will terminated all the processes,
;;	close all the descriptors and release all the pages. After, it will
;;	restore the BASIC stack and give it back the control. It is not
;;	possible to start more than one process neither to start another
;;	process STARTER at the 'BOOT' time (except by writing another
;;	process and give to it the name 'STARTER').
;;

	;; Start entry point for the MMPS kernel
	;; Save the BASIC stack, and LCD indicators
	;; Perform a mmps.DWN, create the kernel
	;; console, activate it, and tries to launch
	;; STARTER, the first process.
	;; Enter FATAL error if something fails
mmps.START:

	;; Save DE, the current BASIC pointer
	PUSH	DE

	;; Save LCD indicators
	LDU	(HIGHLCDFLAG)
	PUSH	HL

	;; Save BASIC stack to &7850
	LD	BC,SP
	STS	(50)

	;; Disable the interrupts. Funny things
	;; starts...
	DI
	;; Set stack here
	LD	SP,kernelinitSTACK

	;; Clear LCD indicators
	AND	(HIGHLCDFLAG),00
	AND	(LOWLCDFLAG),00

	;; Perform a system DWN before starting...
	CALL	mmps.ps.dwn_noSP

	;; Current process is kernel itself
	LDA	ps.PID
	STA	(BC)

	;; All kernel-jump should work. If not, we
	;; enter into FATAL vector, because we have
	;; no way to continue !

	;; Create the kernel CONSOLE
	LDR	BC,mmps.CONSOLE
	CALL	mmps.fs.con
	SBR	NC,(93)

	;; Duplicate CONSOLE in/out descriptor
	LDA	E
	STA	D
	PUSH	DE

	;; Select the CONSOLE active
	CALL	mmps.fs.scn
	POP	DE
	SBR	NC,(93)

	;; Present ourself, like "Hello !"
	PUSH	DE
	LDR	BC,mmps.banner
	LD	L,0F
	CALL	mmps.fs.wri
	POP	DE
	SBR	NC,(93)

	;; Now, launch the first process STARTER
	LDR	BC,mmps.STARTER
	LD	HL,BC
	CALL	mmps.ps.xeq
	SBR	NC,(93)

	;; Yes ! We are happy :-D
	CALL	BEEP1

	;; Call the scheduler and like STARTER is
	;; ready, this process will take the hand
	;; And now begin "The Great Adventure"...
	LD	B,<tablePROC
	CLA
	CALL	krn.ps.SCHEDULE

	;; Exit from MMPS kernel after a DWN.
	;; We return here. So we first restore the
	;; BASIC stack, and the LCD indicators and
	;; back to BASIC. We clear the CARRY to
	;; avoid value to be returned from CALL
	LDS	(50)
	LD	SP,BC
	POP	HL
	STU	(HIGHLCDFLAG)
	POP	DE
	RCF

	;; Bye !
	RET

	;; Fatal error vector
	;; Copy all registers into BASIC variable E$
	;; and jump to the ROM RESET
mmps.FATAL:
	;; Disable the interruptions
	DI
	;; Push all registers and status F
	PUSH	HL
	PUSH	DE
	PUSH	BC
	PUSH	A
	LDA	F
	PUSH	A

	;; Get SP
	LD	BC,SP

	;; Set the SYSTEMSTACK as SP
	LD	SP,SYSTEMSTACKBASE

	;; Load the address of BASIC variable E$
	LDR	DE,E$
	INC	BC

	;; Copy all registers from the stack as
	;; binary into E$
	;;   FF AA BBCC DDEE HHLL PCPC SPSP
	LD	L,09
tmp_1_6ed:
	LDI
	DJC	tmp_1_6ed
	DEC	BC
	LDA	B
	STI	(DE)
	LDA	C
	STI	(DE)

	;; Be careful.... Are you ready ?
	CALL	BEEP1

	;; Game over ;-)
	ROMRESET

	.ORIGIN:	16FD
	;; Compute the checksum and put it into a
	;; CPA     (mmps.checksum)
	.CHECKSUM	+(A7)

	.END
